//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#include <IPAddressResolver.h>
#include <IPv4InterfaceData.h>
#include <IPControlInfo_m.h>
#include <UDPControlInfo_m.h>

#include <YMMOGMessage_m.h>
#include "GameClient.h"

Define_Module (GameClient);

void GameClient::initialize (int stage)
{
// because of IPAddressResolver, we need to wait until interfaces are registered,
// address auto-assignment takes place etc.
	
	if (stage != 5)
		return;

	// fetch parameters
	eventCreationDelay = par("eventCreationDelay"); // create a random Event every eventCreationDelay s
	
	// some initialization	
	movement = NULL;		
	playerId = 0;
	xpos = 0;
	ypos = 0;
		
	// timers:
	loginTimer = new cMessage("login");
	moveTimer = new cMessage("move");
	eventCreationTimer = new cMessage("event_creation");
			
	// Delays
	initialDelay= 1;
	moveDelay = 0.2; // Move every moveDelay
	
	// Statistics
	sentMessages = 0;
	receivedMessages = 0;
	sentEvents = 0;
	receivedEvents = 0;
	reachedWaypoints = 0;
	
	// Watches for gui
	WATCH(playerId);
	WATCH(xpos);
	WATCH(ypos);
	WATCH_LIST(playerMap);
	WATCH(receivedMessages);
	WATCH(sentMessages);
	WATCH(reachedWaypoints);
	WATCH(receivedEvents);
	WATCH(sentEvents);
	
	// begin
	scheduleAt(simTime() + initialDelay, loginTimer );

}

void GameClient::handleMessage (cMessage * msg)
{
	if( msg->isName("login") ) {
		if(debug) ev << "Initiate Login\n";
		ClientLogin* clientLogin = new ClientLogin("Clientlogin");
		clientLogin->setCommand(LOGIN);
		sendMsg( clientLogin );
		cancelEvent(loginTimer);
		scheduleAt(simTime() + 60, loginTimer );
	}
	
	else if ( msg->isName("move") ) {
		Point p = movement->move(xpos, ypos);
		xpos = p.x;
		ypos = p.y;
		if( p.reached ) reachedWaypoints++;
		cancelEvent(moveTimer);
		scheduleAt(simTime() + moveDelay, moveTimer);
	}
	
	else if ( msg->isName("event_creation") ) {
		// Send our actual position to the rest of the world.
		ClientEvent* clientEvent = new ClientEvent("publishEvent");
		clientEvent->setEvent(Event(simTime(), playerId, PLAYERPOSITION, xpos, ypos));
		sentEvents++;
		clientEvent->setCommand(PUBLISH);
		sendMsg(clientEvent);
		// Schedule for next Event creation	
		cancelEvent(eventCreationTimer);
		scheduleAt(simTime() + eventCreationDelay, eventCreationTimer);
	}
	
	else if (msg->arrivedOn ("from_lowerTier"))
	{
		receivedMessages++;
		YMMOGMessage *ymmog = (YMMOGMessage *) (msg);
		if (ymmog != NULL)
		{
			if( ymmog->getCommand() == LOGIN ) {
				ClientLogin* clientLogin = (ClientLogin*) ymmog;
				playerId = clientLogin->getPlayerId();
				xpos = clientLogin->getXpos();
				ypos = clientLogin->getYpos();
				gameSpaceSize = clientLogin->getGameSpaceSize();
				movement = new Movement(gameSpaceSize);
				cancelEvent(moveTimer);
				scheduleAt(simTime()+0.1, moveTimer);
				cancelEvent(eventCreationTimer);
				scheduleAt(simTime(), eventCreationTimer);
				cancelEvent(loginTimer);
				delete msg;
			}
			else if (ymmog->getCommand() == PUBLISHEVENTLIST ) {
				ClientEventList* clientEventList = (ClientEventList*) ymmog;
				/** Berechne die Events.... */
				receivedEvents += clientEventList->getEventListArraySize();
				for(unsigned int i=0; i < clientEventList->getEventListArraySize(); ++i) {
					Event event = clientEventList->getEventList(i);
					if( event.type == PLAYERPOSITION ) {
						for(pM=playerMap.begin(); pM != playerMap.end(); ++pM) {
							if(pM->playerId == event.playerId) {
								pM->xpos = event.xpos;
								pM->ypos = event.ypos;
								pM->timeStamp = event.timestamp;
								break;
							}
						}
						if( pM == playerMap.end() ) {
							PlayerPosition player;
							player.playerId = event.playerId;
							player.xpos = event.xpos;
							player.ypos = event.ypos;
							player.timeStamp = event.timestamp;
							playerMap.push_back(player);
						}
					}
				}
				pM=playerMap.begin();
				// Delete too old positions
				while(pM != playerMap.end()) {
					if ( pM->timeStamp < simTime() - 10) {
						pM = playerMap.erase(pM);
					}
					else { ++pM; }
				}
				delete msg;
			}
			else {
				delete msg;
			}
		}
		else {
			delete msg;
		}
	}
}


void GameClient::finish ()
{
	//CleanUp	
	playerMap.clear();
	delete movement;
	
	cancelEvent( loginTimer );
	cancelEvent( moveTimer );
	cancelEvent( eventCreationTimer );
	delete loginTimer;
	delete moveTimer;
	delete eventCreationTimer;
	
}

void GameClient::sendMsg(cMessage* msg) {
	send( msg, "to_lowerTier");
	sentMessages++;
}

std::ostream& operator<<(std::ostream& Stream, const PlayerPosition player)
{
    return Stream << "PlayerId: " << player.playerId << " xpos: " << player.xpos << " ypos: " << player.ypos << " last update:" << player.timeStamp << "\n";
}



