//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Repository.cc
 * @author Helge Backhaus
 */

#include "Repository.h"

Define_Module(Repository);

void Repository::initialize()
{
    Position_size = 0;
    Counter = 0;
    testCounter = 0;
    Position = NULL;

    // fetch parameters
    connectivityCheckTimeout = par("connectivityCheckTimeout");
    connectivityCheckTimeout *= 60.0;

    neighborsList = check_and_cast<NeighborsList*>(parentModule()->submodule("neighborsList"));
    neighborsList->setDebug(false);

    // self-messages
    check_timer = new cMessage("check_timer");
    scheduleAt(simulation.simTime() + connectivityCheckTimeout, check_timer);
}

int selectFunction(cModule *mod, void *)
{
    if(dynamic_cast<Vast*>(mod)) {
        Vast *vast = check_and_cast<Vast*>(mod);
        return vast->getState() == READY;
    }
    return false;
}

void Repository::handleMessage(cMessage* msg)
{
    // catch self timer messages
    if(msg->isName("check_timer")) {
        //reset timer
        cancelEvent(check_timer);
        scheduleAt(simulation.simTime() + connectivityCheckTimeout, msg);
        // handle event
        int i, j, connected, localNeighbors, globalNeighbors, superfluousNeighbors, missingNeighbors;
        std::stringstream vecName, vecName2;
        Vast *vast;
        Site *site;

        vecName << "Neighbor Count " << ++testCounter;
        vecName2 << "Neighbor Deviation " << testCounter;

        cOutVector cOV_NeighborCount(vecName.str().c_str());
        cOutVector cOV_NeighborDeviation(vecName2.str().c_str(), 2);

        topology.extractFromNetwork(selectFunction, NULL);
        // depth-search through topology
        if(topology.nodes()) testConnectivity(topology.node(0));

        connected = 0;
        for(i=0; i<topology.nodes(); i++) {
            if(topology.node(i)->enabled()) {
                connected++;
            }
        }
        recordScalar("Topology disconnected nodes", connected);

        for(i=0; i<topology.nodes(); i++) {
            vast = check_and_cast<Vast*>(topology.node(i)->module());
            neighborsList->initializeList(vast->getPosition(), vast->getThisNode(), vast->getAOI());
            for(j=0; j<topology.nodes(); j++) {
                if(j != i) {
                    vast = check_and_cast<Vast*>(topology.node(j)->module());
                    neighborsList->addNode(vast->getPosition(), vast->getThisNode());
                }
            }
            neighborsList->buildVoronoi();
            neighborsList->removeNeighbors();

            // compare global with local voronoi
            vast = check_and_cast<Vast*>(topology.node(i)->module());

            localNeighbors = vast->getList()->getSize();
            globalNeighbors = neighborsList->getSize();

            // output total neighbors
            cOV_NeighborCount.record((double)localNeighbors);

            // iterate through local voronoi and delete global entries which match
            SiteMap::iterator itSites = vast->getList()->Sites.begin();
            while(itSites != vast->getList()->Sites.end()) {
                neighborsList->removeNode(itSites->second.addr);
                ++itSites;
            }

            // output results (number of neighbors missing in local voronoi, number of superfluous neighbors in local voronoi
            missingNeighbors = neighborsList->getSize();
            superfluousNeighbors = localNeighbors - globalNeighbors + missingNeighbors;
            cOV_NeighborDeviation.record((double)missingNeighbors, (double)superfluousNeighbors);

            // reset the list
            neighborsList->clearList();
        }
        topology.clear();
    }
}

cTopology::Node* Repository::getNodeByHandle(NodeHandle nodeHandle)
{
    Vast *vast;
    for(int i=0; i<topology.nodes(); i++) {
        vast = check_and_cast<Vast*>(topology.node(i)->module());
        if(vast->getThisNode() == nodeHandle) {
            return topology.node(i);
        }
    }
    return NULL;
}

void Repository::testConnectivity(cTopology::Node *node)
{
    Vast *vast;
    Site *site;

    node->disable();
    vast = check_and_cast<Vast*>(node->module());

    SiteMap::iterator itSites = vast->getList()->Sites.begin();
    while(itSites != vast->getList()->Sites.end()) {
        cTopology::Node *newNode;
        newNode = getNodeByHandle(itSites->second.addr);
        if(newNode != NULL && newNode->enabled()) testConnectivity(newNode);
        ++itSites;
    }
}

void Repository::setPosition_size(int size)
{
    Enter_Method_Silent();
    Vector *Temp = (size == 0) ? NULL : new Vector[size];
    int min_size = Position_size < size ? Position_size : size;
    for(int i=0; i<min_size; i++) Temp[i] = this->Position[i];
    Position_size = size;
    delete[] this->Position;
    this->Position = Temp;
}

int Repository::getPosition_size()
{
    Enter_Method_Silent();
    return Position_size;
}

Vector& Repository::getPosition(int k)
{
    Enter_Method_Silent();
    if(k >= Position_size || k < 0) throw new cException("Array out of bounds exception! getPosition(%d)", k);
    return Position[k];
}

void Repository::setPosition(int k, const Vector& Position)
{
    Enter_Method_Silent();
    if(k >= Position_size || k < 0) throw new cException("Array out of bounds exception! setPosition(%d, ...)", k);
    this->Position[k] = Position;
}

int Repository::getCounter()
{
    Enter_Method_Silent();
    return Counter;
}

void Repository::setCounter(int Counter)
{
    Enter_Method_Silent();
    this->Counter = Counter;
}

Repository::~Repository()
{
    // destroy self timer messages
    cancelAndDelete(check_timer);
}
