//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SimpleClient.cc
 * @author Helge Backhaus
 */

#include "SimpleClient.h"

Define_Module(SimpleClient);

void SimpleClient::initializeApp(int stage)
{
    WATCH_MAP(Neighbors);
    WATCH(position);

    // fetch parameters
    moveDelay = par("moveDelay");

    repository = check_and_cast<Repository*>(simulation.moduleByPath("globalObserver.globalFunctions[0].repository"));

    if(strcmp(par("movementGenerator"), "greatGathering") == 0) {
        // debug output
        if(debugOutput) ev << "SIMPLECLIENT: Movement generator: greatGathering" << endl;
        Generator = new greatGathering(par("areaDimension"), par("movementSpeed"), &Neighbors, repository);
    } else if(strcmp(par("movementGenerator"), "groupRoaming") == 0) {
        // debug output
        if(debugOutput) ev << "SIMPLECLIENT: Movement generator: groupRoaming" << endl;
        Generator = new groupRoaming(par("areaDimension"), par("movementSpeed"), &Neighbors, repository, (int)par("groupSize"));
    } else {
        // debug output
        if(debugOutput) {
            if(strcmp(par("movementGenerator"), "randomRoaming") == 0) ev << "SIMPLECLIENT: Movement generator: randomRoaming" << endl;
            else ev << "SIMPLECLIENT: Movement generator: <unknown>. Defaulting to: randomRoaming" << endl;
        }
        Generator = new randomRoaming(par("areaDimension"), par("movementSpeed"), &Neighbors);
    }

    // self-messages
    move_timer = new cMessage("move_timer");

    onlyCommonAPIMessages = false;
}

void SimpleClient::handleTimerEvent(cMessage* msg)
{
    if(msg->isName("move_timer")) {
        //reset timer
        cancelEvent(move_timer);
        scheduleAt(simulation.simTime() + moveDelay, msg);
        // handle event
        Generator->move();
        position = Generator->getPosition();
        SimpleClientMoveMessage *scMsg = new SimpleClientMoveMessage("MOVE_INDICATION");
        scMsg->setCommand(MOVE_INDICATION);
        scMsg->setPosX(position.getX());
        scMsg->setPosY(position.getY());
        scMsg->setLength(SIMPLECLIENTMOVE_L(scMsg));
        sendMessageToOverlay(scMsg);
    }
    else delete msg;
}

void SimpleClient::handleAppMessage(cMessage* msg)
{
    if(dynamic_cast<SimpleClientMessage*>(msg)) {
        SimpleClientMessage* simpleClientMsg = check_and_cast<SimpleClientMessage*>(msg);
        switch(simpleClientMsg->getCommand()) {
            case MOVE_REQUEST: {
                Generator->move();
                position = Generator->getPosition();
                SimpleClientMoveMessage *scMsg = new SimpleClientMoveMessage("MOVE_INDICATION");
                scMsg->setCommand(MOVE_INDICATION);
                scMsg->setPosX(position.getX());
                scMsg->setPosY(position.getY());
                scMsg->setLength(SIMPLECLIENTMOVE_L(scMsg));
                sendMessageToOverlay(scMsg);
            } break;
            case UPDATE_NEIGHBORS: {
                SimpleClientListMessage* simpleClientListMsg = check_and_cast<SimpleClientListMessage*>(msg);
                updateNeighbors(simpleClientListMsg);
            } break;
            case OVERLAY_READY: {
                scheduleAt(simulation.simTime() + moveDelay, move_timer);
            } break;
        }
    }
    delete msg;
}

void SimpleClient::updateNeighbors(SimpleClientListMessage* scMsg)
{
    unsigned int i;
    Vector v;
    for(i=0; i<scMsg->getRemove_neighborArraySize(); i++) {
        Neighbors.erase(scMsg->getRemove_neighbor(i));
    }
    for(i=0; i<scMsg->getAdd_neighborArraySize(); i++) {
        v.setValue(scMsg->getAdd_neighbor_posX(i), scMsg->getAdd_neighbor_posY(i));
        itNeighbors = Neighbors.find(scMsg->getAdd_neighbor(i));
        if(itNeighbors != Neighbors.end()) Neighbors.erase(scMsg->getAdd_neighbor(i));
        Neighbors.insert(std::make_pair(scMsg->getAdd_neighbor(i), v));
    }
}

SimpleClient::~SimpleClient()
{
    // destroy self timer messages
    cancelAndDelete(move_timer);
}
