//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseApp.cc
 * @author Bernhard Heep
 */

#include <IPAddressResolver.h>
#include <CommonAPIMessages_m.h>
#include <BaseOverlayMessage_m.h>

#include "BaseApp.h"
#include "BaseRpc.h"
#include "BootstrapOracleAccess.h"
#include "GlobalStatisticsAccess.h"
#include "UnderlayConfiguratorAccess.h"


BaseApp::~BaseApp()
{
    finishRpcs();
}

int BaseApp::numInitStages() const
{
    return MAX_STAGE_APP + 1;
}

void BaseApp::initialize(int stage)
{
    if(stage == MIN_STAGE_APP) {
        // fetch parameters
        debugOutput = par("debugOutput");
        onlyCommonAPIMessages = true;

        bootstrapOracle = BootstrapOracleAccess().get();
        underlayConfigurator = UnderlayConfiguratorAccess().get();
        globalStatistics = GlobalStatisticsAccess().get();

	overlay = check_and_cast<BaseOverlay*>(gate("to_lowerTier")->destinationGate()->ownerModule());

        // determine the terminal's transport address
        thisNode.ip = IPAddressResolver().addressOf(parentModule()->parentModule()).get4();
	WATCH(thisNode);
        //thisNode.key = //solution?

        // statistics
        numSent = 0;
        numReceived = 0;
        bytesSent = 0;
        bytesReceived = 0;
		creationTime = simTime();

        WATCH(numSent);
        WATCH(numReceived);
        WATCH(bytesSent);
        WATCH(bytesReceived);
    }
    if(stage >= MIN_STAGE_APP && stage <= MAX_STAGE_APP)
        initializeApp(stage);

    // init rpcs
    initRpcs();
}

void BaseApp::initializeApp(int stage)
{
    // ...
}

// Process messages passed up from the overlay.
void BaseApp::handleMessage(cMessage* msg)
{
    // Process self-messages.
    if(msg->isSelfMessage()) {
        // process rpc self-messages
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
        // process all other self-messages
        handleTimerEvent(msg);
        return;
    }
    
    if(msg->arrivedOn("from_lowerTier")) {
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
	// common API
	CommonAPIMessage* commonAPIMsg = dynamic_cast<CommonAPIMessage*>(msg);
	if(commonAPIMsg != NULL)
	    handleCommonAPIMessage(commonAPIMsg);
	    
	else if (onlyCommonAPIMessages == false) {
	    RECORD_STATS(numReceived++; bytesReceived += msg->byteLength());
	    handleAppMessage(msg);
	}
	else
	    delete msg;
    }
    else if(msg->arrivedOn("from_upperTier")) {
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
	handleUpperMessage(msg);
    }
}

void BaseApp::handleTimerEvent(cMessage* msg)
{
    // process self-messages
}

void BaseApp::callRoute(const OverlayKey& key, cMessage* msg, const NodeHandle& hint)
{
    // create route-message (common API)
    KBRroute* routeMsg = new KBRroute();
    routeMsg->setDestKey(key);
    routeMsg->setHint(hint);
    routeMsg->encapsulate(msg);

    routeMsg->setType(KBR_ROUTE);

    send(routeMsg, "to_lowerTier");

    // debug output
    if (debugOutput)
        EV << "(BaseApp::callRoute()) Node " << thisNode.ip << " sent message "
        << id() << "-" << numSent <<  " to destination key "
        << key.toString(16) << "." << endl;

    // count
    RECORD_STATS(numSent++; bytesSent += msg->byteLength());
}

void BaseApp::deliver(OverlayKey& key, cMessage* msg)
{
    // deliver...

    delete msg;
}

void BaseApp::forward(OverlayKey* key, cMessage** msg, NodeHandle* nextHopNode)
{
    // usually do nothing
}

void BaseApp::forwardResponse(const OverlayKey& key, cMessage* msg,
			      const NodeHandle& nextHopNode)
{
    OverlayCtrlInfo* ctrlInfo =
	check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());

    //create forwardResponse message (common API)
    KBRforward* forwardMsg = new KBRforward();
    forwardMsg->setDestKey(key);
    forwardMsg->setNextHopNode(nextHopNode);
    forwardMsg->setControlInfo(ctrlInfo);
    forwardMsg->encapsulate(msg);

    forwardMsg->setType(KBR_FORWARD_RESPONSE);

    send(forwardMsg, "to_lowerTier");
}

void BaseApp::update(const NodeHandle& node, bool joined)
{
    EV << "BaseApp::update(" << node << ", " << joined
       << ") called on node " << thisNode.ip << "." << endl;
}

void BaseApp::handleCommonAPIMessage(CommonAPIMessage* commonAPIMsg)
{
    cMessage* tempMsg = commonAPIMsg->decapsulate();

    // process interface control information
    OverlayCtrlInfo* overlayCtrlInfo =
	dynamic_cast<OverlayCtrlInfo*>(commonAPIMsg->removeControlInfo());
    if(overlayCtrlInfo != NULL) {
	tempMsg->setControlInfo(overlayCtrlInfo);

	if(debugOutput)
	    EV << "(BaseApp) Node " << thisNode.ip << " received message "
	       << " from node " << overlayCtrlInfo->getSrcNode().ip
	       << "." << endl;
    }

    switch (commonAPIMsg->getType()) {

    case KBR_DELIVER:
	{
	    KBRdeliver* apiMsg = dynamic_cast<KBRdeliver*>(commonAPIMsg);
	    OverlayKey key = apiMsg->getDestKey();
	    NodeHandle nextHopNode = thisNode;
	    //first call forward, then deliver
	    forward(&key, &tempMsg, &nextHopNode);

	    if(tempMsg != NULL) {
		//if key or nextHopNode is changed send msg back to overlay
		if(key != apiMsg->getDestKey() ||
		   (!nextHopNode.isUnspecified() &&
		    nextHopNode != thisNode)) {
		    forwardResponse(key, tempMsg, nextHopNode);		
		}
		else {
		    RECORD_STATS(numReceived++; bytesReceived += tempMsg->byteLength());
                    //handle RPC first
		    BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(tempMsg);
		    if (rpcMessage!=NULL) {
			internalHandleRpcMessage(rpcMessage);
		    }
                    else {
        	        deliver(apiMsg->getDestKey(), tempMsg);
                    }
		}			
	    }		
	    break;
	}

    case KBR_FORWARD:
	{
	    KBRforward* apiMsg = dynamic_cast<KBRforward*>(commonAPIMsg);
	    OverlayKey key = apiMsg->getDestKey();
	    NodeHandle nextHopNode = apiMsg->getNextHopNode();

	    forward(&key, &tempMsg, &nextHopNode);

	    //if message ist not deleted send it back
	    if(tempMsg != NULL) {
		if(nextHopNode == apiMsg->getNextHopNode())
		    //do we need this?
		    nextHopNode = NodeHandle::UNSPECIFIED_NODE;
		forwardResponse(key, tempMsg, nextHopNode);		
	    }
	    break;
	}

    case KBR_UPDATE:
	{
	    KBRupdate* apiMsg = dynamic_cast<KBRupdate*>(commonAPIMsg);
	    if(apiMsg->getJoined() &&
	       ((apiMsg->getNode().ip == thisNode.ip && 
		 (thisNode.port == -1 || apiMsg->getNode().port == thisNode.port)) ||
		(!thisNode.key.isUnspecified() && apiMsg->getNode().key == thisNode.key)))
		thisNode = apiMsg->getNode();
	    else
		update(apiMsg->getNode(), apiMsg->getJoined());
		
	    break;
	}

    default:
	{
	    delete tempMsg;
	}
    }
    delete commonAPIMsg;
}

void BaseApp::handleUpperMessage(cMessage* msg)
{
    delete msg;
}

void BaseApp::handleAppMessage(cMessage* msg)
{
    delete msg;
}

void BaseApp::sendMessageToOverlay(cMessage* msg)
{
    RECORD_STATS(numSent++; bytesSent += msg->byteLength());

    send(msg, "to_lowerTier");
}

void BaseApp::finish()
{
    finishRpcs();

    // record scalar data
    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);

    if(time != 0) {    
	globalStatistics->addStdDev("BaseApp: Sent Messages/s", numSent / time);
	globalStatistics->addStdDev("BaseApp: Received Messages/s", numReceived / time);
	globalStatistics->addStdDev("BaseApp: Sent Bytes/s", bytesSent / time);
	globalStatistics->addStdDev("BaseApp: Received Bytes/s", bytesReceived / time);
    }

    finishApp();
}

void BaseApp::finishApp()
{
    // ...
}


void BaseApp::sendRpcMessageToDestination(int destType,
					      const TransportAddress& dest,
					      const OverlayKey& destKey,
					      BaseOverlayMessage* message)
{
    switch(destType) {
      case RPC_TO_UDP:
          callRoute( dynamic_cast<const NodeHandle&>(dest).key, message, dynamic_cast<const NodeHandle&>(dest) );
          break;
      case RPC_TO_KEY:
          callRoute( destKey, message, dynamic_cast<const NodeHandle&>(dest) );
          break;
      case RPC_TO_UPPERTIER:
          send( message, "to_upperTier");
          break;
      case RPC_TO_LOWERTIER:
          send( message, "to_lowerTier");
          break;
    }
}  
