//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseOverlay.cc
 * @author Bernhard Heep (initial)
 * @author Sebastian Mies (rpcs)
 */

#include <UDPAppBase.h>
#include <UDPSocket.h>
#include <UDPControlInfo_m.h>
#include <IPAddressResolver.h>

#include <CommonAPIMessages_m.h>
#include <BootstrapOracleAccess.h>
#include <UnderlayConfiguratorAccess.h>
#include <GlobalStatisticsAccess.h>


#include "BaseOverlay.h"
#include "RpcMacros.h"

enum {
    NOT_FORWARDED,
    ALREADY_FORWARDED
};

//------------------------------------------------------------------------
//--- Initialization & finishing -----------------------------------------
//------------------------------------------------------------------------

BaseOverlay::~BaseOverlay()
{
    finishLookups();
    finishRpcs();
}

int BaseOverlay::numInitStages() const
{
    return MAX_STAGE_OVERLAY + 1;
}

void BaseOverlay::initialize(int stage)
{
    if(stage == MIN_STAGE_OVERLAY) {
        OverlayKey::setKeyLength(par("keyLength"));

        // fetch some parameters
        debugOutput                     = par("debugOutput");
        measureNetwInitPhase            = par("measureNetwInitPhase");
        localPort                       = par("localPort");
        hopCountMax                     = par("hopCountMax");
        drawOverlayTopology             = par("drawOverlayTopology");
        useBaseLookup                   = par("useBaseLookup");
        iterativeLookup                 = par("iterativeLookup");
        useCommonAPIforward             = false;
        onlyCommonAPIMessages           = true;

        // set base lookup parameters
        baseLookupConfig.redundantNodes = par("lookupRedundantNodes");
        baseLookupConfig.parallelPaths  = par("lookupParallelPaths");
        baseLookupConfig.parallelRpcs   = par("lookupParallelRpcs");
        baseLookupConfig.secure         = par("lookupSecure");
        baseLookupConfig.merge          = par("lookupMerge");
	baseLookupConfig.failedNodeRpcs	= par("lookupFailedNodeRpcs");

        // statistics
        numSent = 0;
        bytesSent = 0;
        numSignalingSent = 0;
        bytesSignalingSent = 0;
        numReceived = 0;
        bytesReceived = 0;
        numSignalingReceived = 0;
        bytesSignalingReceived = 0;
        numForwarded = 0;
        bytesForwarded = 0;
        numSignalingForwarded = 0;
        bytesSignalingForwarded = 0;
        numDropped = 0;
        bytesDropped = 0;
	numPingSent = 0;
	bytesPingSent = 0;
	numPingResponseSent = 0;
	bytesPingResponseSent = 0;
	numFindNodeSent = 0;
	bytesFindNodeSent = 0;
	numFindNodeResponseSent = 0;
	bytesFindNodeResponseSent = 0;
	numFailedNodeSent = 0;
	bytesFailedNodeSent = 0;
	numFailedNodeResponseSent = 0;
	bytesFailedNodeResponseSent = 0;

        WATCH(numSent);
        WATCH(bytesSent);
        WATCH(numSignalingSent);
        WATCH(bytesSignalingSent);
        WATCH(numReceived);
        WATCH(bytesReceived);
        WATCH(numSignalingReceived);
        WATCH(bytesSignalingReceived);
        WATCH(numForwarded);
        WATCH(bytesForwarded);
        WATCH(numSignalingForwarded);
        WATCH(bytesSignalingForwarded);
        WATCH(numDropped);
        WATCH(bytesDropped);
	WATCH(numPingSent);
	WATCH(bytesPingSent);
	WATCH(numPingResponseSent);
	WATCH(bytesPingResponseSent);
	WATCH(numFindNodeSent);
	WATCH(bytesFindNodeSent);
	WATCH(numFindNodeResponseSent);
	WATCH(bytesFindNodeResponseSent);
	WATCH(numFailedNodeSent);
	WATCH(bytesFailedNodeSent);
	WATCH(numFailedNodeResponseSent);
	WATCH(bytesFailedNodeResponseSent);

        // set up UDP
        bindToPort(localPort);

        // find friend modules
        bootstrapOracle = BootstrapOracleAccess().get();
        underlayConfigurator = UnderlayConfiguratorAccess().get();
        globalStatistics = GlobalStatisticsAccess().get();
        notificationBoard = NotificationBoardAccess().get();

        thisTerminal = parentModule()->parentModule();

        // set up local nodehandle
        //thisNode.moduleId = parentModule()->id();
        thisNode.ip = IPAddressResolver().
                      addressOf(parentModule()->parentModule()).get4();
        thisNode.port = localPort;
        thisNode.key = OverlayKey::UNSPECIFIED_KEY;

        // set up arrow-gates
        thisOutGateArray = thisTerminal->gate("overlayNeighborArrowOut");
        thisInGateArray = thisTerminal->gate("overlayNeighborArrowIn");
        thisTerminal->setGateSize("overlayNeighborArrowOut" ,1);
        thisTerminal->setGateSize("overlayNeighborArrowIn" ,1);

        // subscribe to the notification board
        notificationBoard->subscribe( this, NF_HOSTPOSITION_UPDATED);
        notificationBoard->subscribe( this, NF_OVERLAY_NODE_LEAVE);

        // init rpcs
        initRpcs();
        initLookups();

	// statistics
	globalStatistics->nodesInitialized++;
	creationTime = simTime();
    }


    if (stage >= MIN_STAGE_OVERLAY && stage <= MAX_STAGE_OVERLAY)
        initializeOverlay(stage);

    if (stage == MAX_STAGE_OVERLAY) {
//    	setReadyIcon(false);
	if ((bool)par("joinOnApplicationRequest") == false) {
	    join();
	}
    }
}

void BaseOverlay::initializeOverlay(int stage)
{
    // ...
}

void BaseOverlay::finish()
{
    finishOverlay();
    finishLookups();
    finishRpcs();

    globalStatistics->nodesFinished++;

    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);
    if(time == 0) return;

    globalStatistics->addStdDev("BaseOverlay: Sent User Messages/s", numSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent User Bytes/s", bytesSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Signaling Messages/s", numSignalingSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Signaling Bytes/s", bytesSignalingSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Total Messages", numSent + numSignalingSent);
    globalStatistics->addStdDev("BaseOverlay: Sent Total Bytes", bytesSent + bytesSignalingSent);
    globalStatistics->addStdDev("BaseOverlay: Sent FindNode Messages/s", numFindNodeSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FindNode Bytes/s", bytesFindNodeSent / time);

    globalStatistics->addStdDev("BaseOverlay: Sent FindNodeResponse Messages/s", numFindNodeResponseSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FindNodeResponse Bytes/s", bytesFindNodeResponseSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNode Messages/s", numFailedNodeSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNode Bytes/s", bytesFailedNodeSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNodeResponse Messages/s", numFailedNodeResponseSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNodeResponse Bytes/s", bytesFailedNodeResponseSent / time);

    globalStatistics->addStdDev("BaseOverlay: Received User Messages/s", numReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received User Bytes/s", bytesReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received Signaling Messages/s", numSignalingReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received Signaling Bytes/s", bytesSignalingReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received Total Messages", numReceived
                 + numSignalingReceived);
    globalStatistics->addStdDev("BaseOverlay: Received Total Bytes", bytesReceived
                 + bytesSignalingReceived);
    globalStatistics->addStdDev("BaseOverlay: Forwarded User Messages/s", numForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded User Bytes/s", bytesForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Signaling Messages/s", numSignalingForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Signaling Bytes/s", bytesSignalingForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Total Messages", numForwarded
                 + numSignalingForwarded);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Total Bytes", bytesForwarded
                 + bytesSignalingForwarded);

    globalStatistics->addStdDev("BaseOverlay: Dropped Messages/s", numDropped / time);
    globalStatistics->addStdDev("BaseOverlay: Dropped Bytes/s", bytesDropped / time);

    globalStatistics->addStdDev("BaseOverlay: Active Time", simTime() - creationTime);

    globalStatistics->doFinish();

}

void BaseOverlay::finishOverlay()
{
    // ...
}

//------------------------------------------------------------------------
//--- General Overlay Parameters (getter and setters) --------------------
//------------------------------------------------------------------------
bool BaseOverlay::isMalicious()
{
    return bootstrapOracle->isMalicious(getThisNode());
}

NodeHandle& BaseOverlay::getThisNode()
{
    return thisNode;
}

//------------------------------------------------------------------------
//--- UDP functions copied from the INET framework .----------------------
//------------------------------------------------------------------------
void BaseOverlay::bindToPort(int port)
{
    EV << "Binding to UDP port " << port << endl;

    // TODO UDPAppBase should be ported to use UDPSocket sometime, but for now
    // we just manage the UDP socket by hand...
    cMessage *msg = new cMessage("UDP_C_BIND", UDP_C_BIND);
    UDPControlInfo *ctrl = new UDPControlInfo();
    ctrl->setSrcPort(port);
    ctrl->setSockId(UDPSocket::generateSocketId());
    msg->setControlInfo(ctrl);
    send(msg, "to_udp");
}

void BaseOverlay::sendToUDP(cMessage *msg, int srcPort, const IPvXAddress& destAddr,
			    int destPort)
{
    // send message to UDP, with the appropriate control info attached
    msg->setKind(UDP_C_DATA);

    UDPControlInfo *ctrl = new UDPControlInfo();
    ctrl->setSrcPort(srcPort);
    ctrl->setDestAddr(destAddr);
    ctrl->setDestPort(destPort);
    msg->setControlInfo(ctrl);

    EV << "Sending packet: ";
    printPacket(msg);

    send(msg, "to_udp");
}

void BaseOverlay::printPacket(cMessage *msg)
{
    UDPControlInfo *ctrl = check_and_cast<UDPControlInfo *>(msg->controlInfo());

    IPvXAddress srcAddr = ctrl->srcAddr();
    IPvXAddress destAddr = ctrl->destAddr();
    int srcPort = ctrl->srcPort();
    int destPort = ctrl->destPort();

    ev  << msg << "  (" << msg->byteLength() << " bytes)" << endl;
    ev  << srcAddr << " :" << srcPort << " --> " << destAddr << ":" << destPort 
<< endl;
}




//------------------------------------------------------------------------
//--- Overlay Common API: Key-based Routing ------------------------------
//------------------------------------------------------------------------

void BaseOverlay::route( const OverlayKey& key, cMessage* msg,
                         const NodeHandle& hint )
{}


void BaseOverlay::callDeliver(BaseOverlayMessage* msg,
			      const OverlayKey& destKey)
{
    KBRdeliver* deliverMsg = new KBRdeliver();

    OverlayCtrlInfo* overlayCtrlInfo =
        check_and_cast<OverlayCtrlInfo*>
        (msg->removeControlInfo());

    deliverMsg->setControlInfo(overlayCtrlInfo);
    deliverMsg->setDestKey(destKey);

    deliverMsg->encapsulate(msg->decapsulate());

    deliverMsg->setType(KBR_DELIVER);

    send(deliverMsg, "to_app");

    delete msg;
}

void BaseOverlay::callForward(const OverlayKey& key, BaseRouteMessage* msg,
                              const NodeHandle& nextHopNode)
{
    KBRforward* forwardMsg = new KBRforward();

    forwardMsg->setDestKey(msg->getDestKey());
    forwardMsg->setNextHopNode(nextHopNode);
    forwardMsg->encapsulate(msg->encapsulatedMsg()->decapsulate());

    OverlayCtrlInfo* overlayCtrlInfo =
        new OverlayCtrlInfo();
    //overlayCtrlInfo->setThisNode(thisNode);
    overlayCtrlInfo->setHopCount(msg->getHopCount());
    overlayCtrlInfo->setSrcNode(msg->getSrcNode());

    if(msg->controlInfo() != NULL) {
	OverlayCtrlInfo* ctrlInfo =
	    check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());
	overlayCtrlInfo->setLastHopAddr(ctrlInfo->getLastHopAddr());
	overlayCtrlInfo->setLastHopPort(ctrlInfo->getLastHopPort());

	delete ctrlInfo;
    }

    forwardMsg->setControlInfo(overlayCtrlInfo);

    forwardMsg->setType(KBR_FORWARD);

    send(forwardMsg, "to_app");

    delete msg;
}

NodeVector* BaseOverlay::local_lookup(const OverlayKey& key,
				      int num, bool safe)
{
    Enter_Method("local_lookup()");

    if (safe == true) {
	error("BaseOverlay::local_lookup(): safe flag is not implemented!");
    }

    NodeVector* nodeVector = findNode(key, num, num);

    if(((int)nodeVector->size()) > num)
	nodeVector->resize(num);

    return nodeVector;
}

void BaseOverlay::join(const OverlayKey& nodeID)
{
    Enter_Method("join()");

    // set nodeID and IP
    thisNode.ip = 
        IPAddressResolver().addressOf(parentModule()->parentModule()).get4();
    if (nodeID.isUnspecified())  {
        thisNode.key = OverlayKey::random();
    }

    callUpdate(thisNode, true);

    joinOverlay();
}


NodeVector* BaseOverlay::neighborSet(int num)
{
    Enter_Method("neighborSet()");

    return local_lookup(thisNode.key, num, false);
}


void BaseOverlay::callUpdate(const NodeHandle& node, bool joined)
{
    KBRupdate* updateMsg = new KBRupdate("UPDATE");

    updateMsg->setNode(node);
    updateMsg->setJoined(joined);

    updateMsg->setType(KBR_UPDATE);

    send(updateMsg, "to_app");//"to_upperTier");
}

bool BaseOverlay::isSiblingFor(const NodeHandle& node, const OverlayKey& key,
			    int numSiblings, bool* err)
{
    Enter_Method("isSiblingFor()");

    opp_error( "isSiblingFor: Not implemented!" );

    return false;
}

int BaseOverlay::getMaxNumSiblings()
{
    Enter_Method("getMaxNumSiblings()");

    opp_error( "getMaxNumSiblings: Not implemented!" );

    return false;
}

int BaseOverlay::getMaxNumRedundantNodes()
{
    Enter_Method("getMaxNumRedundantNodes()");

    opp_error( "getMaxNumRedundantNodes: Not implemented!" );

    return false;
}


//------------------------------------------------------------------------
//--- Message Handlers ---------------------------------------------------
//------------------------------------------------------------------------

//private
void BaseOverlay::handleMessage(cMessage* msg)
{
    // process self-messages
    if (msg->isSelfMessage()) {
        // process rpc self-messages
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
        // process all other self-messages
        handleTimerEvent(msg);
    }

    // process messages from UDP
    else if (msg->arrivedOn("from_udp")) {
        UDPControlInfo* udpControlInfo =
            check_and_cast<UDPControlInfo*>(msg->removeControlInfo());
        OverlayCtrlInfo* overlayCtrlInfo = new OverlayCtrlInfo;
        overlayCtrlInfo->setLastHopAddr(udpControlInfo->srcAddr());
        overlayCtrlInfo->setLastHopPort(udpControlInfo->srcPort());
        msg->setControlInfo(overlayCtrlInfo);
        delete udpControlInfo;

        BaseOverlayMessage* baseOverlayMsg =
            dynamic_cast<BaseOverlayMessage*>(msg);

        if (baseOverlayMsg == NULL) {
            delete msg;
            return;
        }

        // records stats if message is not a UDP "self message"
        if (overlayCtrlInfo->getLastHopAddr() != thisNode.ip) {
            if (baseOverlayMsg->getSignaling() == false)
                RECORD_STATS(numReceived++; bytesReceived +=
                                 baseOverlayMsg->byteLength());
            else
                RECORD_STATS(numSignalingReceived++;bytesSignalingReceived
                             += baseOverlayMsg->byteLength());
        }

        handleBaseOverlayMessage(baseOverlayMsg);
    }

    // process CommonAPIMessages from App
    else if (dynamic_cast<CommonAPIMessage*>(msg) != NULL) {
        if (dynamic_cast<KBRroute*>(msg) != NULL) {
	    KBRroute* apiMsg = dynamic_cast<KBRroute*>(msg);

            internalRoute(apiMsg->getDestKey(), apiMsg->decapsulate(),
			  apiMsg->getHint());
	}

        else if (dynamic_cast<KBRforward*>(msg) != NULL) {
	    KBRforward* apiMsg = dynamic_cast<KBRforward*>(msg);
	    OverlayCtrlInfo* overlayCtrlInfo =
		check_and_cast<OverlayCtrlInfo*>
		(msg->removeControlInfo());

	    BaseAppDataMessage* dataMsg =
		new BaseAppDataMessage();
	    dataMsg->setType(APPDATA);
	    dataMsg->setLength(BASEAPPDATA_L(dataMsg));
	    dataMsg->setName(msg->encapsulatedMsg()->name());
	    dataMsg->encapsulate(msg->decapsulate());
	    dataMsg->setSignaling(false);

	    BaseRouteMessage* routeMsg = new BaseRouteMessage(dataMsg->name());
	    routeMsg->setType(OVERLAYROUTE);
	    routeMsg->setLength(BASEROUTE_L(routeMsg));
	    routeMsg->encapsulate(dataMsg);

	    routeMsg->setSignaling(false);
	    routeMsg->setDestKey(apiMsg->getDestKey());
	    routeMsg->setSrcNode(overlayCtrlInfo->getSrcNode());
	    routeMsg->setHopCount(overlayCtrlInfo->getHopCount());
	    routeMsg->setControlInfo(overlayCtrlInfo);

	    routeMsg->setKind(ALREADY_FORWARDED);

            sendToKey(apiMsg->getDestKey(), routeMsg, 1, apiMsg->getNextHopNode());
	}

        delete msg;
    } 

    else if (dynamic_cast<BaseRpcMessage*>(msg) != NULL) {
            BaseRpcMessage* rpcMessage =
                dynamic_cast<BaseRpcMessage*>(msg);
            internalHandleRpcMessage(rpcMessage);
    }

    // process other messages from App
    else if (msg->arrivedOn("from_app") && onlyCommonAPIMessages == false)
        handleAppMessage(msg);

    else {
	opp_error("BaseOverlay::handleMessage(): Received msg with "
		  "unknown type!");
//        RECORD_STATS(numDropped++; bytesDropped += msg->byteLength());
        delete msg;
    }
}

void BaseOverlay::handleBaseOverlayMessage(BaseOverlayMessage* msg,
					   const OverlayKey& destKey)
{
    switch(msg->getType()) {
    case OVERLAYSIGNALING:
        handleUDPMessage(msg);
        return;

    case RPC: {
            // process rpc-messages
            BaseRpcMessage* rpcMessage =
                check_and_cast<BaseRpcMessage*>(msg);
            internalHandleRpcMessage(rpcMessage);
            return;
        }

    case APPDATA: {
            BaseAppDataMessage* baseAppDataMsg =
                check_and_cast<BaseAppDataMessage*>(msg);
	    assert(!destKey.isUnspecified());
            callDeliver(baseAppDataMsg, destKey);
            return;
        }

    case OVERLAYROUTE: {
            BaseRouteMessage* baseRouteMsg =
                check_and_cast<BaseRouteMessage*>(msg);

            if (baseRouteMsg->getDestKey().isUnspecified())
                opp_error("BaseOverlay: No destination key in ROUTE msg");


	    // if this node is malicious drop the message
	    if (isMalicious()) {
		    EV << "BaseOverlay: BaseRouteMessage gets dropped "
		       << "because this node is malicious!" << endl;
		    RECORD_STATS(numDropped++;
				 bytesDropped += baseRouteMsg->byteLength());
		    delete msg;
		    return;
	    }

	    bool err;
	    if (isSiblingFor(thisNode, baseRouteMsg->getDestKey(),
				1, &err)) {
                OverlayCtrlInfo* overlayCtrlInfo =
                    check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());
		
                overlayCtrlInfo->setHopCount(baseRouteMsg->getHopCount());
                overlayCtrlInfo->setSrcNode(baseRouteMsg->getSrcNode());

                BaseOverlayMessage* tmpMsg
                = check_and_cast<BaseOverlayMessage*>
                  (baseRouteMsg->decapsulate());
                tmpMsg->setControlInfo(overlayCtrlInfo);

                handleBaseOverlayMessage(tmpMsg, baseRouteMsg->getDestKey());//bh
                delete msg;
                return;
            } else {
                // forward msg if this node is not responsible for the key
                sendToKey(baseRouteMsg->getDestKey(), baseRouteMsg, 1);
                return;
            }
            break;
        }

    default:
        EV << "BaseOverlay::handleMessage(): received unknown message "
        << "from UDP of type " << msg->name() << endl;
        break;
    }
}


//virtual protected
void BaseOverlay::handleTimerEvent(cMessage* msg)
{
    // timer...
}

//virtual protected
void BaseOverlay::handleAppMessage(cMessage* msg)
{
    delete msg;
}

//virtual protected
void BaseOverlay::recordOverlaySentStats(BaseOverlayMessage* msg)
{
    // collect statistics ...
}

//------------------------------------------------------------------------
//--- Overlay Topology Drawing -------------------------------------------
//------------------------------------------------------------------------

void BaseOverlay::setReadyIcon(bool ready)
{
	bootstrapOracle->setOverlayReadyIcon(getThisNode(), ready);
}

void BaseOverlay::showOverlayNeighborArrow(const NodeHandle& neighbor,
        bool flush, char* displayString)
{
    if (!ev.isGUI() || !drawOverlayTopology)
        return;

    char red[] = "o=red,1";
    if (displayString == NULL)
        displayString = red;

    cModule* neighborTerminal;
    cGate* neighborInGateArray;

    // flush
    if (flush) {
        for (int l = 0; l < thisOutGateArray->size(); l++) {
            cGate* tempGate
            = thisTerminal->gate("overlayNeighborArrowOut", l)->toGate();
            thisTerminal->gate("overlayNeighborArrowOut", l)->disconnect();
            if (tempGate != NULL)
                compactGateArray(tempGate->ownerModule(), IN);
        }
        thisTerminal->setGateSize("overlayNeighborArrowOut" ,0);
    }

    int neighborModuleId;
    if(bootstrapOracle->getPeerInfo(neighbor) == NULL)
        return;
    else
        neighborModuleId = bootstrapOracle->getPeerInfo(neighbor)->getModuleID();

    if (simulation.module(neighborModuleId) != NULL) {
        neighborTerminal
        = simulation.module(neighborModuleId)->parentModule();
        neighborInGateArray = simulation.module(neighborModuleId)->
                              parentModule()->gate("overlayNeighborArrowIn");
    } else
        return;

    //if (neighborInGateArray == NULL)
	//return;

    if (thisTerminal == neighborTerminal)
        return;

    //do not draw double
    for (int i = 0; i < thisOutGateArray->size(); i++)
        if (thisTerminal->gate("overlayNeighborArrowOut", i)->toGate() != NULL
                && neighborTerminal
                == thisTerminal->gate("overlayNeighborArrowOut", i)->
                toGate()->ownerModule())
            return;

    // IN
    int i = 0;
    if (neighborInGateArray->size() == 0) {
        neighborTerminal->setGateSize("overlayNeighborArrowIn", 1);
	}
    else {
        for (i = 0; i < neighborInGateArray->size() - 1; i++) {
            if (!(neighborTerminal->gate("overlayNeighborArrowIn", i)
                    ->isConnectedOutside()))
                break;
        }
        if (neighborTerminal->gate("overlayNeighborArrowIn", i)
                ->isConnectedOutside()) {
            neighborTerminal->setGateSize("overlayNeighborArrowIn", i + 2);
            i++;
        }
    }

    // OUT
    int j = 0;
    if (thisOutGateArray->size() == 0)
        thisTerminal->setGateSize("overlayNeighborArrowOut", 1);
    else {
        for (j = 0; j < (thisOutGateArray->size() - 1); j++) {
            if (!(thisTerminal->gate("overlayNeighborArrowOut", j)
                    ->isConnectedOutside()))
                break;
        }
        if (thisTerminal->gate("overlayNeighborArrowOut", j)
                ->isConnectedOutside()) {
            thisTerminal->setGateSize("overlayNeighborArrowOut", j + 2);
            j++;
        }
    }

    thisTerminal->gate("overlayNeighborArrowOut", j)->
    connectTo(neighborTerminal->gate("overlayNeighborArrowIn", i));

    thisTerminal->gate("overlayNeighborArrowOut", j)->
    setDisplayString(displayString);
}

void BaseOverlay::deleteOverlayNeighborArrow(const NodeHandle& neighbor)
{
    if (!ev.isGUI() || !drawOverlayTopology)
        return;

    //does neighbor module exist anymore?
    int neighborModuleId;
    if(bootstrapOracle->getPeerInfo(neighbor) == NULL)
        return;
    else
        neighborModuleId = bootstrapOracle->getPeerInfo(neighbor)->getModuleID();

    if (simulation.module(neighborModuleId) == NULL)
        return;

    cModule* neighborTerminal
    = simulation.module(neighborModuleId)->parentModule();

    //find gate
    bool compactOut = false;
    bool compactIn = false;
    for (int i = 0; i < thisOutGateArray->size(); i++) {
        // NULL-Gate?
        if (thisTerminal->gate("overlayNeighborArrowOut", i)->toGate() == NULL) {
            compactOut = true;
            continue;
        }

        if (thisTerminal->gate("overlayNeighborArrowOut", i)->toGate()->ownerModule()->id() == neighborTerminal->id()) {
            thisTerminal->gate("overlayNeighborArrowOut", i)->disconnect();
            compactOut = true;
            compactIn = true;
        }
    }

    //compact OUT-array
    if (compactOut)
        compactGateArray(thisTerminal, OUT);
    //compact IN-array
    if (compactIn)
        compactGateArray(neighborTerminal, IN);
}

void BaseOverlay::compactGateArray(cModule* terminal, enum Direction dir)
{
    cGate* gateArray = (dir == OUT ? terminal->gate("overlayNeighborArrowOut")
                        : terminal->gate("overlayNeighborArrowIn"));
    const char* gateName = (dir == OUT ? "overlayNeighborArrowOut"
                            : "overlayNeighborArrowIn");

    for (int j = 0; j < gateArray->size() - 1; j++) {
        if (terminal->gate(gateName, j)->isConnectedOutside())
            continue;

        cGate* tempGate = NULL;
        int k = 1;
        while ((tempGate == NULL) && ((j + k) != gateArray->size())) {
            tempGate = (dir == OUT ? terminal->gate(gateName, j + k)->toGate()
                        : terminal->gate(gateName, j + k)->fromGate());
            k++;
        }

        if (tempGate == NULL)
            break;

        if (dir == OUT) {
            terminal->gate(gateName, j + k - 1)->disconnect();
            terminal->gate(gateName, j)->connectTo(tempGate);
        } else {
            tempGate->disconnect();
            tempGate->connectTo(terminal->gate(gateName, j));
        }
    }

    int nullGates = 0;
    for (int j = 0; j < gateArray->size(); j++)
        if (!terminal->gate(gateName, j)->isConnectedOutside())
            nullGates++;

    terminal->setGateSize(gateName, gateArray->size() - nullGates);
}

//------------------------------------------------------------------------
//--- Messages -----------------------------------------------------------
//------------------------------------------------------------------------

void BaseOverlay::sendMessageToUDP(const TransportAddress& dest,
                                   BaseOverlayMessage* msg)
{
    if (!iterativeLookup)
    {
	// increase HopCount on outgoing RouteMessages in recursive mode
	// moved here from sendToKey() so that hopCount is only increased
	// when the Overlay decides to actually send the Message in
	// forwardMessageRecursive().
	BaseRouteMessage* routeMsg = dynamic_cast<BaseRouteMessage*>(msg);
	if (routeMsg && (dest != thisNode))
	{
	    routeMsg->setHopCount(routeMsg->getHopCount() + 1);
	}
    }

    // if there's still a control info attached to the message, remove it
    cPolymorphic* ctrlInfo = msg->removeControlInfo();
    if (ctrlInfo != NULL)
        delete ctrlInfo;

    // debug message
    if (debugOutput) {
        EV << "BaseOverlay: Node " << thisNode.ip << ": sends "
        << msg->name() << " to " << dest.ip << "." << endl;
    }

    msg->setKind(UDP_C_DATA);
    UDPControlInfo* udpControlInfo = new UDPControlInfo();
    udpControlInfo->setSrcAddr(thisNode.ip);
    udpControlInfo->setSrcPort(thisNode.port);
    udpControlInfo->setDestAddr(dest.ip);
    udpControlInfo->setDestPort(dest.port);
    msg->setControlInfo(udpControlInfo);

    send(msg, "to_udp");

    if (dest != thisNode) {
        // record statistics, if message is not local
        if (msg->getSignaling() == false) {
            RECORD_STATS(numSent++; bytesSent += msg->byteLength());
        } else {
            RECORD_STATS(numSignalingSent++; bytesSignalingSent +=
                             msg->byteLength());
        }
        recordOverlaySentStats(msg);
    }
}


//------------------------------------------------------------------------
//--- RPC Handling -------------------------------------------------------
//------------------------------------------------------------------------

//private
PingCall* BaseOverlay::createPingCall( const char* caption )
{
    PingCall* call;
    if ( caption )
    {
	call = new PingCall(caption);
    }
    else
    {
	call = new PingCall("PING");
    }
    call->setLength(PINGCALL_L(call));
    return call;
}

//public
void BaseOverlay::pingNode( const TransportAddress& dest, simtime_t timeout,
	int retries, const char* caption, RpcListener* rpcListener, int rpcId )
{
    PingCall* call = createPingCall(caption);
    RECORD_STATS(numPingSent++; bytesPingSent += call->byteLength());
    sendRpcMessage(dest, call, rpcListener, OverlayKey::UNSPECIFIED_KEY,
	    rpcId, timeout, retries);
}

//public
void BaseOverlay::pingKey( const OverlayKey& destKey, simtime_t timeout,
	int retries, const char* caption, RpcListener* rpcListener, int rpcId )
{
    PingCall* call = createPingCall(caption);
    RECORD_STATS(numPingSent++; bytesPingSent += call->byteLength());
    sendRpcMessage(TransportAddress::UNSPECIFIED_NODE, call, rpcListener,
	    destKey, rpcId, timeout, retries);
}


//------------------------------------------------------------------------
//--- Basic Routing ------------------------------------------------------
//------------------------------------------------------------------------

static int pendingLookups = 0;

void BaseOverlay::initLookups()
{
    lookups = LookupSet();
}

void BaseOverlay::finishLookups()
{
    while (lookups.size()!=0)
        delete *lookups.begin();
    lookups.clear();
}

class SendToKeyListener : public LookupListener
{
private:
    BaseOverlay* overlay;
    BaseOverlayMessage* msg;
public:
    SendToKeyListener( BaseOverlay* overlay, BaseOverlayMessage* msg )
    {
        this->overlay = overlay;
        this->msg = msg;
        pendingLookups++;
    }

    ~SendToKeyListener()
    {
        this->overlay = NULL;
        this->msg = NULL;
    }

    virtual void lookupFinished(AbstractLookup *lookup)
    {
        pendingLookups--;

	if (dynamic_cast<BaseRouteMessage*>(msg)) {
	    BaseRouteMessage* routeMsg = static_cast<BaseRouteMessage*>(msg);
	    if (lookup->isValid()) {
		if (lookup->getResult().size()==0) {
		    cout << "[ERROR] SendToKeyListener: Valid result, "
			"but empty array." << endl;
		} else {
		    routeMsg->setHopCount(routeMsg->getHopCount()
				     + lookup->getAccumulatedHops());

		    for (uint i=0; i<lookup->getResult().size(); i++) {
			overlay->sendMessageToUDP(lookup->getResult()[i],
						  (BaseOverlayMessage*)msg->dup());
		    }
		}
	    } else {
		EV << "BaseOverlay: Lookup failed!" << endl;
	    }
	    delete msg;
	} else if (dynamic_cast<LookupCall*>(msg)) {
	    LookupCall* call = static_cast<LookupCall*>(msg);
	    LookupResponse* response = new LookupResponse();

	    if(lookup->isValid()) {
		response->setIsValid(true);
		response->setKey(call->getKey());
		response->setSiblingsArraySize(lookup->getResult().size());
		for (uint i=0; i<lookup->getResult().size(); i++) {
		    response->setSiblings(i, lookup->getResult()[i]);
		}
	    } else {
		response->setIsValid(false);
	    }
	    overlay->sendRpcResponse(RPC_TO_UPPERTIER, call, response);

	} else {
	    opp_error("SendToKeyListener::lookupFinished(): "
		      "Unkown message type!");
	}
	delete this;
    }
};

void BaseOverlay::internalRoute(const OverlayKey& key, cMessage* msg,
                                const NodeHandle& hint)
{
    // check if base lookup should be used
    if (!useBaseLookup) {
        route(key,msg,hint);
        return;
    }

    if(key.isUnspecified())
        error("route(): key unspecified!");

    // create base route message
    BaseAppDataMessage* baseAppDataMsg =
        new BaseAppDataMessage("BaseAppDataMessage");
    baseAppDataMsg->setType(APPDATA);
    baseAppDataMsg->setLength(BASEAPPDATA_L(baseAppDataMsg));
    baseAppDataMsg->setSignaling(false);
    //baseAppDataMsg->setKind(1);//???
    baseAppDataMsg->encapsulate(msg);

    // debug output
    if (debugOutput) {
        EV << "BaseOverlay: Node " << thisNode.ip << " received message from "
        "application." << endl;
    }

    sendToKey( key, baseAppDataMsg, 1, hint );
}

void BaseOverlay::forwardMessageRecursive( const TransportAddress& dest,
	BaseRouteMessage* msg)
{
    sendMessageToUDP(dest, msg);
}

void BaseOverlay::sendToKey(const OverlayKey& key, BaseOverlayMessage* msg,
                            int numSiblings, const TransportAddress& nextHop)
{
    BaseRouteMessage* routeMsg = NULL;

    if (key.isUnspecified())
        error("BaseOverlay::sendToKey(): unspecified destination key!");

    if (msg->getType() != OVERLAYROUTE) {
        routeMsg = new BaseRouteMessage("BaseRouteMessage");
        routeMsg->setType(OVERLAYROUTE);
        routeMsg->setDestKey(key);
        routeMsg->setSrcNode(thisNode);
        routeMsg->setSignaling(msg->getSignaling());
        //routeMsg->setKind(routeMsg->getSignaling() ? 0 : 1);//???
	routeMsg->setKind(NOT_FORWARDED);
        // copy the name of the inner message
        routeMsg->setName(msg->name());
        routeMsg->setLength(BASEROUTE_L(routeMsg));
        routeMsg->encapsulate(msg);

	OverlayCtrlInfo* routeCtrlInfo = new OverlayCtrlInfo;
	routeCtrlInfo->setLastHopAddr(thisNode.ip);
	routeCtrlInfo->setLastHopPort(thisNode.port);

	routeMsg->setControlInfo(routeCtrlInfo);
    } else {
        routeMsg = check_and_cast<BaseRouteMessage*>(msg);
    }

    if (!nextHop.isUnspecified()) {
        // send msg to nextHop if specified (used for join rpcs)
	if (iterativeLookup && (nextHop != thisNode))
	{
	    routeMsg->setHopCount(routeMsg->getHopCount() + 1);
	}
        sendMessageToUDP(nextHop, routeMsg);
        return;
    }

    if (iterativeLookup) {
        // create lookup and sent to key
        AbstractLookup* lookup = createLookup(routeMsg);
        lookup->lookup( routeMsg->getDestKey(), numSiblings, hopCountMax,
                        new SendToKeyListener( this, routeMsg ) );
    } else  {
        NodeVector* nextHops = findNode(routeMsg->getDestKey(), 1,
					numSiblings, routeMsg);
	
        if (nextHops->size() == 0) {
            EV <<   "BaseOverlay::sendToKey(): findNode() returned NULL "
            "- dropping message" << endl;
            // statistics
            RECORD_STATS(numDropped++; bytesDropped += routeMsg->byteLength());
            delete routeMsg;
        } else {
            // delete message if the hop count maximum is exceeded
            if (routeMsg->getHopCount() >= hopCountMax) {
		EV << "BaseOverlay: Node " << thisNode.ip << " discards "
		   << routeMsg->name() << " from "
		   << routeMsg->getSrcNode().ip
		   << ". The hop count maximum has been exceeded. ("
		   << routeMsg->getHopCount() << ">="
		   << hopCountMax << ")" << endl;
		
                // statistics
                RECORD_STATS(numDropped++;
                             bytesDropped += routeMsg->byteLength());
                delete routeMsg;
		delete nextHops;
                return;
            }

	    // callForward to app
	    if(useCommonAPIforward && !iterativeLookup
	       && dynamic_cast<BaseAppDataMessage*>(routeMsg->encapsulatedMsg())
	       && routeMsg->kind() != ALREADY_FORWARDED) {
		//routeMsg->setKind(ALREADY_FORWARDED);
		callForward(routeMsg->getDestKey(), routeMsg, (*nextHops)[0]);
		delete nextHops;
		return;
	    }
	    routeMsg->setKind(NOT_FORWARDED);

            // forward msg if this node is not responsible for the key

            if ((*nextHops)[0] != thisNode) {
                OverlayCtrlInfo* overlayCtrlInfo =
                    dynamic_cast<OverlayCtrlInfo*>(msg->removeControlInfo());//why msg??? routeMsg???

                // records statistics, if we forward this message
                if ((overlayCtrlInfo != NULL) &&
                        (overlayCtrlInfo->getLastHopAddr() != thisNode.ip)) {
                    if (routeMsg->getSignaling() == false) {
                        RECORD_STATS(numForwarded++; bytesForwarded +=
                                         routeMsg->byteLength());
                    } else {
                        RECORD_STATS(numSignalingForwarded++;
                                     bytesSignalingForwarded +=
                                         routeMsg->byteLength());
                    }
                }
                if (overlayCtrlInfo != NULL)
                    delete overlayCtrlInfo;
            }

	    bool err;
	    if (((*nextHops)[0] == thisNode) &&
		!isSiblingFor(thisNode,routeMsg->getDestKey(),
				 numSiblings, &err)) {
		error("nextHop = thisNode but isSiblingsFor() is false!");
	    }

//	    EV << "!!!! Node " << thisNode << " forwards msg for key "
//	       << routeMsg->getDestKey() << " to node " << (*nextHops)[0]
//	       << endl;

	    forwardMessageRecursive((*nextHops)[0], routeMsg);
        }
        delete nextHops;
    }
}

//protected: create a iterative lookup class
AbstractLookup* BaseOverlay::createLookup(const BaseOverlayMessage* msg)
{
    AbstractLookup* newLookup = new BaseLookup( this, baseLookupConfig );
    lookups.insert( newLookup );
    return newLookup;
}

void BaseOverlay::removeLookup( AbstractLookup* lookup )
{
    lookups.erase(lookup);
}

//virtual public
OverlayKey BaseOverlay::distance( const OverlayKey& x,
                                  const OverlayKey& y ) const
{
    return x > y ? x-y : y-x;
}

//protected: find closest nodes
NodeVector* BaseOverlay::findNode(const OverlayKey& key,
				  int numRedundantNodes,
				  int numSiblings,
                                  BaseOverlayMessage* msg)
{
    opp_error( "findNode: Not implemented!" );
    return NULL;
}

//protected: join the overlay with a given nodeID
void BaseOverlay::joinOverlay()
{
    std::cout << "BaseOverlay::joinOverlay(): Not implemented!" << endl;
    //joinOverlay(): Not implemented
    return;
}

bool BaseOverlay::handleFailedNode(const TransportAddress& failed)
{
    return false;
}

PingResponse* BaseOverlay::ping( PingCall* call )
{
    std::string pongName("PONG: ");
    pongName += call->name();
    PingResponse* response = new PingResponse(pongName.c_str());

    response->setLength( PINGRESPONSE_L(response) );
    return response;
}


//------------------------------------------------------------------------
//--- RPCs ---------------------------------------------------------------
//------------------------------------------------------------------------

void BaseOverlay::sendRpcMessageToDestination(int destType,
					      const TransportAddress& dest,
					      const OverlayKey& destKey,
					      BaseOverlayMessage* message)
{
    switch(destType) {
      case RPC_TO_UDP:
          sendMessageToUDP( dest, message );
          break;
      case RPC_TO_KEY:
          sendToKey( destKey, message, 1, dest );
          break;
      case RPC_TO_UPPERTIER:
          send( message, "to_app");
          break;
      case RPC_TO_LOWERTIER:
          break;
    }
}  

//private
bool BaseOverlay::internalHandleRpc( BaseCallMessage* msg )
{
    // call rpc stubs
    RPC_SWITCH_START( msg );
    RPC_DELEGATE( Ping, pingRpc );
    RPC_DELEGATE( FindNode, findNodeRpc );
    RPC_DELEGATE( FailedNode, failedNodeRpc );
    RPC_DELEGATE( Lookup, lookupRpc );
    RPC_SWITCH_END( );

    // check if rpc has been handled
    IF_RPC_HANDLED return true;
    else
        return false;
}


//protected: statistic helpers for BaseLookup
void BaseOverlay::countFindNodeCall( const FindNodeCall* call )
{
    RECORD_STATS(numFindNodeSent++;
	    bytesFindNodeSent += call->byteLength());
}

void BaseOverlay::countFailedNodeCall( const FailedNodeCall* call )
{
    RECORD_STATS(numFailedNodeSent++;
	    bytesFailedNodeSent += call->byteLength());
}

//private: rpc stub
void BaseOverlay::findNodeRpc( FindNodeCall* call )
{
    // if this node is malicious don't answer a findNodeCall
    if (isMalicious()) {
	EV << "BaseOverlay: Node ignores findNodeCall "
	   << "because this node is malicious!" << endl;
	delete call;
	return;
    }

    FindNodeResponse* findNodeResponse =
        new FindNodeResponse("FindNodeResponse");

    NodeVector* nextHops = findNode(call->getLookupKey(),
				    call->getNumRedundantNodes(),
				    call->getNumSiblings(), call);

    findNodeResponse->setClosestNodesArraySize(nextHops->size());

    for (uint i=0; i < nextHops->size(); i++) {
        findNodeResponse->setClosestNodes(i, (*nextHops)[i]);
    }

    bool err;
    if (isSiblingFor(thisNode, call->getLookupKey(), call->getNumSiblings(),
			&err)) {
        findNodeResponse->setSiblings(true);
    }

    findNodeResponse->setLength(FINDNODERESPONSE_L(findNodeResponse));

    if (call->hasObject("findNodeExt")) {
	cMessage* findNodeExt = (cMessage*)call->removeObject("findNodeExt");
	findNodeResponse->addObject(findNodeExt);
	findNodeResponse->addLength(findNodeExt->length());
    }

    RECORD_STATS(numFindNodeResponseSent++; bytesFindNodeResponseSent +=
		 findNodeResponse->byteLength());

    delete nextHops;

    sendRpcResponse( call, findNodeResponse );
}

void BaseOverlay::failedNodeRpc( FailedNodeCall* call )
{
    FailedNodeResponse* failedNodeResponse =
	new FailedNodeResponse("FailedNodeResponse");
    failedNodeResponse->setTryAgain(handleFailedNode(call->getFailedNode()));
    failedNodeResponse->setLength(FAILEDNODERESPONSE_L(failedNodeResponse));
    if (call->hasObject("findNodeExt"))
    {
	cMessage* findNodeExt = check_and_cast<cMessage*>(
		call->removeObject("findNodeExt"));
	failedNodeResponse->addObject(findNodeExt);
	failedNodeResponse->addLength(findNodeExt->length());
    }

    RECORD_STATS(numFailedNodeResponseSent++; bytesFailedNodeResponseSent +=
	    failedNodeResponse->byteLength());

    sendRpcResponse( call, failedNodeResponse );
}

//private: rpc stub
void BaseOverlay::pingRpc( PingCall* call )
{
    PingResponse* response = ping(call);
    RECORD_STATS(numPingResponseSent++; bytesPingResponseSent +=
	    response->byteLength());
    sendRpcResponse( call, response );
}

void BaseOverlay::lookupRpc(LookupCall* call)
{
    // create lookup and sent to key
    AbstractLookup* lookup = createLookup(call);
    lookup->lookup( call->getKey(), call->getNumSiblings(), hopCountMax,
		    new SendToKeyListener( this, call ) );
}


