// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseOverlay.h
 * @author Bernhard Heep
 * @author Sebastian Mies
 */

#ifndef __BASEOVERLAY_H_
#define __BASEOVERLAY_H_

class BootstrapOracle;
class UnderlayConfigurator;

#include <omnetpp.h>

// global modules
#include <BootstrapOracle.h>
#include <GlobalStatistics.h>
#include <UnderlayConfigurator.h>
#include <InitStages.h>
#include <NotificationBoard.h>
#include <UDPAppBase.h>

// common structures
#include <NodeHandle.h>
#include <NodeVector.h>
#include <OverlayKey.h>
#include <NodeVector.h>
#include <BaseOverlayMessage_m.h>
#include <CommonAPIMessages_m.h>

// rpc includes
#include <BaseOverlayMessage_m.h>
#include <BaseRpc.h>
#include <ext/hash_map>

// base lookup
#include <BaseLookup.h>

using namespace __gnu_cxx;


/**
 * Base class for overlays
 *
 * Base class for overlay modules, with KBR-API, statistics and
 * pointers to the BootstrapOracle and the UnderlayConfigurator.
 * A minimal overlay has to implement numInitStages(), route(),
 * handleUDPMessage() and receiveChangeNotification(). Derived
 * classes must use BaseOverlayMessage as base class for own
 * message types.
 *
 * @author Bernhard Heep (initial)
 * @author Sebastian Mies (rpc, lookup)
 */
class BaseOverlay : public INotifiable,
            public BaseRpc
{

    friend class BaseLookup;
friend class BasePathLookup;

    //------------------------------------------------------------------------
    //--- Construction / Destruction -----------------------------------------
    //------------------------------------------------------------------------

public:

    /** Virtual destructor */
    virtual ~BaseOverlay();

    //------------------------------------------------------------------------
    //--- Statistics ---------------------------------------------------------
    //------------------------------------------------------------------------

private://fields: statistics

    int numSent;   //*< number of sent packets (incl.\ forwarded packets) */
    int bytesSent; //*< number of sent bytes (incl.\ forwarded bytes) */
    int numSignalingSent;   //*< number of sent signaling packets (incl.\ forwarded packets) */
    int bytesSignalingSent; //*< number of sent signaling bytes (incl.\ forwarded bytes) */

    int numReceived;   //*< number of received packets (incl.\ packets to be forwarded ) */
    int bytesReceived; //*< number of received bytes (incl.\ bytes to be forwarded) */
    int numSignalingReceived;   //*< number of received signaling packets (incl.\ packets to be forwarded) */
    int bytesSignalingReceived; //*< number of received signaling bytes (incl.\ bytes to be forwarded) */

protected:

    int numForwarded;   //*< number of forwarded packets */
    int bytesForwarded; //*< number of forwarded bytes at out-gate */
    int numSignalingForwarded;   //*< number of forwarded signaling packets */
    int bytesSignalingForwarded; //*< number of forwarded signaling bytes at out-gate */
    int numPingSent; /**< */
    int bytesPingSent; /**< */
    int numPingResponseSent; /**< */
    int bytesPingResponseSent; /**< */
    int numFindNodeSent; /**< */
    int bytesFindNodeSent; /**< */
    int numFindNodeResponseSent; /**< */
    int bytesFindNodeResponseSent; /**< */
    int numFailedNodeSent; /**< */
    int bytesFailedNodeSent; /**< */
    int numFailedNodeResponseSent; /**< */
    int bytesFailedNodeResponseSent; /**< */

	simtime_t creationTime; //*< simtime when the node has been created */


    //------------------------------------------------------------------------
    //--- Common Overlay Attributes & Global Module References ---------------
    //------------------------------------------------------------------------

protected://fields: overlay attributes

    // overlay identity
    cModule* thisTerminal; //*< pointer to corresponding node */

    // references to global modules
    BootstrapOracle* bootstrapOracle;           //*< pointer to BootstrapOracle in this node  */
    GlobalStatistics* globalStatistics;         //*< pointer to GlobalStatistics module in this node */
    NotificationBoard* notificationBoard;       //*< pointer to NotificationBoard in this node */
    UnderlayConfigurator* underlayConfigurator; //*< pointer to UnderlayConfigurator in this node */

    // overlay common parameters
    bool debugOutput;           //*< debug output ? */
    bool measureNetwInitPhase;  //*< fetch statistics in init phase ? */
    bool onlyCommonAPIMessages; //*< drop non-commonAPIMessages ? */
    bool useBaseLookup;		    //*< use the base lookup class */
    bool iterativeLookup;       //*< do iterative instead of recursive lookups */
    bool useCommonAPIforward;   //*< forward messages to applications? */
    int localPort;              //*< used UDP-port */
    int hopCountMax;            //*< maximum hop count */
    int numDropped;             //*< number of dropped packets */
    int bytesDropped;           //*< number of dropped bytes */
    cOutVector delayVector;     //*< statistical output vector for packet-delays */
    cOutVector hopCountVector;  //*< statistical output vector for hop-counts */

    //------------------------------------------------------------------------
    //--- Initialization & finishing -----------------------------------------
    //------------------------------------------------------------------------

private://methods: cSimpleModule initialization

    /**
     * initializes base-class-attributes
     *
     * @param stage the init stage
     */
    void initialize(int stage);

    /**
     * collects statistical data
     */
    void finish();

protected://methods: overlay initialization


    //  Default number of stages is 6.
    //      *
    //      * Stage 0-4 is needed to initialize OMNet and BaseOverlay.<br>
    //      * Stage 5-X can be used by the overlay implementation.<br>

    /**
     * Sets init stage.
     *
     * @see InitStages.h
     */
    int numInitStages() const;

    /**
     * Initializes derived-class-attributes.<br>
     *
     * Initializes derived-class-attributes, called by
     * BaseOverlay::initialize(). By default this method is called
     * once. If more stages are needed one can overload
     * numInitStages() and add more stages.
     * 
     * @param stage the init stage
     */
    virtual void initializeOverlay( int stage );

    /**
     * collects statistical data in derived class
     */
    virtual void finishOverlay();

    //------------------------------------------------------------------------
    //--- General Overlay Parameters (getter and setters) --------------------
    //------------------------------------------------------------------------

public://methods: getters and setters

    /**
     * Returns true, if node is malicious.
     * 
     * @return true, if node is malicious.
     */
    bool isMalicious();

    /**
     * Returns the NodeHandle of this node.
     * 
     * @return the NodeHandle of this node.
     */
    NodeHandle& getThisNode();

   //------------------------------------------------------------------------
   //--- UDP functions copied from the INET framework .----------------------
   //------------------------------------------------------------------------

  protected:
    /**
     * Tells UDP we want to get all packets arriving on the given port
     */
    virtual void bindToPort(int port);

    /**
     * Sends a packet over UDP
     */
    virtual void sendToUDP(cMessage *msg, int srcPort, const IPvXAddress& destAddr,
			   int destPort);

    /**
     * Prints a brief about packets having an attached UDPControlInfo
     * (i.e. those which just arrived from UDP, or about to be send to UDP).
     */
    virtual void printPacket(cMessage *msg);



    //------------------------------------------------------------------------
    //--- Overlay Common API: Key-based Routing ------------------------------
    //------------------------------------------------------------------------

protected://methods: KBR

    /**
     * Routes message through overlay.
     * 
     * The default implementation uses FindNode to determine next
     * hops and a generic greedy routing algorithm provides with
     * SendToKey.
     *
     * @param key destination key
     * @param msg message to route
     * @param hint next hop (usually unused)
     */
    virtual void route( const OverlayKey& key, cMessage* msg,
                        const NodeHandle& hint
                        = NodeHandle::UNSPECIFIED_NODE );


    /**
     * Calls deliver function in application.
     *
     * Encapsulates messages in KBRdeliver messages and sends them
     * to application.
     * 
     * @param msg delivered message
     */
    void callDeliver( BaseOverlayMessage* msg,
		      const OverlayKey& destKey);

    /**
     * Calls forward function in application
     *
     * Encapsulates messages in KBRforward messages and sends them
     * to application. <br> the message to be sent through the API
     * must be encapsulated in <code>msg</code>.
     * 
     * @param key destination key
     * @param msg message to forward
     * @param nextHopNode next hop
     */
    void callForward( const OverlayKey& key, BaseRouteMessage* msg,
                      const NodeHandle& nextHopNode);

    /**
     * Informs application about state changes of nodes or newly joined nodes
     * 
     * Creates a KBRUpdate message and sends it up to the application
     * 
     * @param node the node that has joined or changed its state
     * @param joined has the node joined or changed its state?
     */
    void callUpdate(const NodeHandle& node, bool joined);

public:

    /**
     * Join the overlay with a given nodeID
     * 
     * Join the overlay with a given nodeID. This method may be called
     * by an application to join the overlay with a specific nodeID. It is
     * also called if the node's IP address changes.
     * 
     * @param nodeID The new nodeID for this node.
     */
    void join(const OverlayKey& nodeID = OverlayKey::UNSPECIFIED_KEY);

    /**
     * finds nodes closest to the given OverlayKey
     * 
     * calls findNode() (that should be overridden in derived overlay)
     * and returns a list with (num) nodes ordered by distance to the
     * node defined by key.
     * 
     * @param key the given node
     * @param num number of nodes that are returned
     * @param safe 
     */
    virtual NodeVector* local_lookup(const OverlayKey& key, int num, bool safe);

    
    /**
     * @param num
     */
    virtual NodeVector* neighborSet(int num);
    
    /**
     * Query if a node is among the siblings for a given key.
     * 
     * Query if a node is among the siblings for a given key.
     * This means, that the nodeId of this node is among the closest
     * numSiblings nodes to the key and that by a local findNode() call
     * all other siblings to this key can be retrieved. 
     *
     * @param node the NodeHandle 
     * @param key destination key
     * @param numSiblings The nodes knows all numSiblings nodes close
     *                    to this key
     * @param err return false if the range could not be determined
     * @return bool true, if the node is responsible for the key.
     */
    virtual bool isSiblingFor(const NodeHandle& node,
			   const OverlayKey& key, int numSiblings, bool* err);

    /**
     * Query the maximum number of siblings (nodes close to a key)
     * that are maintained by this overlay protocol.
     *
     * @return int number of siblings.
     */
    virtual int getMaxNumSiblings();

    /**
     * Query the maximum number of redundant next hop nodes that
     * are returned by findNode().
     *
     * @return int number of redundant nodes returned by findNode().
     */
    virtual int getMaxNumRedundantNodes();

    //------------------------------------------------------------------------
    //--- Message Handlers ---------------------------------------------------
    //------------------------------------------------------------------------

private://methods: message handling

    /**
     * Checks for message type and calls corresponding method.<br>
     *
     * Checks for message type (from UDP/App or selfmessage) and
     * calls corresponding method like route(), get(), put(),
     * remove(), handleTimerEvent(), handleAppMessage() and
     * handleUDPMessage().
     * 
     * @param msg The message to be handled
     */
    void handleMessage(cMessage* msg);


    /**
     * Handles a BaseOverlayMessage<br>
     *
     * Handles BaseOverlayMessages of type OVERLAYSIGNALING, RPC,
     * APPDATA or OVERLAYROUTE.
     * 
     * @param msg The message to be handled
     */
    void handleBaseOverlayMessage(BaseOverlayMessage* msg,
				  const OverlayKey& destKey = OverlayKey::UNSPECIFIED_KEY);

protected://methods: message handling

    /**
     * Processes messages from underlay
     *
     * @param msg Message from UDP
     */
    virtual void handleUDPMessage(BaseOverlayMessage* msg) = 0;

    /**
     * Processes "timer" self-messages
     *
     * @param msg A self-message
     */
    virtual void handleTimerEvent(cMessage* msg);

    /**
     * Processes non-commonAPI messages.<br>
     *
     * Processes non-commonAPI messages if parameter
     * "onlyCommonAPIMessages" is false.
     * 
     * @param msg non-commonAPIMessage
     */
    virtual void handleAppMessage(cMessage* msg);

    /**
     * Collect overlay specific sent messages statistics
     *
     * This method is called from BaseOverlay::sendMessageToUDP()
     * for every overlay message that is sent by a node. Use this
     * to collect statistical data for overlay protocol specific
     * message types.
     *
     * @param msg The overlay message to be sent to the UDP layer
     */
    virtual void recordOverlaySentStats(BaseOverlayMessage* msg);


    //------------------------------------------------------------------------
    //--- Overlay Topology Drawing -------------------------------------------
    //------------------------------------------------------------------------

private://fields: draw overlay topology

    cGate* thisOutGateArray;  //*< out-gate array for vis. overlay topology */
    cGate* thisInGateArray;   //*< in-gate array for vis. overlay topology */
    bool drawOverlayTopology; //*< draw overlay topology? */

    enum Direction { IN, OUT };

    /**
     * compacts arrow gate-array
     *
     * @param terminal node
     * @param dir in- or out-array?
     */
    void compactGateArray(cModule* terminal, enum Direction dir);

protected://methods: icons and ui support

    /**
     * Colors module-icon blue (ready) or red (not ready)
     *
     * @param ready state to visualize
     */
    void setReadyIcon( bool ready );

    /**
     * Draws an arrow from this node to neighbor
     *
     * @param neighbor neighbor to point to
     * @param flush delete all previous drawn arrows starting at this node?
     * @param displayString display string to define the arrow drawing style
     * @todo add neighbor's kind to distinguish arrows pointing to the
     * same neighbor
     */
    void showOverlayNeighborArrow( const NodeHandle& neighbor,
                                   bool flush = true,
                                   char* displayString = NULL );

    /**
     * Removes an arrow from this node to neighbor
     *
     * @param neighbor neighbor to remove arrow to
     * @todo add neighbor's kind to distinguish arrows pointing to the
     * same neighbor
     */
    void deleteOverlayNeighborArrow( const NodeHandle& neighbor );

    //------------------------------------------------------------------------
    //--- Messages -----------------------------------------------------------
    //------------------------------------------------------------------------

public://methods: sending packets over udp

    /**
     * Sends message to underlay
     * 
     * @param dest destination node
     * @param msg message to send
     */
    void sendMessageToUDP( const TransportAddress& dest, BaseOverlayMessage* msg );


    /**
     * ping a node by its TransportAddress
     * 
     * Statistics are collected by this method.
     *
     * @param dest the node to ping
     * @param timeout RPC timeout
     * @param retries how often to retry after timeout
     * @param caption special name for the ping call (instead of "PING")
     * @param rpcListener RPC Listener
     * @param rpcId RPC id
     */
    void pingNode( const TransportAddress& dest,
	           simtime_t timeout = -1,
		   int retries = 0,
		   const char* caption = NULL,
		   RpcListener* rpcListener = NULL,
		   int rpcId = -1 );

    /**
     * ping a node by its OverlayKey
     *
     * Statistics are collected by this method.
     *
     * @param destKey key of the node to ping
     * @param timeout RPC timeout
     * @param retries how often to retry after timeout
     * @param caption special name for the ping call (instead of "PING")
     * @param rpcListener RPC Listener
     * @param rpcId RPC id
     */
    void pingKey( const OverlayKey& destKey,
		  simtime_t timeout = -1,
		  int retries = 0,
		  const char* caption = NULL,
		  RpcListener* rpcListener = NULL,
		  int rpcId = -1 );

    //------------------------------------------------------------------------
    //--- Basic Routing ------------------------------------------------------
    //------------------------------------------------------------------------
protected://fields: config
    BaseLookupConfiguration baseLookupConfig;

    class lookupHashFcn
    {
    public:
        size_t operator()( const AbstractLookup* l1 ) const
        {
            return (size_t)l1;
        }
        bool operator()(const AbstractLookup* l1,
                        const AbstractLookup* l2) const
        {
            return (l1 == l2);
        }
    };

    typedef hash_set<AbstractLookup*, lookupHashFcn, lookupHashFcn> LookupSet;

    LookupSet lookups;

private://methods: internal routing

    /**
     * creates a LookupSet
     */
    void initLookups();
    
    /**
     * deletes entries in lookups
     */
    void finishLookups();

    /**
     * calls route(const OverlayKey& key, cMessage* msg, const NodeHandle& hint) 
     * if useBaseLookup is false,
     * else sendToKey(const OverlayKey& key, BaseOverlayMessage* msg,
     *                uint numNeighbors, const TransportAddress& nextHop)
     */
    void internalRoute( const OverlayKey& key, cMessage* msg,
                        const NodeHandle& hint
                        = NodeHandle::UNSPECIFIED_NODE );

    /**
     * Hook for forwarded message in recursive lookup mode
     *
     * Default implementation just calls sendMessageToUDP(). This hook
     * can for example be used to detect failed nodes and call
     * handleFailedNode() before the actual forwarding takes place.
     *
     * @param dest destination node
     * @param msg message to send
     */
    virtual void forwardMessageRecursive( const TransportAddress& dest,
	    BaseRouteMessage* msg);

public://methods: basic message routing

    /**
     * Sends a message to an overlay node, with the generic routing
     * algorithm.
     * 
     * @param key The destination key
     * @param message Message to be sent
     * @param numSiblings number of siblings to send message to
     *                     (numSiblings > 1 means multicast)
     * @param nextHop If nextHop is given, the message gets sent to this
     *                node before it is routed (nextHop is used as a proxy)
     */
    void sendToKey( const OverlayKey& key, BaseOverlayMessage* message,
                    int numSiblings = 1,
                    const TransportAddress& nextHop = TransportAddress::UNSPECIFIED_NODE);

protected://methods: routing class factory

    /**
     * Creates an abstract iterative lookup instance.
     * 
     * @param msg pointer to the message for which the lookup is created.
     *            Derived classes can use it to construct an object with
     *            additional info for the lookup class.
     * @return AbstractLookup* The new lookup instance.
     */
    virtual AbstractLookup* createLookup(const BaseOverlayMessage* msg = NULL);

    /**
     * Removes the abstract lookup instance.
     * 
     * @param lookup
     */
    virtual void removeLookup( AbstractLookup* lookup );

    /**
     * This method should implement the distance between two keys.
     * It may be overloaded to implement a new metric. The default
     * implementation uses the standard-metric d = abs(x-y).
     * 
     * @param x Left-hand-side Key
     * @param y Right-hand-side key
     * @return OverlayKey Distance between x and y
     */
    virtual OverlayKey distance( const OverlayKey& x,
                                 const OverlayKey& y ) const;

    /**
     * Implements the find node call.
     * 
     * This method simply returns the closest nodes known in the
     * corresponding routing topology. If the node is a sibling for
     * this key (isSiblingFor(key) = true), this method returns all
     * numSiblings siblings, with the closest neighbor to the key
     * first.
     * 
     * @param key The lookup key.
     * @param numRedundantNodes Maximum number of next hop nodes to return.
     * @param numSiblings number of siblings to return
     * @param msg A pointer to the BaseRouteMessage or FindNodeCall
     *                   message of this lookup.
     * @return NodeVector with closest nodes.
     */
    virtual NodeVector* findNode( const OverlayKey& key,
				  int numRedundantNodes,
				  int numSiblings,
                                  BaseOverlayMessage* msg = NULL);


    /**
     * Join the overlay with a given nodeID in thisNode.key
     * 
     * Join the overlay with a given nodeID in thisNode.key.  This
     * method may be called by an application to join the overlay with
     * a specific nodeID. It is also called if the node's IP address
     * changes.
     * 
     */
    virtual void joinOverlay();

    /**
     * Handles a failed node.
     *
     * This method is called whenever a node given by findNode() was
     * unreachable. The default implementation does nothing at all.
     *
     * @param failed the failed node
     * @return true if lookup should retry here
     */
    virtual bool handleFailedNode( const TransportAddress& failed);

    /**
     * Implements a ping call.<br>
     * 
     * This method implements a simple ping call.
     * 
     * @param call The ping message
     * @return The ping response or NULL if no response should be
     *         sent.
     */
    virtual PingResponse* ping( PingCall* call );

    virtual void lookupRpc(LookupCall* call);


protected://methods: statistic helpers for BaseLookup

    void countFindNodeCall( const FindNodeCall* call );
    void countFailedNodeCall( const FailedNodeCall* call );

protected:
    void sendRpcMessageToDestination(int destType,
				     const TransportAddress& dest,
				     const OverlayKey& destKey,
				     BaseOverlayMessage* message);
    bool internalHandleRpc( BaseCallMessage* msg );

private://rpc stubs
    void findNodeRpc( FindNodeCall* call );
    void failedNodeRpc( FailedNodeCall* call );
    void pingRpc( PingCall* call );
    PingCall* createPingCall( const char* caption );

};


#endif
