//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseRpc.cc
 * @author Bernhard Heep (initial)
 * @author Sebastian Mies (rpcs)
 * @author Gregoire Menuel (separation from BaseRpc)
 */

#include <CommonAPIMessages_m.h>
#include <BootstrapOracleAccess.h>
#include <UnderlayConfiguratorAccess.h>
#include <GlobalStatisticsAccess.h>


#include "BaseRpc.h"
#include "RpcMacros.h"

enum {
    NOT_FORWARDED,
    ALREADY_FORWARDED
};



//------------------------------------------------------------------------
//--- Initialization & finishing -----------------------------------------
//------------------------------------------------------------------------


//------------------------------------------------------------------------
//--- RPC Handling -------------------------------------------------------
//------------------------------------------------------------------------

//private
void BaseRpc::initRpcs()
{
    rpcsPending = 0;
    rpcStates.clear();

    defaultRpcListener = new RpcListener();
}

//private
void BaseRpc::finishRpcs()
{
    // stop all rpcs
    for (RpcStates::iterator i = rpcStates.begin();
        i != rpcStates.end(); i++) {
        cancelAndDelete(i->second.callMsg);
        cancelAndDelete(i->second.timeoutMsg);
    }
    rpcStates.clear();

    // delete default rpc listener
    if (defaultRpcListener!=NULL) {
        delete defaultRpcListener;
        defaultRpcListener = NULL;
    }
}

//public
uint32_t BaseRpc::sendRpcMessage( const TransportAddress& dest,
                                  BaseCallMessage* msg,
                                  RpcListener* rpcListener,
                                  const OverlayKey& destKey,
                                  int rpcId, simtime_t timeout,
                                  int retries)
{
    if (destKey.isUnspecified()) {
        return sendRpcMessage( RPC_TO_UDP, dest, msg, rpcListener, destKey, rpcId, timeout, retries );
    } else {
        // send message to key
        return sendRpcMessage( RPC_TO_KEY, dest, msg, rpcListener, destKey, rpcId, timeout, retries );
    }
}  


uint32_t BaseRpc::sendRpcMessage( int destType, const TransportAddress& dest,
                                  BaseCallMessage* msg,
                                  RpcListener* rpcListener,
                                  const OverlayKey& destKey,
                                  int rpcId, simtime_t timeout,
                                  int retries)
{

    // create nonce, timeout and set default parameters
    uint nonce;
    do {
        nonce = intuniform( 0, 2147483647);
    } while (rpcStates.count(nonce) > 0);


    if (timeout == -1)
        timeout = destKey.isUnspecified() ? 1.0 : 5.0;
    if (rpcListener == NULL)
        rpcListener = defaultRpcListener;

    // create state
    RpcState state;
    state.id = rpcId;
    state.timeSent = simulation.simTime();
    state.dest = dest;
    state.destKey = destKey;
    state.listener = rpcListener;
    state.timeoutMsg = new RpcTimeoutMessage();
    state.timeoutMsg->setNonce(nonce);
    state.retries = retries;
    state.destType = destType;



    if (rpcStates.count(nonce) > 0)
        error("RPC nonce collision");

    // set message parameters
    msg->setNonce( nonce );
    msg->setSrcNode( thisNode );
    msg->setType(RPC);

    // save copy of call message in RpcState
    state.callMsg = dynamic_cast<BaseCallMessage*>(msg->dup());

    // register state
    rpcStates[nonce] = state;

    // schedule timeout message
    if (timeout!=0)
        scheduleAt( simulation.simTime() + timeout, state.timeoutMsg);

    sendRpcMessageToDestination(destType, dest, destKey, msg);

    return nonce;
}

//public
void BaseRpc::cancelRpcMessage( uint32_t nonce )
{
    if (rpcStates.count(nonce)==0)
        return;
    RpcState state = rpcStates[nonce];
    rpcStates.erase(nonce);
    cancelAndDelete(state.callMsg);
    cancelAndDelete(state.timeoutMsg);
}

//protected
void BaseRpc::internalHandleRpcMessage( BaseRpcMessage* msg)
{
    // check if this is a rpc call message
    BaseCallMessage* rpCall = dynamic_cast<BaseCallMessage*>(msg);
    if (rpCall != NULL) {
        bool rpcHandled = true;
        if (!handleRpc(rpCall)) rpcHandled = internalHandleRpc(rpCall);
        if (!rpcHandled) {
            EV << "Error: RPC '" << msg->fullName()
	       << "' was not handled" << endl;
	    delete msg;
	}
        return;
    }

    // get nonce
    int nonce = msg->getNonce();

    // nonce known? no -> delete message and return
    if (rpcStates.count(nonce)==0) {
        EV << "RPC: Nonce Unknown!" << endl;
        delete msg;
        return;
    }

    // get state and remove from map
    RpcState state = rpcStates[nonce];
    rpcStates.erase(nonce);

    // is timeout message?
    if ( msg->isSelfMessage() && (dynamic_cast<RpcTimeoutMessage*>(msg) != NULL)) { // yes-> inform listener

        // retry?
        state.retries--;
        simtime_t timeout = simulation.simTime() - state.timeSent;
        if (state.retries>=0) {
            sendRpcMessageToDestination(state.destType, state.dest,
                                        state.destKey,
                                        dynamic_cast<BaseCallMessage*>
                                        (state.callMsg->dup()));
            if (timeout!=0)
                scheduleAt( simulation.simTime() + timeout, msg );
            rpcStates[nonce] = state;
            return;
        }

        // inform listener
        if ( state.listener != NULL )
            state.listener->handleRpcTimeout( state.callMsg, state.dest,
                                              state.id, state.destKey );

        // inform overlay
        handleRpcTimeout( state.callMsg, state.dest, state.id, state.destKey );

    } else { // no-> handle rpc response

        // get parameters
        simtime_t rtt = simulation.simTime() - state.timeSent;
        BaseResponseMessage* response
        = dynamic_cast<BaseResponseMessage*>(msg);

        // inform listener
        if ( state.listener != NULL )
            state.listener->handleRpcResponse( response, state.id, rtt );

        // inform overlay
        handleRpcResponse( response, state.id, rtt );

        // delete response
        delete response;
    }

    // delete messages
    cancelAndDelete(state.callMsg);
    cancelAndDelete(state.timeoutMsg);

    // clean up pointers
    state.callMsg = NULL;
    state.timeoutMsg = NULL;
}

//private
bool BaseRpc::internalHandleRpc( BaseCallMessage* msg )
{
    return false;
}

//virtual protected
bool BaseRpc::handleRpc( BaseCallMessage* msg )
{
    return false;
}

void BaseRpc::sendRpcResponse( int destType, BaseCallMessage* call,
                               BaseResponseMessage* response )
{
    if (call==NULL || response==NULL) {
        if ( call!=NULL )
            delete call;
        return;
    }
    response->setSrcNode( thisNode );
    response->setType( RPC );
    response->setNonce( call->getNonce() );
    sendRpcMessageToDestination(destType, call->getSrcNode(), call->getSrcNode().getKey(), response );
    delete call;

}


//protected
void BaseRpc::sendRpcResponse( BaseCallMessage* call,
                               BaseResponseMessage* response )
{
    sendRpcResponse(RPC_TO_UDP, call, response);
}

//virtual private
void BaseRpc::sendRpcMessageToDestination(int destType,
					  const TransportAddress& dest,
					  const OverlayKey& destKey,
					  BaseOverlayMessage* message)
{
    switch(destType) {
      case RPC_TO_UDP:
          opp_error( "sendRpcMessageToDestination with UDP: Not implemented!" );
          delete message;
          break;
      case RPC_TO_KEY:
          opp_error( "sendRpcMessageToDestination with KEY: Not implemented!" );
          delete message;
          break;
      case RPC_TO_UPPERTIER:
          send( message, "to_upperTier");
          break;
      case RPC_TO_LOWERTIER:
          send( message, "to_lowerTier");
          break;
    }
}

