// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseRpc.h
 * @author Bernhard Heep
 * @author Sebastian Mies
 * @author Gregoire Menuel
 */

#ifndef __BASERPC_H_
#define __BASERPC_H_


class BootstrapOracle;
class UnderlayConfigurator;

#include <omnetpp.h>

// rpc includes
#include <RpcListener.h>
#include <RpcMacros.h>

using namespace __gnu_cxx;

enum {
    RPC_TO_UDP,
    RPC_TO_KEY,
    RPC_TO_UPPERTIER,
    RPC_TO_LOWERTIER,
};

/**
 * Base class for RPCs
 *
 * Base class for RPCs.
 * 
 * @author Bernhard Heep (initial)
 * @author Sebastian Mies (rpc, lookup)
 * @author Gregoire Menuel (separation from BaseRpc)
 */
class BaseRpc : public RpcListener,
                public cSimpleModule
{

    //------------------------------------------------------------------------
    //--- Construction / Destruction -----------------------------------------
    //------------------------------------------------------------------------

public:

    /** Virtual destructor */
    //------------------------------------------------------------------------
    //--- Statistics ---------------------------------------------------------
    //------------------------------------------------------------------------

private://fields: statistics


protected://fields: overlay attributes

    // overlay identity
    NodeHandle thisNode;   //*< NodeHandle to this node */

    // overlay common parameters
    bool debugOutput;           //*< debug output ? */
    //------------------------------------------------------------------------
    //--- RPC Handling -------------------------------------------------------
    //------------------------------------------------------------------------

private://classes: rpc handling

    class RpcState
    {
    public:
        int id;
        int retries;
        int destType;
        RpcListener* listener;
        TransportAddress dest;
        OverlayKey destKey;
        BaseCallMessage* callMsg;
        RpcTimeoutMessage* timeoutMsg;
        simtime_t timeSent;
    };

private://fields: rpc handling

    typedef hash_map<int,RpcState> RpcStates;

    int rpcsPending;
    RpcListener* defaultRpcListener;
    RpcStates rpcStates;

protected://methods: rpc handling

    /**
     * Handles internal rpc requests.<br>
     * 
     * This method is used to implement basic functionionality in
     * the BaseRpc.
     * 
     * @param msg The call message
     * @return bool true, if call has been handled.
     */
    virtual bool internalHandleRpc( BaseCallMessage* msg );

    /**
     * Initializes Remote-Procedure state.
     */
    void initRpcs( );

    /**
     * Deinitializes Remote-Procedure state.
     */
    void finishRpcs( );

protected://methods: rpc handling

    /**
     * Handles incoming rpc messages and delegates them to the
     * corresponsing listeners or handlers.
     * 
     * @param msg The message to handle.
     */
    virtual void internalHandleRpcMessage( BaseRpcMessage* msg);

public:
    /**
     * Sends a Remote-Procedure-Call message to the underlay.<br>
     *
     * Compatibility function.
     * See the other sendRpcMessage function for more information
     * 
     * @param dest Destination node handle
     * @param msg RPC Call Message
     * @param rpcListener RPC Listener
     * @param destKey first route the RPC to the node that is
     *        responsible for kestkey
     * @param rpcId RPC id
     * @param timeout RPC timeout
     * @param retries How often we try to resent rpc call, if it gets lost
     * @return The nonce of the RPC
     */
    uint32_t sendRpcMessage( const TransportAddress& dest, BaseCallMessage* msg,
                             RpcListener* rpcListener = NULL,
                             const OverlayKey& destKey
                             = OverlayKey::UNSPECIFIED_KEY,
                             int rpcId = -1, simtime_t timeout = -1,
                             int retries = 0);


    /**
     * Sends a Remote-Procedure-Call message to the underlay.<br>
     * 
     * If no timeout is provided, a default value of 1.0 for
     * underlay and 5.0 for a overlay rpc is used. After a timeout
     * the message gets retransmitted for at maximum retries times.
     * 
     * The destKey
     * attribute is kept untouched. 
     * 
     * @param destType The type of destination (RPC_TO_UDP, RPC_TO_KEY, RPC_TO_LOWERTIER, RPC_TO_UPPERTIER)
     * @param dest Destination node handle
     * @param msg RPC Call Message
     * @param rpcListener RPC Listener
     * @param destKey first route the RPC to the node that is
     *        responsible for kestkey
     * @param rpcId RPC id
     * @param timeout RPC timeout
     * @param retries How often we try to resent rpc call, if it gets lost
     * @return The nonce of the RPC
     */
    uint32_t sendRpcMessage( int destType, const TransportAddress& dest, BaseCallMessage* msg,
                             RpcListener* rpcListener = NULL,
                             const OverlayKey& destKey
                             = OverlayKey::UNSPECIFIED_KEY,
                             int rpcId = -1, simtime_t timeout = -1,
                             int retries = 0 );

    /**
     * Cancels a Remote-Procedure-Call.
     * 
     * @param nonce The nonce of the RPC
     */
    void cancelRpcMessage( uint32_t nonce );

    /**
     * Send Remote-Procedure response message and deletes call
     * message.<br>
     * 
     * The destKey attribute is kept untouched. If call or reponse
     * is NULL no reponse will be sent.
     * 
     * @param destType the destination type of the RPC
     * @param call The corresponding call message to the reponse
     * @param response The call return value
     */
    void sendRpcResponse( int destType, BaseCallMessage* call,
		      BaseResponseMessage* response );

    void sendRpcResponse( BaseCallMessage* call,
                          BaseResponseMessage* response );


protected:
    /**
     * Processes Remote-Procedure-Call invokation messages.<br>
     * 
     * This method should be overloaded when the overlay provides
     * RPC functionality.
     * 
     * @return true, if rpc has been handled
     */
    virtual bool handleRpc( BaseCallMessage* msg );

protected://methods: rpc handling

    virtual void sendRpcMessageToDestination(int destType,
					     const TransportAddress& dest,
					     const OverlayKey& destKey,
					     BaseOverlayMessage* message);
};

#endif
