//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BootstrapOracle.cc
 * @author Markus Mauch, Robert Palmer
 */

#include <omnetpp.h>

#include <OverlayKey.h>

#include "BootstrapOracle.h"


Define_Module(BootstrapOracle);

std::ostream& operator<<(std::ostream& os, const bootstrapEntry entry)
{
    NodeHandle* nodeHandle = dynamic_cast<NodeHandle*>(entry.node);
   
    os << "Address: " << entry.node->ip
       << " Port: " << entry.node->port;

    if (nodeHandle) {
	os << " NodeId: " << nodeHandle->key;
    }

    os << " ModuleID: "
       << entry.info->getModuleID() << " Bootstrapped: "
       << (entry.info->isBootstrapped() ? "true" : "false");

    return os;
}

std::ostream& operator<<(std::ostream& stream, const DHTEntry entry)
{
    return stream << "Value: " << *(entry.value)
                  << " Endtime: " << entry.endtime; 
}

void BootstrapOracle::initialize()
{
    maxNumberOfKeys = par("maxNumberOfKeys");
    keyProbability = par("keyProbability");
    WATCH_HASH_MAP(peerSet);
    WATCH_MAP(dataMap);
    WATCH_VECTOR(keyList);
    WATCH(bootstrappedPeerSize);
    WATCH(bootstrappedMaliciousNodes);
    WATCH(maliciousNodes);
    createKeyList(maxNumberOfKeys);
    bootstrappedPeerSize = 0;
    bootstrappedMaliciousNodes = 0;
    maliciousNodes = 0;

	if (par("maliciousNodeChange"))
	{
		if ((double) par("maliciousNodeProbability") > 0)
			error("maliciousNodeProbability and maliciousNodeChange are not supported concurrently");
		
		cMessage* msg = new cMessage("maliciousNodeChange");
		scheduleAt(simulation.simTime() + (int) par("maliciousNodeChangeStartTime"), msg);
		maliciousNodesVector.setName("MaliciousNodeRate");
		maliciousNodesVector.record(0);
		maliciousNodeRatio = 0;
	}

    min_ip =0xFFFFFFFF;
    max_ip =0x00000000;

    globalStatistics = GlobalStatisticsAccess().get();

    cMessage* timer = new cMessage("oracleTimer");

    scheduleAt(simulation.simTime(), timer);
}

void BootstrapOracle::handleMessage(cMessage* msg)
{
	if (msg->isName("maliciousNodeChange"))
	{
		double newRatio = maliciousNodeRatio + (double) par("maliciousNodeChangeRate"); // ratio to obtain
		if (maliciousNodeRatio < (double) par("maliciousNodeChangeStartValue"))
			newRatio = (double) par("maliciousNodeChangeStartValue");
		
		if (newRatio < (double) par("maliciousNodeChangeStopValue")) // schedule next event
			scheduleAt(simulation.simTime() + (int) par("maliciousNodeChangeInterval"), msg);

		int nodesNeeded = (int) (((double) par("maliciousNodeChangeRate")) * peerSet.size());

		EV << "BootstrapOracle: Changing " << nodesNeeded << " Nodes to be malicious\n";
		
		for (int i = 0; i < nodesNeeded; i++)
		{
			// search a node that is not yet malicious
			NodeHandle node;
			do node = getRandomNode(false);
			while (isMalicious(node));

			setMalicious(node, true);
		}

		maliciousNodesVector.record(newRatio);
		maliciousNodeRatio = newRatio;
		
		return;
	}

    else if (msg->isName("oracleTimer")) {
	RECORD_STATS(globalStatistics->recordOutVector(
			 "BootstrapOracle: Number of nodes", peerSet.size()));
	scheduleAt(simulation.simTime() + 10, msg);
    } else {
	opp_error("BootstrapOracle::handleMessage: Unknown message type!");
    }
}

const NodeHandle& BootstrapOracle::getBootstrapNode()
{
    return getRandomNode(true);
}

const NodeHandle& BootstrapOracle::getRandomNode(bool bootstrappedNeeded = false)
{
	if (peerSet.size() == 0)
		return NodeHandle::UNSPECIFIED_NODE;
	if (bootstrappedNeeded && bootstrappedPeerSize == 0)
		return NodeHandle::UNSPECIFIED_NODE;
	else {
        // return random TransportAddress in O(log n)
		PeerHashMap::iterator it = peerSet.end();
		bootstrapEntry tempEntry = {NULL, NULL};

		while(it == peerSet.end() || (bootstrappedNeeded && !it->second.info->isBootstrapped())) {
            IPvXAddress randomAddr(intuniform(min_ip, max_ip));

			it = peerSet.find(randomAddr);

			if (it == peerSet.end()) {
				it = peerSet.insert(std::make_pair(randomAddr,tempEntry)).first;
				peerSet.erase(it++);
			}

			if (it == peerSet.end())
				it = peerSet.begin();
		}

		if(dynamic_cast<NodeHandle*>(it->second.node))
			return *dynamic_cast<NodeHandle*>(it->second.node);
		else
			return NodeHandle::UNSPECIFIED_NODE;
	}
}

void BootstrapOracle::addPeer(const IPvXAddress& ip, PeerInfo* info)
{
    bootstrapEntry temp;
    temp.node = new TransportAddress(ip);
    temp.info = info;
    peerSet.insert(std::make_pair(temp.node->ip, temp));
    // set bounds for random node retrival
    if(ip.get4().getInt() < min_ip) min_ip = ip.get4().getInt();
    if(ip.get4().getInt() > max_ip) max_ip = ip.get4().getInt();

    if (uniform(0, 1) < (double) par("maliciousNodeProbability") ||
       (par("maliciousNodeChange") && uniform(0, 1) < maliciousNodeRatio))
	setMalicious(*temp.node, true);
}

void BootstrapOracle::registerPeer(const TransportAddress& peer)
{
    PeerHashMap::iterator it = peerSet.find(peer.ip);
    if (it == peerSet.end())
        error("unable to bootstrap peer, peer is not in peer set");
    else {
        PeerInfo* info = it->second.info;

        if (!info->isBootstrapped()) {
            bootstrappedPeerSize++;
            info->setBootstrapped();
            if (info->isMalicious())
                bootstrappedMaliciousNodes++;
        }

        delete it->second.node;
        peerSet.erase(it);

        bootstrapEntry temp;
        temp.node = new TransportAddress(peer);
        temp.info = info;
        peerSet.insert(std::make_pair(temp.node->ip, temp));
    }
}

void BootstrapOracle::registerPeer(const NodeHandle& peer)
{
    PeerHashMap::iterator it = peerSet.find(peer.ip);
    if (it == peerSet.end())
        error("unable to bootstrap peer, peer is not in peer set");
    else {
        PeerInfo* info = it->second.info;

        if (!info->isBootstrapped()) {
            bootstrappedPeerSize++;
            info->setBootstrapped();
            if (info->isMalicious())
		bootstrappedMaliciousNodes++;
        }

        delete it->second.node;
        peerSet.erase(it);

        bootstrapEntry temp;
        temp.node = new NodeHandle(peer);
        temp.info = info;
        peerSet.insert(std::make_pair(temp.node->ip, temp));
    }
}

void BootstrapOracle::removePeer(const TransportAddress& peer)
{
    PeerHashMap::iterator it = peerSet.find(peer.ip);
	if(it != peerSet.end() && it->second.info->isBootstrapped()) {
		bootstrappedPeerSize--;
		it->second.info->setBootstrapped(false);

        if(it->second.info->isMalicious())
            bootstrappedMaliciousNodes--;
        it->second.info->setBootstrapped(false);
    }
}

void BootstrapOracle::killPeer(const IPvXAddress& ip)
{
    PeerHashMap::iterator it = peerSet.find(ip);
    if(it != peerSet.end()) {
        if(it->second.info->isBootstrapped()) {
            bootstrappedPeerSize--;
            if(it->second.info->isMalicious())
                bootstrappedMaliciousNodes--;
            it->second.info->setBootstrapped(false);
        }

        delete it->second.node;
        delete it->second.info;
        peerSet.erase(it);
    }
}

void BootstrapOracle::createKeyList(uint size)
{
    for(uint i = 0; i < size; i++)
        keyList.push_back(OverlayKey::random());
}

BootstrapOracle::DataMap* BootstrapOracle::getDataMap()
{
    return &dataMap;
}

BootstrapOracle::KeyList* BootstrapOracle::getKeyList(uint maximumKeys)
{
    if (maximumKeys > keyList.size()) {
        maximumKeys = keyList.size();
    }
    // copy keylist to temporary keylist
    KeyList tmpKeyList;
    tmpKeyList.clear();
    for ( uint i=0; i < keyList.size(); i++ )
        tmpKeyList.push_back(keyList[i]);

    KeyList* returnList = new KeyList;

    for ( uint i=0; i < ((float)maximumKeys * keyProbability); i++) {
        uint index = intuniform(0, tmpKeyList.size()-1);
        returnList->push_back(tmpKeyList[index]);
        tmpKeyList.erase(tmpKeyList.begin()+index);
    }

    return returnList;
}

const OverlayKey& BootstrapOracle::getRandomKeyListItem() const
{
    return keyList[intuniform(0,keyList.size()-1)];
}

PeerInfo* BootstrapOracle::getPeerInfo(const TransportAddress& peer)
{
    PeerHashMap::iterator it = peerSet.find(peer.ip);
    if(it == peerSet.end())
        return NULL;
    else
        return it->second.info;
}

PeerInfo* BootstrapOracle::getPeerInfo(const IPvXAddress& ip)
{
    PeerHashMap::iterator it = peerSet.find(ip);
    if(it == peerSet.end())
        return NULL;
    else
        return it->second.info;
}

PeerInfo* BootstrapOracle::getRandomPeerInfo(uint32_t nodeType) {
    // return random TransportAddress in O(log n)
    PeerHashMap::iterator it;
    bootstrapEntry tempEntry = {NULL, NULL};

    IPvXAddress randomAddr(intuniform(min_ip, max_ip));

    it = peerSet.find(randomAddr);
    if (it == peerSet.end()) {
	it = peerSet.insert(std::make_pair(randomAddr,tempEntry)).first;
	peerSet.erase(it++);
    }
    if (it == peerSet.end())
	it = peerSet.begin();

    // if nodeType != 0, search for next node with the given type
    if (nodeType) {
	while( it->second.info->getTypeID() != nodeType ) {
	    ++it;
	    if (it == peerSet.end()) it = peerSet.begin();
	}
    }

    return it->second.info;
}

void BootstrapOracle::setMalicious(const TransportAddress& address, bool malicious)
{
    PeerInfo* peer = getPeerInfo(address);
    if(peer != NULL)
    {
        if(malicious && !peer->isMalicious())
		{
			maliciousNodes++;
			if (peer->isBootstrapped())
				bootstrappedMaliciousNodes++;
		}
		if (!malicious && peer->isMalicious())
		{
			maliciousNodes--;
			if (peer->isBootstrapped())
				bootstrappedMaliciousNodes--;
		}
        peer->setMalicious(malicious);
    }
}

bool BootstrapOracle::isMalicious(const TransportAddress& address)
{
    PeerInfo* peer = getPeerInfo(address);
    if(peer != NULL)
        return peer->isMalicious();

    return false;
}

void BootstrapOracle::setOverlayReadyIcon(const TransportAddress& address, bool ready)
{
	if (ev.isGUI()) {
		const char* color;
		if (ready) {
			color = isMalicious(address) ? "green" : ""; // change color if node is malicious
		} else {
			color = "red";
		}
		
		cModule* module = simulation.module(getPeerInfo(address)->getModuleID());
		
		module->parentModule()->displayString().setTagArg("i2", 1, color);
		module->displayString().setTagArg("i", 1, color);
	}
}
