//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BootstrapOracle.h
 * @author Markus Mauch, Robert Palmer
 */

#ifndef __BOOTSTRAPORACLE_H__
#define __BOOTSTRAPORACLE_H__

class PeerInfo;

#include <omnetpp.h>

#include <HashFunc.h>
#include <NodeHandle.h>
#include <OverlayKey.h>
#include <PeerInfo.h>
#include <vector>
#include <hashWatch.h>
#include <GlobalStatisticsAccess.h>


/**
 * bootstrapEntry consists of
 * TransportAddress and PeerInfo
 * and is used (together with
 * IPvXAddress) as an entry in the peerSet
 */
struct bootstrapEntry
{
    TransportAddress* node;
    PeerInfo* info;
    friend std::ostream& operator<<(std::ostream& Stream, const bootstrapEntry entry);
};


struct DHTEntry
{
    BinaryValue* value;
    simtime_t endtime;
    friend std::ostream& operator<<(std::ostream& Stream, const DHTEntry entry);
};


/**
 * Global modul that supports bootstrap process and key distribution
 *
 * @author Markus Mauch, Robert Palmer
 */
class BootstrapOracle : public cSimpleModule
{
    public:

        /**
        * holds all OverlayKeys
        */
        typedef std::vector<OverlayKey> KeyList;
        typedef std::map<OverlayKey, DHTEntry> DataMap;

        /**
        * Adds new peers to the peer set.
        *
        * Called automatically by the underlay,
        * when new peers are created.
        *
        * @param addr IPvXAddress of the peer to add
        * @param info underlay specific info of the peer to add
        */
        void addPeer(const IPvXAddress& ip, PeerInfo* info);

        /**
        * Removes a peer from the peerSet.
        *
        * Called automatically by the underlay,
        * when peers are removed.
        *
        * @param ip IPvXAddress of the peer to remove
        */
        virtual void killPeer(const IPvXAddress& ip);

	/**
	 * Returns a random NodeHandle
	 *
	 * Returns a random NodeHandle from the peerSet if at least one peer has
	 * been registered, an empty TransportAddress otherwise.
	 *
	 * @param bootstrappedNeeded does the node need to be bootstrapped?
	 *
	 * @return NodeHandle of the node
	 */
	virtual const NodeHandle& getRandomNode(bool bootstrappedNeeded);
	
        /**
	 * Returns a random NodeHandle
	 *
	 * Returns a random NodeHandle of an already bootstrapped node from the
	 * peerSet if at least one peer has been registered, an empty
	 * TransportAddress otherwise.
	 *
	 * @return NodeHandle of the node
	 */
        virtual const NodeHandle& getBootstrapNode();
	
        /**
	 * Bootstraps peers in the peer set.
	 *
	 * @param peer node to register
	 */
        virtual void registerPeer(const TransportAddress& peer);
	
        /**
	 * Bootstraps peers in the peer set.
	 *
	 * @param peer node to register
	 */
        virtual void registerPeer(const NodeHandle& peer);

        /**
	 * Debootstraps peers in the peer set
	 *
	 * @param peer node to remove
	 */
        virtual void removePeer(const TransportAddress& peer);
	
        /**
        * Returns a keylist
        *
        * @param maximumKeys maximum number of keys in new keylist
        * @return pointer to new keylist
        */
        virtual KeyList* getKeyList(uint maximumKeys);
	
        virtual DataMap* getDataMap();
	
        /**
	 * Returns random key from list
	 *
	 * @return the key
        */
        virtual const OverlayKey& getRandomKeyListItem() const;
	
	/**
	 * Colors module-icon blue (ready), green (ready, malicious) or red (not ready)
	 *
	 * @param address TransportAddress of the specified node
	 * @param ready state to visualize
	 */
	virtual void setOverlayReadyIcon(const TransportAddress& address, bool ready);
	
        /**
	 * Searches the peerSet for the specified node
	 *
	 * @param peer TransportAddress of the specified node
	 * @return PeerInfo of the node or NULL if node is not in peerSet
	 */
        virtual PeerInfo* getPeerInfo(const TransportAddress& peer);
	
	/**
	 * Set a node to be malicious
	 *
	 * @param address TransportAddress of the node
	 * @param malicious state to set
	 */
	virtual void setMalicious(const TransportAddress& address, bool malicious);
	
	/**
	 * Check if a node is malicious
	 *
	 * @param address TransportAddress of the node
	 * @return if the node is malicious
	 */
	virtual bool isMalicious(const TransportAddress& address);
	
	/**
	 * Selects a random node from the peerSet
	 *
	 * @param type If != 0, return a node of that type
	 * @returns The peerInfo of a random node
	 */
	virtual PeerInfo* getRandomPeerInfo(uint32_t nodeType = 0);
	
	/**
	 * Searches the peerSet for the specified node
	 *
	 * @param ip IPvXAddress of the specified node
	 * @return PeerInfo of the node or NULL if node is not in peerSet
	 */
	virtual PeerInfo* getPeerInfo(const IPvXAddress& ip);
protected:
	
	/**
	 * Init member function of module
	 */
	virtual void initialize();
	
	/**
	 * HandleMessage member function of module
	 *
	 * @param msg messag to handle
	 */
	virtual void handleMessage(cMessage* msg);
	
	/**
	 * Member function to create keylist
	 *
	 * @param size size of new keylist
	 */
	virtual void createKeyList(uint size);
	
	KeyList keyList; /**< the keylist */
	DataMap dataMap; /**< the datamap */
	uint bootstrappedPeerSize; /**< number of bootstrapped peers in the peer set */
	uint bootstrappedMaliciousNodes; /**< number of bootstrapped malicious nodes in the peer set */
	uint maliciousNodes; /**< number of malicious nodes in the peer set */
	double maliciousNodeRatio; /**< ratio of current malicious nodes when changing the ratio dynamically */
	cOutVector maliciousNodesVector; /**< vector that records the cange of malicious node rate */
	typedef __gnu_cxx::hash_map<IPvXAddress, bootstrapEntry> PeerHashMap; /**< Set of nodes participating in the overlay */
	PeerHashMap peerSet; /**< Set of nodes participating in the overlay */
	
	// OMNeT++ parameters
	uint maxNumberOfKeys; /**< parameter used by createKeyList() */
	double keyProbability; /**< probability of keys to be owned by nodes */
	
private:
	uint32 min_ip, max_ip; /**< used for random node choosing */
	GlobalStatistics* globalStatistics; /**< pointer to GlobalStatistics module in this node */
	
};

#endif
