
#ifndef __COMPARATOR_H
#define __COMPARATOR_H

#include<OverlayKey.h>

/**
 * Default Comparator
 */
template<class T>
class Comparator
{
public:
    /**
     * virtual destructor
     */
    virtual ~Comparator()
    {}
    
    /**
     * compares two variables of type T and indicates
     * which one is smaller or if they are equal
     * 
     * @param lhs first variable to compare
     * @param rhs second variable to compare
     * @return -1 if rhs is smaller, 0 if lhs and rhs are
     *         equal and 1 if rhs is greater
     */
    virtual int compare( const T& lhs, const T& rhs ) const
    {
        return lhs.compareTo(rhs);
    }
};

/**
 * OverlayKey comparator
 */
class KeyComparator : public Comparator<OverlayKey>
    {}
;

/**
 * OverlayKey standard metric
 */
class KeyStdMetric
{
public:
    /**
     * calculates the difference between x and y
     * 
     * @param x first parameter 
     * @param y second parameter
     * @return |x-y|
     */
    inline OverlayKey distance( const OverlayKey& x,
				const OverlayKey& y ) const
    {
        return x > y ? (x-y) : (y-x);
    }
};

/**
 * OverlayKey XOR Metric
 */
class KeyXorMetric
{
public:
    /**
     * calculates the difference between x and y
     * 
     * @param x first parameter
     * @param y second parameter
     * @return x XOR y
     */
    inline OverlayKey distance( const OverlayKey& x, 
				const OverlayKey& y ) const
    {
        return x^y;
    }
};

/**
 * OverlayKey Ring Metric
 */
class KeyRingMetric
{
public:
    /**
     * calculates the difference between x and y on a bidirectional ring
     * 
     * @param x first parameter
     * @param y second parameter
     * @return |x-y| on a bidirectional ring
     */
    inline OverlayKey distance( const OverlayKey& x, 
				const OverlayKey& y ) const
    {
	const OverlayKey* smaller;
	const OverlayKey* bigger;
	
	if (x > y)
	{
	    smaller = &y;
	    bigger = &x;
	}
	else
	{
	    smaller = &x;
	    bigger = &y;
	}
	
	OverlayKey diff1(*bigger - *smaller);
	OverlayKey diff2(*smaller + (OverlayKey::max() - *bigger) + 1);
	
	if (diff1 > diff2)
	    return diff2;
	else
	    return diff1;
    }
};

/**
 * OverlayKey distance comparator
 */
template<class Metric = KeyStdMetric>
class KeyDistanceComparator : public Comparator<OverlayKey>
{
private:
    Metric m; /**< indicates which metric to use for the comparison */
    OverlayKey key; /**< the relative key to which distances are compared */ 
public:

    /**
     * constructor
     */
    KeyDistanceComparator( const OverlayKey& relativeKey )
    {
        this->key = relativeKey;
    }

    /**
     * indicates which of the two given keys has a longer distance to 
     * the relative key
     * 
     * @param lhs first key
     * @param rhs second key
     * @return -1 if lhs is closer, 0 if lhs and rhs are equal and 1
     *         if rhs closer to the relative key
     */
    int compare( const OverlayKey& lhs, const OverlayKey& rhs ) const
    {
        return m.distance( key, lhs ) .compareTo( 
               m.distance( key, rhs ) );
    }
};

#endif

