//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file GlobalStatistics.h
 * @author Ingmar Baumgart
 */

#ifndef __GLOBALSTATISTICS_H__
#define __GLOBALSTATISTICS_H__

#include <set>
#include <vector>
#include <map>

#include <omnetpp.h>

#include <OverlayKey.h>

/**
 * Macro used for recording statistics considering measureNetwIn
 * parameter. The do-while-loop is needed for compatability in
 * outer if-else-structures.
 */
#define RECORD_STATS(x) \
    do { \
        if (globalStatistics->isMeasuring()){ x; } \
    } while(false)


/**
 * Modul to record global statistics
 *
 * @author Ingmar Baumgart
 */
class GlobalStatistics : public cSimpleModule
{
public:
    double sentKBRTestAppMessages; //!< total number of messages sent by KBRTestApp
    double deliveredKBRTestAppMessages; //!< total number of messages delivered by KBRTestApp
    int testCount;
    cOutVector currentDeliveryVector; //!< statistical output vector for current delivery ratio
    
    
    int nodesInitialized; //!< nodes that have run initialize() and "registered" with GlobalStatistics
    int nodesFinished; //!< nodes that have run finished() and have "deregistered" with GlobalStatistics
    
	/**
	 * Destructor
	 */
	~GlobalStatistics();

	/**
	 * Do the actual finish() call and record scalars
	 * needed because finish() gets called before all nodes have transmitted their data to GlobalStatistics
	 */
	void doFinish();

	/**
	 * add a value to a cStdDev
	 */
	void addStdDev(const char *name, double value);


	/**
	 * record a value to a global cOutVector defined by name
	 */
	void recordOutVector(const char *name, double value);

    void startMeasuring();

    inline bool isMeasuring() { return measuring; }; 
    inline simtime_t getMeasureStartTime() { return measureStartTime; };

    simtime_t calcMeasuredLifetime(simtime_t creationTime);

protected:

	class OutVector //!< struct for cOutVectors and cummulated values
	{
	public:
		cOutVector* vector; //!< output vector
		int count; //!< number of recorded values
		double value; //!< sum of values

		OutVector(const char* name)
		{
			vector = new cOutVector(name);
			value = count = 0;
		}

		~OutVector() { delete vector; }
	};

	map<const char*, cStdDev*> cStdDevMap; //!< map to store and access the statistics data
	map<const char*, OutVector*> cOutVectorMap; //!< map to store and access the output vectors
	cMessage* globalStatTimer; //!< timer for periodic statistic updates
	double globalStatTimerInterval; //!< interval length of periodic statistic timer

    /**
     * Init member function of module
     */
    virtual void initialize();

    /**
     * HandleMessage member function of module
     */
    virtual void handleMessage(cMessage* msg);

    /**
     * Finish member function of module
     */
    virtual void finish();    
    
    bool measureNetwInitPhase;
    
    bool measuring;

    simtime_t measureStartTime;
};

#endif
