//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file ParetoChurn.cc
 * @author Helge Backhaus
 * @author Ingmar Baumgart
 * @author Stephan Krause
 */

#include "ParetoChurn.h"
#include "GlobalStatisticsAccess.h"

Define_Module(ParetoChurn);

void ParetoChurn::initializeChurn()
{
    Enter_Method_Silent();

    initialMean = par("initialMobilityDelay");
    initialDeviation = initialMean / 3;
    lifetimeMean = par("lifetimeMean");
    deadtimeMean = par("deadtimeMean");

    WATCH(lifetimeMean);
    WATCH(deadtimeMean);

    lastCreatetime = 0;
    lastDeletetime = 0;

    globalStatistics = GlobalStatisticsAccess().get();

    double initFinishedTime = initialMean * targetOverlayTerminalNum;
    if (initialOverlayTerminalNum) opp_warning("ParetoChurn ignores parameter "
	    "initialOverlayTerminalNum, but it is set to a number != 0");

   
    randomNodeOffset = intuniform(0, 0xFFFFFFFF);

    // try to create a stable equilibrium of nodes in init phase
    //
    // check for each node if he is present in initial state
    for (int i = 0; i < (int)par("targetOverlayTerminalNum"); i++) {

	double nodeLifetimeMean = individualMeanTime(lifetimeMean);
	double nodeDeadtimeMean = individualMeanTime(deadtimeMean);
	double nodeAvailability = nodeLifetimeMean/(nodeLifetimeMean
						    + nodeDeadtimeMean);

	RECORD_STATS(globalStatistics->recordOutVector("Node availability",
						       nodeAvailability));

	// draw a random value. if it is smaller than availability,
	// then node is alive
        // note: residual lifetime is shiftedPareto(2, beta) and not
	// shiftedPareto(3, beta)
	if (uniform(0, 1) < nodeAvailability) {
	    double scheduleTime = truncnormal(initialMean*i, initialDeviation);
	    scheduleCreateNodeAt(scheduleTime, initFinishedTime - scheduleTime
				 + residualLifetime(nodeLifetimeMean), 
				 nodeLifetimeMean, nodeDeadtimeMean);
	} else {
	    // node is dead, schedule creation at the end of residual dead time
	    scheduleCreateNodeAt(initFinishedTime
				 + residualLifetime(nodeDeadtimeMean),
				 individualLifetime(nodeLifetimeMean),
				 nodeLifetimeMean, nodeDeadtimeMean);
	}
    }

    initFinishedTimer = new cMessage("initFinishTimer");
    scheduleAt(initFinishedTime, initFinishedTimer);
}

void ParetoChurn::handleMessage(cMessage* msg)
{
    if (!msg->isSelfMessage()) {
        delete msg;
        return;
    }

    // init phase finished
    if (msg ==  initFinishedTimer) {
	underlayConfigurator->initFinished();
	delete msg;
	return;
    }

    ParetoChurnMessage* churnMsg = check_and_cast<ParetoChurnMessage*>(msg);
    
    if (churnMsg->getCreateNode() == true) {
	createNode(churnMsg->getLifetime(), churnMsg->getMeanLifetime(),
		churnMsg->getMeanDeadtime(), false);
    } else {
	deleteNode(churnMsg->getAddr(), churnMsg->getMeanLifetime(),
		churnMsg->getMeanDeadtime());
    }
    
    delete msg;
}

void ParetoChurn::createNode(double lifetime, double meanLifetime,
			     double meanDeadtime, bool initialize)
{
    ParetoChurnMessage* churnMsg = new ParetoChurnMessage("DeleteNode");
    churnMsg->setAddr(*underlayConfigurator->createNode(type, initialize));
    churnMsg->setCreateNode(false);
    churnMsg->setMeanLifetime(meanLifetime);
    churnMsg->setMeanDeadtime(meanDeadtime);
    scheduleAt(simulation.simTime() + lifetime, churnMsg);

    RECORD_STATS(globalStatistics->recordOutVector("BaseOverlay: Session Time",
						   lifetime));

    RECORD_STATS(globalStatistics->recordOutVector(
		     "BaseOverlay: Interarrivaltime",
		     simulation.simTime() - lastCreatetime));

    lastCreatetime = simulation.simTime();
}

void ParetoChurn::deleteNode(TransportAddress& addr, double meanLifetime,
			     double meanDeadtime)
{
    // Kill node
    underlayConfigurator->preKillNode(NodeType(), &addr);

    RECORD_STATS(globalStatistics->recordOutVector(
		     "BaseOverlay: Interdeletiontime",
		     simulation.simTime() - lastDeletetime));
    lastDeletetime = simulation.simTime();
    scheduleCreateNodeAt(simulation.simTime()+individualLifetime(meanDeadtime),
			 individualLifetime(meanLifetime), meanLifetime,
			 meanDeadtime);
}

void ParetoChurn::scheduleCreateNodeAt(double creationTime, double lifetime,
				       double meanLifetime, double meanDeadtime)
{
    ParetoChurnMessage* churnMsg = new ParetoChurnMessage("CreateNode");
    churnMsg->setCreateNode(true);
    churnMsg->setLifetime(lifetime);
    churnMsg->setMeanLifetime(meanLifetime);
    churnMsg->setMeanDeadtime(meanDeadtime);
    scheduleAt(creationTime, churnMsg);
}

double ParetoChurn::betaByMean(double mean, double alpha)
{
    return 1/(mean*(alpha -1));
}

double ParetoChurn::shiftedPareto(double a, double b, int rng)
{
    // What OMNET calles "pareto_shifted" in reality is a gerneralized pareto,
    // not a shifted pareto...
    return (pareto_shifted(a, b, 0, rng)/b - 1) / b;
}

double ParetoChurn::individualMeanTime(double mean)
{
    return shiftedPareto(3, betaByMean(mean));
}

double ParetoChurn::individualLifetime(double mean)
{
    return shiftedPareto(3, betaByMean(mean));
}

// Residual lifetime is shifted pareto with the same beta, but
// decreased alpha i.e. beta calculation is based on "old alpha" (in
// this case 3), while "actual alpha" is 2
double ParetoChurn::residualLifetime(double mean)
{
    return shiftedPareto(2, betaByMean(mean));
}

void ParetoChurn::updateDisplayString()
{
    char buf[80];
    sprintf(buf, "pareto churn");
    displayString().setTagArg("t", 0, buf);
}

ParetoChurn::~ParetoChurn() {
    // destroy self timer messages
    cancelAndDelete(mobilityTimer);
    cancelAndDelete(initFinishedTimer);
}
