//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file RandomChurn.cc
 * @author Helge Backhaus
 */

#include "RandomChurn.h"

Define_Module(RandomChurn);

void RandomChurn::initializeChurn()
{
    Enter_Method_Silent();

    creationProbability = par("creationProbability");
    migrationProbability = par("migrationProbability");
    removalProbability = par("removalProbability");
    initialMean = par("initialMobilityDelay");
    initialDeviation = initialMean / 3;
    targetMean = par("targetMobilityDelay");
    targetOverlayTerminalNum = par("targetOverlayTerminalNum");
    //targetDeviation = targetMean / 3;
    WATCH(targetMean);

    churnTimer = NULL;
    churnIntervalChanged = false;
    churnChangeInterval = (double)(par("churnChangeInterval"));
    initAddMoreTerminals = true;

    // add the overlay nodes
    //underlayConfigurator->firstNodeId = createRandomNode(true)->getModuleID();
    underlayConfigurator->createNode(type, true);
    for (int i = 1; i < initialOverlayTerminalNum; i++) {
        underlayConfigurator->createNode(type, true);
    }

    // initialize simulation
    mobilityTimer = NULL;
    mobilityTimer = new cMessage("mobilityTimer");
    scheduleAt(simulation.simTime(), mobilityTimer);
    if((double)(par("churnChangeInterval")) > 0) {
	churnTimer = new cMessage("churnTimer");
	scheduleAt(simulation.simTime() + churnChangeInterval, churnTimer);
    }
}

void RandomChurn::handleMessage(cMessage* msg)
{
    if (!msg->isSelfMessage()) {
        delete msg;
        return;
    }

    if (msg == churnTimer) {
        cancelEvent(churnTimer);
        scheduleAt(simulation.simTime() + churnChangeInterval, churnTimer);
        if (churnIntervalChanged) {
            targetMean = par("targetMobilityDelay");
            churnIntervalChanged = false;
        }
        else {
            targetMean = par("targetMobilityDelay2");
            churnIntervalChanged = true;
        }
        std::stringstream temp;
        temp << "Churn-rate changed to " << targetMean;
        bubble(temp.str().c_str());
    } else if (msg == mobilityTimer) {
        if (initAddMoreTerminals) {
            // increase the number of nodes steadily during setup
            if (terminalCount < targetOverlayTerminalNum) {
                underlayConfigurator->createNode(type);
	    }

	    if (terminalCount >= targetOverlayTerminalNum) {
		initAddMoreTerminals = false;
		underlayConfigurator->initFinished();
	    }
            scheduleAt(simulation.simTime()
                    + truncnormal(initialMean, initialDeviation), msg);
        }
	else {
	    double random = uniform(0, 1);

	    // modify the number of nodes according to user parameters
	    if (random < creationProbability) {
		underlayConfigurator->createNode(type);
	    } else if (creationProbability <= random &&
		    random < creationProbability + removalProbability &&
		    terminalCount > 1) {
		underlayConfigurator->preKillNode(type);
	    } else if (creationProbability + removalProbability <= random &&
		       random < creationProbability + removalProbability
		       + migrationProbability && terminalCount > 1) {
		underlayConfigurator->migrateNode(type);
	    }
	    scheduleAt(simulation.simTime()
		       + truncnormal(targetMean, targetMean / 3), msg);
	}
    }
}

void RandomChurn::updateDisplayString()
{
    char buf[80];
    sprintf(buf, "random churn");
    displayString().setTagArg("t", 0, buf);
}

RandomChurn::~RandomChurn() {
    // destroy self timer messages
    cancelAndDelete(churnTimer);
    cancelAndDelete(mobilityTimer);
}

