//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file TransportAddress.cc
 * @author Markus Mauch
 * @author Sebastian Mies
 */

#include <omnetpp.h>

#include <IPAddress.h>

#include "TransportAddress.h"
#include "BaseOverlay.h"

class BaseOverlay;

// predefined transport address
const TransportAddress TransportAddress::UNSPECIFIED_NODE;

std::ostream& operator<<(std::ostream& os, const TransportAddress& n)
{
    if (n.isUnspecified()) {
        os << "<unspec>";
    } else {
        os << n.ip << ":" << n.port;
    }

    return os;
};


//default-constructor
TransportAddress::TransportAddress()
{
    port = -1;
}

//copy constructor
TransportAddress::TransportAddress( const TransportAddress& handle )
{
    port = handle.port;
    ip = handle.ip;
}


//complete constructor
TransportAddress::TransportAddress( const IPvXAddress& ip, int port)
{
    this->ip = ip;
    this->port = port;
}

//public
bool TransportAddress::isUnspecified() const
{
    return (ip.isUnspecified() || (port == -1));
}

//public
TransportAddress& TransportAddress::operator=(const TransportAddress& rhs)
{
    this->ip = rhs.ip;
    this->port = rhs.port;

    return *this;
}

//public
bool TransportAddress::operator==(const TransportAddress& rhs) const
{
    assertUnspecified(rhs);
    return (this->ip == rhs.ip && this->port == rhs.port);
}

//public
bool TransportAddress::operator!=(const TransportAddress& rhs) const
{
    assertUnspecified(rhs);
    return !(this->ip == rhs.ip && this->port == rhs.port );
}

//public
bool TransportAddress::operator<(const TransportAddress &rhs) const
{
    assertUnspecified(rhs);
    return this->ip < rhs.ip;
}

//public
bool TransportAddress::operator>(const TransportAddress &rhs) const
{
    assertUnspecified(rhs);
    return !(this->ip < rhs.ip || this->ip == rhs.ip);
}

//public
bool TransportAddress::operator<=(const TransportAddress &rhs) const
{
    assertUnspecified(rhs);
    return this->ip < rhs.ip || this->ip == rhs.ip;
}

//public
bool TransportAddress::operator>=(const TransportAddress &rhs) const
{
    assertUnspecified(rhs);
    return !(this->ip < rhs.ip);
}

//public
void TransportAddress::setAddress( const IPvXAddress& ip, int port )
{
    this->ip = ip;
    if (port!=-1)
        this->port = port;
}

//public
void TransportAddress::setPort( int port )
{
    this->port = port;
}

//public
const IPvXAddress& TransportAddress::getAddress() const
{
    return ip;
}

//public
int TransportAddress::getPort() const
{
    return port;
}

//public
size_t TransportAddress::hash() const
{
    size_t iphash;
    if (ip.isIPv6()) {
        uint32_t* addr = (uint32_t*) ip.get6().words();
        iphash = (size_t)(addr[0]^addr[1]^addr[2]^addr[3]);
    } else {
        iphash = (size_t)ip.get4().getInt();
    }

    return (size_t)(iphash^port);
}


//private
inline void TransportAddress::assertUnspecified( const TransportAddress& handle ) const
{
    if ( this->isUnspecified() || handle.isUnspecified() )
        opp_error("TransportAddress: Trying to compare unspecified TransportAddress!");
}


// Automatically supply array (un)packing functions
template<typename T>
void doPacking(cCommBuffer *b, T *t, int n) {
    for (int i=0; i<n; i++)
        doPacking(b,t[i]);
}
template<typename T>
void doUnpacking(cCommBuffer *b, T *t, int n) {
    for (int i=0; i<n; i++)
        doUnpacking(b,t[i]);
}
inline void doPacking(cCommBuffer *, cPolymorphic&) {}
inline void doUnpacking(cCommBuffer *, cPolymorphic&) {}

#define DOPACKING(T,R) \
    inline void doPacking(cCommBuffer *b, T R a) {b->pack(a);}  \
    inline void doPacking(cCommBuffer *b, T *a, int n) {b->pack(a,n);}  \
    inline void doUnpacking(cCommBuffer *b, T& a) {b->unpack(a);}  \
    inline void doUnpacking(cCommBuffer *b, T *a, int n) {b->unpack(a,n);}
#define _
DOPACKING(char,_)
DOPACKING(unsigned char,_)
DOPACKING(bool,_)
DOPACKING(short,_)
DOPACKING(unsigned short,_)
DOPACKING(int,_)
DOPACKING(unsigned int,_)
DOPACKING(long,_)
DOPACKING(unsigned long,_)
DOPACKING(float,_)
DOPACKING(double,_)
DOPACKING(long double,_)
DOPACKING(char *,_)
DOPACKING(const char *,_)
DOPACKING(opp_string,&)
//DOPACKING(std::string,&)
#undef _
#undef DOPACKING

void TransportAddress::netPack(cCommBuffer *b)
{
    //cMessage::netPack(b);
    doPacking(b,this->ip);
    doPacking(b,this->port);
}

void TransportAddress::netUnpack(cCommBuffer *b)
{
    //cMessage::netUnpack(b);
    doUnpacking(b,this->ip);
    doUnpacking(b,this->port);
}
