//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file UnderlayConfigurator.cc
 * @author Stephan Krause, Bernhard Heep
 */

#include <omnetpp.h>
#include <BootstrapOracleAccess.h>
#include <ChurnGeneratorAccess.h>
#include <GlobalStatisticsAccess.h>

#include "UnderlayConfigurator.h"


int UnderlayConfigurator::numInitStages() const
{
    return MAX_STAGE_UNDERLAY + 1;
}

void UnderlayConfigurator::initialize(int stage)
{
    if(stage == MIN_STAGE_UNDERLAY) {
	gracefullLeaveDelay = par("gracefullLeaveDelay");
	gracefullLeavePropability = par("gracefullLeavePropability");
	channelTypes = cStringTokenizer(par("channelTypes"), " ").asVector();

	transitionTime = par("transitionTime");
	measurementTime = par("measurementTime");

	bootstrapOracle = BootstrapOracleAccess().get();
	globalStatistics =GlobalStatisticsAccess().get();

	keepFirstNode = par("keepFirstNode");

	if (keepFirstNode)
	    opp_error("UnderlayConfigurator: keepFirstNode is not supported in this version!");

	endSimulationTimer = new cMessage("endSimulationTimer");
	endTransitionTimer = new cMessage("endTransitionTimer");
 
	gettimeofday(&initStartTime, NULL);	 
	init = true;
	//transitionTimeFinished = false;
	initCounter = 0;

	firstNodeId = -1;
	WATCH(firstNodeId);
    }

    if(stage >= MIN_STAGE_UNDERLAY && stage <= MAX_STAGE_UNDERLAY) {
	initializeUnderlay(stage);
    }

    if(stage == MAX_STAGE_UNDERLAY) {
       // Create churn gernerators
	NodeType t;
	std::vector<std::string> churnGeneratorTypes = 
	    cStringTokenizer(par("churnGeneratorTypes"), " ").asVector();

	for( std::vector<std::string>::iterator it=churnGeneratorTypes.begin();
		it != churnGeneratorTypes.end(); ++it) {
	    
	    ChurnGenerator* gen = dynamic_cast<ChurnGenerator*> 
		(findModuleType(it->c_str())->create(it->c_str(), parentModule()));
	    
	    // Name the generator
	    char buf[80];
	    sprintf(buf, "churnGenerator[%i]", t.typeID);
	    gen->setName(buf);

	    // Add it to the list of generators and initialize it
	    churnGenerator.push_back(gen);
	    t.typeID++;
	    gen->setNodeType(t);
	    gen->buildInside();
	    gen->setDisplayString("i=block/cogwheel");
	    gen->callInitialize(stage);
	    gen->terminalCount = 0;
	    gen->initializeChurn();
	}
    }
}

void UnderlayConfigurator::initFinished()
{
    Enter_Method_Silent();

    if(++initCounter = churnGenerator.size()) {
	init = false;
	gettimeofday(&initFinishedTime, NULL);

	scheduleAt(simulation.simTime() + transitionTime,
		   endTransitionTimer);
	
	//transitionTimeFinished = true;	

	if (measurementTime >= 0) {
	    scheduleAt(simulation.simTime() + transitionTime + measurementTime,
		       endSimulationTimer);
	}
	consoleOut("INIT phase finished");	
    }
}

void UnderlayConfigurator::handleMessage(cMessage* msg)
{
    if(msg == endSimulationTimer) {
	endSimulation();
    } else if(msg == endTransitionTimer) {
	consoleOut("transition time finished");
	//transitionTimeFinished = true;
	globalStatistics->startMeasuring();
    } else {
	handleTimerEvent(msg);
    }
}

void UnderlayConfigurator::handleTimerEvent(cMessage* msg)
{
    delete msg;
}

void UnderlayConfigurator::finish()
{
    cancelAndDelete(endSimulationTimer);
    cancelAndDelete(endTransitionTimer);

    finishUnderlay();
}

void UnderlayConfigurator::finishUnderlay()
{
    //...
}

void UnderlayConfigurator::consoleOut(const std::string& text)
{
    if(!ev.isGUI()) {
	struct timeval now, diff;
	gettimeofday(&now, NULL);
	timersub(&now, &initStartTime, &diff);

	std::stringstream ss;
	std::string line1(71, '*');
	std::string line2(71, '*');

	ss << "   " << text << "   ";
	line1.replace(35 - ss.str().size() / 2,
		      ss.str().size(),
		      ss.str());
	ss.str("");

	ss << "   (sim time: " << simTime()
		<< ", real time: " << diff.tv_sec
		<< "." << diff.tv_usec << ")   ";
	line2.replace(35 - ss.str().size() / 2,
		      ss.str().size(),
		      ss.str());

	std::cout << "\n" << line1 << "\n"
		  << line2 << "\n" << std::endl;
    } else {
	EV << "(UnderlayConfigurator) " << text;
    }
    
}
