//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file UnderlayConfigurator.h
 * @author Stephan Krause, Bernhard Heep
 */

#ifndef _UNDERLAYCONFIGURATOR_H_
#define _UNDERLAYCONFIGURATOR_H_

#include <sys/time.h>
#include <time.h>

#include <omnetpp.h>
#include <BootstrapOracle.h>
#include <ChurnGenerator.h>
#include <PeerInfo.h>

#include <InitStages.h>

class ChurnGenerator;
class NodeType;

/**
 * Base class for configurators of different underlay models
 *
 * @author Stephan Krause, Bernhard Heep
 */
class UnderlayConfigurator : public cSimpleModule
{
public:

    /**
     * still in initialization phase?
     */
    bool isInit() { return init; };

    bool isTransitionTimeFinished() { return transitionTimeFinished; };

    /**
     * Creates an overlay node
     *
     * @param initialize
     */
    virtual TransportAddress* createNode(NodeType type, bool initialize=false) = 0;

    /**
     * Notifies and schedules overlay nodes for removal.
     *
     * @param addr NULL for random node
     */
    virtual void preKillNode(NodeType type, TransportAddress* addr=NULL) = 0;

    /**
     * Migrates overlay nodes from one access net to another.
     *
     * @param addr NULL for random node
     */
    virtual void migrateNode(NodeType type, TransportAddress* addr=NULL) = 0;

    void initFinished();

protected:

    /**
     * OMNeT number of init stages
     */
    int numInitStages() const;

    /**
     * OMNeT init methods
     */
    virtual void initialize(int stage);

    /**
     * Init method for derived underlay configurators
     */
    virtual void initializeUnderlay(int stage) = 0;

    virtual void handleTimerEvent(cMessage* msg);

    /**
     * Cleans up configurator
     */
    void finish();

    /**
     * Cleans up concrete underlay configurator
     */
    virtual void finishUnderlay();

    /**
     * Sets display string
     */
    virtual void setDisplayString() = 0;

    /**
     * Node mobility simulation
     *
     * @param msg timer-message
     */
    void handleMessage(cMessage* msg);

    std::vector<std::string> channelTypes; //!< possible access types

    int overlayTerminalCount; //!< current number of overlay terminals
    bool keepFirstNode; //!< is the first node kept during simulation?
    int firstNodeId; //!< the Id of the overlayTerminal created first in the overlay
    double gracefullLeaveDelay; //!< delay until scheduled node is removed from overlay
    double gracefullLeavePropability; //!< propability that node is notified befor removal

    BootstrapOracle* bootstrapOracle; //!< pointer to BootstrapOracle
    GlobalStatistics* globalStatistics; //!< pointer to GlobalStatistics
    vector<ChurnGenerator*> churnGenerator; //!< pointer to the ChurnGenerators

    cMessage* endSimulationTimer; //!< timer to signal end of simulation
    cMessage* endTransitionTimer; //!< timer to signal end of transition time

    struct timeval initFinishedTime; //!< timestamp at end of init phase
    struct timeval initStartTime; //!< timestamp at begin of init phase

    double transitionTime; //!< time to wait before measuring after init phase is finished
    double measurementTime; //!< duration of the simulation after init and transition phase

 private:

    bool init;
    bool transitionTimeFinished;
    unsigned int initCounter;
    
    void consoleOut(const std::string& text);
};

#endif
