//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BrooseBucket.cc
 * @author Jochen Schenk
 */

#include <omnetpp.h>
#include "BrooseBucket.h"

using namespace std;

Define_Module(BrooseBucket);

const int MAXBITS = 1024;

void BrooseBucket::initialize(int stage)
{
    // because of IPAddressResolver, we need to wait until interfaces
    // are registered, address auto-assignment takes place etc.
    if (stage != 5)
        return;

    WATCH_MAP(bucket);

}

void BrooseBucket::handleMessage(cMessage* msg)
{
    error("BrooseBucket::handleMessage() shouldn't be called!");
}

void BrooseBucket::initializeBucket(int shiftingBits, uint prefix,
				    BrooseHandle node, int size,
				    Broose* overlay, bool isBBucket)
{
    maxSize = size;
    thisNode = node;
    bits = shiftingBits;
    this->overlay = overlay;
    this->isBBucket = isBBucket;

    if (shiftingBits < 0) {
        key = thisNode.key << -shiftingBits;
    } else {
        key = thisNode.key >> shiftingBits;
    }

    if (prefix != 0) {
        OverlayKey tmp(prefix); // constraint
        tmp = tmp << (thisNode.key.getLength() - shiftingBits);
        key = key + tmp;
    }
    bucket.clear();
}

void BrooseBucket::add(BrooseHandle node)
{
    OverlayKey tmp = key ^ node.key;
    if (bucket.size() < maxSize) {
        if (bucket.insert(make_pair(tmp,node)).second) {
	    if (isBBucket) {
		overlay->callUpdate(node, true);
	    }
	}
    } else {
        if (bucket.find(tmp) == bucket.end()) {
            bucketIter = bucket.end();
            bucketIter--;

            // is the new node closer than the one farthest away,
	    // remove the one and add the other

            if (bucketIter->first > tmp) {
		if (isBBucket) {
		    overlay->callUpdate(bucketIter->second, false);
		}
                bucket.erase(bucketIter);
		if (bucket.insert(make_pair(tmp,node)).second) {
		    if (isBBucket) {
			overlay->callUpdate(node, true);
		    }
		}
            }
        }
    }

}

void BrooseBucket::remove(BrooseHandle node)
{
    for (bucketIter = bucket.begin(); bucketIter != bucket.end();) {
        if (bucketIter->second.ip == node.ip) {
	    if (isBBucket) {
		overlay->callUpdate(node, false);
	    }
            bucket.erase(bucketIter++);
	} else {
            ++bucketIter;
	}
    }
}



const BrooseHandle& BrooseBucket::get(uint pos)
{
    if (pos > bucket.size()) {
        error("Index out of bounds(BrooseBucket).");
    }

    uint i = 0;
    for (bucketIter = bucket.begin(); bucketIter != bucket.end(); bucketIter++){
        if (pos == i) {
            return bucketIter->second;
	}
        i++;
    }
    return BrooseHandle::UNSPECIFIED_NODE;
}

const OverlayKey& BrooseBucket::getDist(uint pos)
{
    if(pos > bucket.size()) {
        error("Index out of bounds(BrooseBucket).");
    }

    uint i = 0;
    for (bucketIter = bucket.begin(); bucketIter != bucket.end(); bucketIter++){
        if (pos == i) {
            return bucketIter->first;
	}
        i++;
    }
    return OverlayKey::UNSPECIFIED_KEY;
}


uint BrooseBucket::getSize()
{
    return bucket.size();
}

uint BrooseBucket::getMaxSize()
{
    return maxSize;
}

bool BrooseBucket::isEmpty()
{
    return bucket.empty();
}

void BrooseBucket::clear()
{
    bucket.clear();
}

BrooseHandle BrooseBucket::getClosestNode(const OverlayKey& destKey)
{
    BrooseHandle closestNode;
    OverlayKey lastDist = OverlayKey::max();
    OverlayKey dist;

    if (bucket.empty()) {
        EV << "--- BUCKET EMPTY ---" << endl;
    }

    for (bucketIter = bucket.begin(); bucketIter!=bucket.end(); bucketIter++) {
        dist = bucketIter->second.key ^ destKey;
        if (dist < lastDist) {
            closestNode = bucketIter->second;
            lastDist = dist;
        }
    }
    return closestNode;
}

int BrooseBucket::longestPrefix()
{
    int bits = 0;
    int maxBits = MAXBITS;
    int tmp = 0;
    BrooseHandle node1, node2;

    if (bucket.size() <= 1)
        return 0;

    for (bucketIter = bucket.begin(); bucketIter != bucket.end(); ++bucketIter){
        tmp = 0;
        node1 = bucketIter->second;
        ++bucketIter;

        if (bucketIter != bucket.end()) {
            node2 = bucketIter->second;
            --bucketIter;

            for (uint i = 1; i <= node1.key.getLength(); i++) {
                if (node1.key.bitAtPlace(i) == node2.key.bitAtPlace(i))
                    tmp++;
                else
                    break;
            }

            if (tmp < maxBits) {
                bits = tmp;
                maxBits = tmp;
            }
        } else {
            --bucketIter;
	}
    }
    return bits;
}

void BrooseBucket::output(int maxEntries)
{
    EV << "(" << bucket.size() << "/" << maxSize <<")" << endl;
    BrooseHandle node;
    OverlayKey dist;

    int max;
    max = bucket.size();
    if (maxEntries != 0 && maxEntries < max) {
        max = maxEntries;
    }
    int i;

    for (bucketIter = bucket.begin(), i = 0; i < max; bucketIter++, i++) {
        dist = bucketIter->first;
        node = bucketIter->second;
        EV << dist << " " << node.key << " " << node.ip << " RTT: "
	   << node.rtt << " LS: " << node.lastSeen << endl;
    }
}

bool BrooseBucket::keyInRange (const OverlayKey& key)
{
    OverlayKey dist;

    if (bucket.size() == 0)
        return false;

    // check if the function was called to perform on a B bucket
    if (bits == 0) {
        if (bucket.size() <=  (maxSize / 7))
            dist = OverlayKey::max();
        else
            dist = getDist((maxSize / 7) - 1);
    } else
        dist = getDist(bucket.size()-1);

    if ((key ^ thisNode.key) <= dist)
        return true;
    else
        return false;

}

bool BrooseBucket::nodeInBucket(const BrooseHandle& node)
{
    for (bucketIter = bucket.begin(); bucketIter!=bucket.end(); bucketIter++) {
        if (node.ip == bucketIter->second.ip)
            return true;
    }
    return false;
}

int BrooseBucket::getFailedResponses (const BrooseHandle& node)
{
    for (bucketIter = bucket.begin(); bucketIter!=bucket.end(); bucketIter++) {
        if (node.ip == bucketIter->second.ip)
            return bucketIter->second.failedResponses;
    }
    return -1;
}

void BrooseBucket::increaseFailedResponses (const BrooseHandle& node)
{
    for (bucketIter = bucket.begin(); bucketIter!=bucket.end(); bucketIter++) {
        if (node.ip == bucketIter->second.ip)
            bucketIter->second.failedResponses++;
    }
}

void BrooseBucket::resetFailedResponses (const BrooseHandle& node)
{
    for (bucketIter = bucket.begin(); bucketIter!=bucket.end(); bucketIter++) {
        if (node.ip == bucketIter->second.ip)
            bucketIter->second.failedResponses = 0;
    }
}


void BrooseBucket::setRTT(BrooseHandle node, simtime_t rpcRTT)
{
    for (bucketIter = bucket.begin(); bucketIter!=bucket.end(); bucketIter++) {
        if (node.ip == bucketIter->second.ip) {
            bucketIter->second.rtt = rpcRTT;
        }
    }
}

simtime_t BrooseBucket::getRTT(BrooseHandle node)
{
    for (bucketIter = bucket.begin(); bucketIter!=bucket.end(); bucketIter++) {
        if (node.ip == bucketIter->second.ip)
            return bucketIter->second.rtt;
    }
    return -2;
}

void BrooseBucket::setLastSeen(BrooseHandle node, simtime_t time)
{
    for (bucketIter = bucket.begin(); bucketIter!=bucket.end(); bucketIter++) {
        if (node.ip == bucketIter->second.ip)
            bucketIter->second.lastSeen = time;
    }
}

simtime_t BrooseBucket::getLastSeen(BrooseHandle node)
{
    for (bucketIter = bucket.begin(); bucketIter!=bucket.end(); bucketIter++) {
        if (node.ip == bucketIter->second.ip)
            return bucketIter->second.lastSeen;
    }
    return -2;
}

void BrooseBucket::binaryOutput(const OverlayKey& key)
{
    if(key.isUnspecified())
        cout << "<unspec>";
    else {
        for (unsigned int i = 1; i <= key.getLength(); i++)
            cout << key.bitAtPlace(i);
    }
}
