//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file ChordFingerTable.cc
 * @author Markus Mauch, Ingmar Baumgart
 */

#include <cfloat>

#include <omnetpp.h>
#include <IPAddressResolver.h>

#include <NodeHandle.h>
#include <OverlayKey.h>

#include "ChordFingerTable.h"


Define_Module(ChordFingerTable);

void ChordFingerTable::initialize(int stage)
{
    // because of IPAddressResolver, we need to wait until interfaces
    // are registered, address auto-assignment takes place etc.
    if(stage != MIN_STAGE_OVERLAY)
        return;

    WATCH_VECTOR(fingerTable);
}

void ChordFingerTable::handleMessage(cMessage* msg)
{
    error("this module doesn't handle messages, it runs only in initialize()");
}

void ChordFingerTable::initializeTable(uint size, const NodeHandle& owner,
				       BaseOverlay* overlay)
{
    Successors temp;
    //temp.insert(std::make_pair(-1, owner));
    //fingerTable.assign(size, temp);
    fingerTable.assign(size, FingerEntry(owner, temp));
    //std::cout << fingerTable[0].second.size() << std::endl;
    this->overlay = overlay;
}

uint ChordFingerTable::getSize()
{
    return fingerTable.size();
}

void ChordFingerTable::setFinger(uint pos, const NodeHandle& node)
{
    if (pos < fingerTable.size()) {
	Successors temp;
        fingerTable[pos] = FingerEntry(node, temp);
    } else {
        error("Index out of bound (ChordFingerTable, setFinger())");
    }
}

void ChordFingerTable::setFinger(uint pos, const Successors& nodes)
{
    if (pos < fingerTable.size()) {
        fingerTable[pos] = FingerEntry(nodes.begin()->second, nodes);
    } else {
        error("Index out of bound (ChordFingerTable, setFinger())");
    }
}

bool ChordFingerTable::updateFinger(uint pos, const NodeHandle& node, double rtt)
{
    if(rtt < 0)
	return false;

    Successors::iterator it;
    for(it = fingerTable[pos].second.begin(); it != fingerTable[pos].second.end(); it++)
	if(it->second == node)
	    break;
    if(it == fingerTable[pos].second.end())
	return false;

    fingerTable[pos].second.erase(it);
    fingerTable[pos].second.insert(std::make_pair(rtt, node));

    return true;
}
  
const NodeHandle& ChordFingerTable::getFinger(uint pos)
{
    if (pos >= fingerTable.size())
        error("Index out of bound (ChordFingerTable, getFinger())");

    return fingerTable[pos].first;
}

NodeVector* ChordFingerTable::getFinger(uint pos, const OverlayKey& key)
{
    if (pos >= fingerTable.size())
        error("Index out of bound (ChordFingerTable, getFinger())");

    NodeVector* nextHop = new NodeVector();

    for(Successors::const_iterator it = fingerTable[pos].second.begin();
	it != fingerTable[pos].second.end(); it++) {
	if(!key.isBetweenR(fingerTable[pos].first.key, it->second.key)) {
	    nextHop->push_back(it->second);
	}
    }
    if(nextHop->size() == 0)
	nextHop->push_back(fingerTable[pos].first);
    return nextHop;
}

std::ostream& operator<<(std::ostream& os, const Successors& suc)
{
    for (Successors::const_iterator i = suc.begin(); i != suc.end(); i++) {
	if(i != suc.begin())
	    os << endl;	
	os << i->second;
	if(i->first == -1)
	    continue;
	else if(i->first == DBL_MAX) 
	    os << "; RTT:  --- ";
	else
	    os << "; RTT: " << i->first;
    }

    return os;
}

std::ostream& operator<<(std::ostream& os, const FingerEntry& entry)
{
    if(entry.second.size() > 0)
	os << "[ " << entry.first << " ]\n" << entry.second;
    else 
	os << entry.first;

    return os;
}
