//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Pastry.h
 * @author Felix Palmen
 */

#ifndef __PASTRY_H_
#define __PASTRY_H_

#include <vector>
#include <map>
#include <algorithm>

#include <omnetpp.h>
#include <IPvXAddress.h>

#include <OverlayKey.h>
#include <NodeHandle.h>
#include <BaseOverlay.h>

#include "PastryTypes.h"
#include "PastryMessage_m.h"
#include "PastryRoutingTable.h"
#include "PastryLeafSet.h"
#include "PastryNeighborhoodSet.h"

/**
 * protocol states
 */
enum
{
    INIT,
    JOIN,
    READY,
};

/**
 * Pastry overlay module
 *
 * @author Felix Palmen
 * @see BaseOverlay
 */
class Pastry : public BaseOverlay
{
  public:
    virtual void initializeOverlay(int stage);

    /**
     * handles self-messages
     *
     * @param msg the self-message
     */
    virtual void handleTimerEvent(cMessage* msg);

    /**
     * processes messages from underlay
     *
     * @param msg message from UDP
     */
    virtual void handleUDPMessage(BaseOverlayMessage* msg);

    /**
     * process RPC responses
     *
     * @param msg the RPC response message
     * @param rpcId the RPC identifier
     * @param rtt round-trip time
     */
    virtual void handleRpcResponse(BaseResponseMessage* msg, int rpcId,
	    simtime_t rtt);

    /**
     * handle RPC timeouts
     *
     * @param msg the RPC call that timed out
     * @param dest the destination node of the RPC call
     * @param rpcID the RPC identifier
     */
    virtual void handleRpcTimeout(BaseCallMessage* msg,
	    const TransportAddress& dest, int rpcID, const OverlayKey& destKey);

    /**
     * processes state messages, merging with own state tables
     *
     * @param msg the pastry state message
     */
    void handleStateMessage(PastryStateMessage* msg);

    /**
     * processes messages from application
     *
     * @param msg message from application
     */
    virtual void handleAppMessage(BaseOverlayMessage* msg);

    /**
     * collects statisticts
     */
    virtual void finishOverlay();

    /**
     * callback-method for events at the NotificationBoard
     *
     * @param category ... TODO ...
     * @param details ... TODO ...
     */
    virtual void receiveChangeNotification(int category,
                                           cPolymorphic* details);
    /**
     * updates information shown in tk-environment
     */
    virtual void updateTooltip();

    /**
     * implement the findNode call
     *
     * @param key the destination key
     * @param msg the message to route
     * @return NodeVector containing the next hop
     */
    virtual NodeVector* findNode(const OverlayKey& key,
				 int numRedundantNodes,
				 int numSiblings,
				 BaseOverlayMessage* msg);

    /**
     * Query the maximum number of siblings (nodes close to a key)
     * that are maintained by this overlay protocol.
     *
     * @return int number of siblings.
     */
    int getMaxNumSiblings();

    /**
     * Query the maximum number of redundant next hop nodes that
     * are returned by findNode().
     *
     * @return int number of redundant nodes returned by findNode().
     */
    int getMaxNumRedundantNodes();


  protected:

    /**
     * changes node state
     *
     * @param toState state to change to
     */
    virtual void changeState(int toState);

    /**
     * send a PastryStateMessage directly to a node
     *
     * @param destination destination node
     * @param type the type of the state message to be sent
     * @param ... additional arguments for some types: @n
     *        PASTRY_STATE_JOIN: @c int @a hops number of hops to
     *        destination node @n
     *        PASTRY_STATE_JOIN: @c bool @a last mark the state
     *        message to originate from closest node found @n
     *        PASTRY_STATE_UPDATE: @c simtime_t @a timestamp use
     *        this timestamp for the uptade message
     */
    void sendStateTables(const TransportAddress& destination,
	    int type = PASTRY_STATE_STD, ...);

    /**
     * send a standard state message with a small delay
     *
     * @param destination destination node
     */
    void sendStateDelayed(const TransportAddress& destination);


    /**
     * see BaseOverlay.cc
     */
    virtual bool isSiblingFor(const NodeHandle& node,
			     const OverlayKey& key,
			     int numSiblings,
			     bool* err);

    virtual AbstractLookup* createLookup(const BaseRouteMessage* msg = NULL);

    // statistics
    int joins;
    int joinTries;
    int joinPartial;
    int joinSeen;
    int joinBytesSeen;
    int joinReceived;
    int joinBytesReceived;
    int joinSent;
    int joinBytesSent;
    int stateSent;
    int stateBytesSent;
    int stateReceived;
    int stateBytesReceived;
    int repairReqSent;
    int repairReqBytesSent;
    int repairReqReceived;
    int repairReqBytesReceived;
    int stateReqSent;
    int stateReqBytesSent;
    int stateReqReceived;
    int stateReqBytesReceived;
    int totalLookups;
    int responsibleLookups;
    int routingTableLookups;
    int closerNodeLookups;
    int closerNodeLookupsFromNeighborhood;

  private:

    PastryRoutingTable* routingTable;
    PastryLeafSet* leafSet;
    PastryNeighborhoodSet* neighborhoodSet;

    uint state;
    uint bitsPerDigit;
    uint numberOfLeaves;
    uint numberOfNeighbors;
    uint pingRetries;
    double pingTimeout;
    double readyWaitAmount;
    double joinTimeoutAmount;
    double secondStageWaitAmount;
    double pingCacheExpireTime;
    double repairTimeout;
    double ringCheckInterval;
    double sendStateWaitAmount;
    bool enableNewLeafs;
    bool optimizeLookup;
    bool optimisticForward;
    bool avoidDuplicates;
    bool partialJoinPath;

    NodeHandle bootstrapNode;
    std::vector<PastryStateMessage*> stReceived;
    std::vector<TransportAddress> notifyList;
    PastryStateMessage* stateCache;
    cQueue stateCacheQueue;
    PastryStateMsgProximity proxCache;
    PastryStateMsgProximity aliveTable;
    std::vector<PastryStateMsgProximity> stateReceivedProximities;
    PastryPingCache pingCache;
    std::map<TransportAddress, BaseRouteMessage*> recFwdQueue;

    uint joinHopCount;
    cMessage* joinTimeout;
    cMessage* readyWait;
    cMessage* secondStageWait;
    cMessage* joinUpdateWait;
    cMessage* ringCheck;
    std::vector<PastrySendState*> sendStateWait;

    simtime_t lastStateChange;

    virtual void forwardMessageRecursive(const TransportAddress& dest,
	    BaseRouteMessage* msg);

    /**
     * send updated state to all nodes when entering ready state
     */
    void doJoinUpdate(void);

    /**
     * do the second stage of initialization as described in the paper
     */
    void doSecondStage(void);

    /**
     * delete all information/messages caching vectors, used for restarting
     * overlay or finish()
     */
    void purgeVectors(void);

    /**
     * ping all nodes in a given state message. this is called when a state
     * message arrives while another one is still being processed.
     */
    void prePing(const PastryStateMessage* stateMsg);

    /**
     * ping all nodes in the pastry state message pointed to by
     * private member stateCache
     */
    void pingNodes(void);

    /**
     * change the aliveTable to match the given stateMsg.
     *
     * each node that's knowm to be dead from our pingCache gets a
     * value of PASTRY_PROX_INFINITE, all other nodes just get a
     * value of 1
     */
    void determineAliveTable(const PastryStateMessage* stateMsg);

    /**
     * checks whether proxCache is complete, takes appropriate actions
     * depending on the protocol state
     */
    void checkProxCache(void);

    /**
     * send a repair request to a given node
     *
     * @param ask request repair from this node
     */
    void sendRepairRequest(const TransportAddress& ask);

    /**
     * notifies leafset and routingtable of a failed node and sends out
     * a repair request if possible
     *
     * @param failed the failed node
     * @return true as long as local state is READY (signals lookup to try
     * again)
     */
    bool handleFailedNode(const TransportAddress& failed);

    /**
     * @see BaseOverlay.cc
     */
    virtual void joinOverlay();

    /**
     * Pastry API: send newLeafs() to application if enabled
     */
    void newLeafs(void);
};

/**
 * predicate for comparing two pointers to PastryStateMessages based on their
 * joinHopCount. Needed for sorting the received PastryStateMessages.
 */
bool stateMsgIsSmaller(const PastryStateMessage* msg1,
	const PastryStateMessage* msg2);

std::ostream& operator<<(std::ostream& os, const PastryStateMsgProximity pr);

#endif
