//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file PastryNeighborhoodSet.cc
 * @author Felix Palmen
 */

#include "PastryNeighborhoodSet.h"
#include "PastryTypes.h"

Define_Module(PastryNeighborhoodSet);

void PastryNeighborhoodSet::earlyInit(void)
{
    WATCH_VECTOR(neighbors);
}

void PastryNeighborhoodSet::initializeSet(uint numberOfNeighbors,
	const NodeHandle& owner)
{
    this->owner = owner;
    this->numberOfNeighbors = numberOfNeighbors;
    
    if (!neighbors.empty()) neighbors.clear();

    // fill Set with unspecified node handles
    for (uint i = numberOfNeighbors; i>0; i--)
	neighbors.push_back(unspecNode);
}

void PastryNeighborhoodSet::dumpToStateMessage(PastryStateMessage* msg) const
{
    uint i = 0;
    std::vector<PastryExtendedNode>::const_iterator it;

    msg->setNeighborhoodSetArraySize(numberOfNeighbors);
    for (it = neighbors.begin(); it != neighbors.end(); it++)
	msg->setNeighborhoodSet(i++, it->node);
}

const NodeHandle& PastryNeighborhoodSet::findCloserNode(
	const OverlayKey& destination, bool optimize)
{
    std::vector<PastryExtendedNode>::const_iterator it;

    if (optimize)
    {
        // pointer to later return value, initialize to unspecified, so
        // the specialCloserCondition() check will be done against our own
	// node as long as no node closer to the destination than our own was
	// found.
	const NodeHandle* ret = &NodeHandle::UNSPECIFIED_NODE;

	for (it = neighbors.begin(); it != neighbors.end(); it++)
	{
	    if (it->node.isUnspecified()) break;
	    if (specialCloserCondition(it->node, destination, *ret))
		ret = &(it->node);
	}
        return *ret;
    }
    else
    {
	for (it = neighbors.begin(); it != neighbors.end(); it++)
	{
	    if (it->node.isUnspecified()) break;
	    if (specialCloserCondition(it->node, destination)) return it->node;
	}
        return NodeHandle::UNSPECIFIED_NODE;
    }
}

bool PastryNeighborhoodSet::mergeNode(const NodeHandle& node, simtime_t prox)
{
    std::vector<PastryExtendedNode>::iterator it;

    for (it = neighbors.begin(); it != neighbors.end(); it++)
    {
	if ((! it->node.isUnspecified()) && (it->node == node)) break;
	if ((it->rtt < 0) || (it->rtt > prox))
	{
	    neighbors.insert(it, PastryExtendedNode(node, prox));
	    neighbors.pop_back();
	    return true;
	}
    }
    return false;
}

void PastryNeighborhoodSet::dumpToVector(
	std::vector<TransportAddress>& affected) const
{
    std::vector<PastryExtendedNode>::const_iterator it;

    for (it = neighbors.begin(); it != neighbors.end(); it++)
	if (! it->node.isUnspecified())
	    affected.push_back(it->node);
}

const TransportAddress& PastryNeighborhoodSet::failedNode(
	const TransportAddress& failed)
{
    std::vector<PastryExtendedNode>::iterator it;

    for (it = neighbors.begin(); it != neighbors.end(); it++)
    {
	if (it->node.isUnspecified()) break;
	if (it->node.ip == failed.ip)
	{
	    neighbors.erase(it);
	    neighbors.push_back(unspecNode);
	    break;
	}
    }

    // never ask for repair
    return TransportAddress::UNSPECIFIED_NODE;
}

std::ostream& operator<<(std::ostream& os, const PastryExtendedNode& n)
{
    os << n.node << "; Ping: " << n.rtt;
    return os;
}

