//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file NeighborsList.cc
 * @author Helge Backhaus
 */

#include "NeighborsList.h"

Define_Module(NeighborsList);

void NeighborsList::initialize(int stage)
{
    // because of IPAddressResolver, we need to wait until interfaces are registered,
    // address auto-assignment takes place etc.
    if(stage != MIN_STAGE_OVERLAY) return;

    WATCH(thisSite);
    WATCH_MAP(Sites);
    WATCH_SET(Positions);
}

void NeighborsList::handleMessage(cMessage* msg)
{
    error("this module doesn't handle messages, it runs only in initialize()");
}

void  NeighborsList::initializeList(Point p, NodeHandle owner, double AOI_size)
{
    Enter_Method_Silent();
    // Initializes the neigbors list.
    thisSite.coord = p;
    thisSite.type = THIS;
    thisSite.addr = owner;

    //Sites.insert(std::make_pair(temp_site.addr, temp_site));
    //Positions.insert(temp_site.coord);
    this->AOI_size = AOI_size;
    updateDisplayString();
}

void NeighborsList::setDebug(bool debugOutput)
{
    Enter_Method_Silent();
    this->debugOutput = debugOutput;
    geom.setDebug(debugOutput);
}

void NeighborsList::addNode(Point p, NodeHandle node)
{
    Enter_Method_Silent();
    if(node != thisSite.addr) {
        if(Sites.find(node) == Sites.end()) {
            Site temp_site;
            temp_site.coord = p;
            temp_site.addr = node;

            Sites.insert(std::make_pair(temp_site.addr, temp_site));
            Positions.insert(temp_site.coord);

            updateDisplayString();
        }
        else {
            SiteMap::iterator itSites = Sites.find(node);
            Positions.erase(itSites->second.coord);
            itSites->second.coord = p;
            Positions.insert(itSites->second.coord);
        }
    }
}

void NeighborsList::updateTimeStamp(NodeHandle node)
{
    Enter_Method_Silent();
    SiteMap::iterator itSites = Sites.find(node);
    if(itSites != Sites.end()) itSites->second.tstamp = simulation.simTime();
}

void NeighborsList::removeNode(NodeHandle node)
{
    Enter_Method_Silent();
    SiteMap::iterator itSites = Sites.find(node);
    if(itSites != Sites.end()) {
        Positions.erase(itSites->second.coord);
        Sites.erase(itSites);
        updateDisplayString();
    }
}

void NeighborsList::changePosition(Point p)
{
    Enter_Method_Silent();
    thisSite.coord = p;
}

bool NeighborsList::validatePosition(Point p)
{
    Enter_Method_Silent();
    return Positions.find(p) == Positions.end();
}

void NeighborsList::buildVoronoi(Point old_pos, Point new_pos)
{
    Enter_Method_Silent();
    int sqrt_nsites = 1;
    double xmin, xmax, ymin, ymax;
    double deltax, deltay;

    SiteMap::iterator itTemp;
    std::map<Point, Site*> sortedSites;
    std::map<Point, Site*>::iterator itSortedSites;

    // no use building with less than 2 sites
    if(Sites.size() < 2) return;

    xmin = xmax = thisSite.coord.x;
    ymin = ymax = thisSite.coord.y;
    sortedSites.insert(std::make_pair(thisSite.coord, &thisSite));

    for(itTemp = Sites.begin(); itTemp != Sites.end(); ++itTemp) {
        // determine min/max site coordinates
        if(itTemp->second.coord.x < xmin) xmin = itTemp->second.coord.x;
        if(itTemp->second.coord.x > xmax) xmax = itTemp->second.coord.x;
        if(itTemp->second.coord.y < ymin) ymin = itTemp->second.coord.y;
        if(itTemp->second.coord.y > ymax) ymax = itTemp->second.coord.y;
        // reset all sites to UNDEF
        itTemp->second.type = UNDEF;
        // fill sorted List
        sortedSites.insert(std::make_pair(itTemp->second.coord, &itTemp->second));
        sqrt_nsites++;
    }

    // needed to determine appropriate hashtable size
    deltax = xmax - xmin;
    deltay = ymax - ymin;
    sqrt_nsites = (int)sqrt((double)(sqrt_nsites+4));

    // start to calculate the voronoi
    Site *newsite, *bot, *top, *temp, *p, *v, *bottomsite;
    Point newintstar;
    int pm;
    Halfedge *lbnd, *rbnd, *llbnd, *rrbnd, *bisector;
    Edge *e;

    newintstar.x = newintstar.y = 0.0;

    itSortedSites = sortedSites.begin();

    geom.initialize(deltax, deltay, thisSite.coord, old_pos, new_pos, AOI_size);
    heap.PQinitialize(sqrt_nsites, ymin, deltay);
    bottomsite = itSortedSites->second;
    ++itSortedSites;
    edgelist.initialize(sqrt_nsites, xmin, deltax, bottomsite);

    newsite = itSortedSites->second;
    ++itSortedSites;
    while(1) {
        if(!heap.PQempty()) newintstar = heap.PQ_min();

        if(newsite != NULL && (heap.PQempty() ||
           newsite->coord.y < newintstar.y ||
           (newsite->coord.y == newintstar.y && newsite->coord.x < newintstar.x))) {
            lbnd = edgelist.ELleftbnd(&(newsite->coord));
            rbnd = edgelist.ELright(lbnd);
            bot = edgelist.rightreg(lbnd);
            e = geom.bisect(bot, newsite);
            bisector = edgelist.HEcreate(e, le);
            edgelist.ELinsert(lbnd, bisector);
            if ((p = geom.intersect(lbnd, bisector)) != NULL) {
                heap.PQdelete(lbnd);
                heap.PQinsert(lbnd, p, geom.dist(p, newsite));
            }
            lbnd = bisector;
            bisector = edgelist.HEcreate(e, re);
            edgelist.ELinsert(lbnd, bisector);
            if ((p = geom.intersect(bisector, rbnd)) != NULL) heap.PQinsert(bisector, p, geom.dist(p, newsite));
            if(itSortedSites != sortedSites.end()) {
                newsite = itSortedSites->second;
                ++itSortedSites;
            }
            else newsite = NULL;
        }
        else if (!heap.PQempty()) {
            lbnd = heap.PQextractmin();
            llbnd = edgelist.ELleft(lbnd);
            rbnd = edgelist.ELright(lbnd);
            rrbnd = edgelist.ELright(rbnd);
            bot = edgelist.leftreg(lbnd);
            top = edgelist.rightreg(rbnd);
            v = lbnd->vertex;
            geom.endpoint(lbnd->ELedge, lbnd->ELpm, v);
            geom.endpoint(rbnd->ELedge, rbnd->ELpm, v);
            edgelist.ELdelete(lbnd);
            heap.PQdelete(rbnd);
            edgelist.ELdelete(rbnd);
            pm = le;
            if (bot->coord.y > top->coord.y) {
                temp = bot;
                bot = top;
                top = temp;
                pm = re;
            }
            e = geom.bisect(bot, top);
            bisector = edgelist.HEcreate(e, pm);
            edgelist.ELinsert(llbnd, bisector);
            geom.endpoint(e, re-pm, v);
            if((p = geom.intersect(llbnd, bisector)) != NULL) {
                heap.PQdelete(llbnd);
                heap.PQinsert(llbnd, p, geom.dist(p, bot));
            }
            if ((p = geom.intersect(bisector, rrbnd)) != NULL) heap.PQinsert(bisector, p, geom.dist(p, bot));
        }
        else break;
    }

    // process the generated edgelist
    for(lbnd = edgelist.ELright(edgelist.ELleftend); lbnd != edgelist.ELrightend; lbnd = edgelist.ELright(lbnd)) {
        e = lbnd -> ELedge;
        geom.processEdge(e);
    }
    // process sites in order to determine our neighbors
    for(itTemp = Sites.begin(); itTemp != Sites.end(); ++itTemp) {
        if(itTemp->second.innerEdge[0]) {
            if(itTemp->second.outerEdge) {
                itTemp->second.type |= BOUNDARY;
                // Debug output
                if(debugOutput)
                    ev << "  VAST: Site at [" << itTemp->second.coord.x << ", "
                       << itTemp->second.coord.y << "] is a boundary neighbor." << endl;
            }
            else {
                itTemp->second.type |= NEIGHBOR;
                // Debug output
                if(debugOutput)
                    ev << "  VAST: Site at [" << itTemp->second.coord.x << ", "
                       << itTemp->second.coord.y << "] is a neighbor." << endl;
            }
        }
        if(!itTemp->second.innerEdge[1] && itTemp->second.innerEdge[2]) {
            itTemp->second.type |= NEW;
            // Debug output
            if(debugOutput)
                ev << "  VAST: Site at [" << itTemp->second.coord.x << ", "
                   << itTemp->second.coord.y << "] is a new neighbor for site at " << new_pos.x << ":" << new_pos.y << "." << endl;
        }
        // reset inner- and outeredge indicator
        itTemp->second.innerEdge[0] = false;
        itTemp->second.innerEdge[1] = false;
        itTemp->second.innerEdge[2] = false;
        itTemp->second.outerEdge = false;
    }
    // clean up
    edgelist.reset();
    heap.PQreset();
    geom.reset();
}

void NeighborsList::buildVoronoi()
{
    buildVoronoi(thisSite.coord, thisSite.coord);
}

void NeighborsList::removeNeighbors()
{
    Enter_Method_Silent();
    SiteMap::iterator itSites;
    for(itSites = Sites.begin(); itSites != Sites.end();) {
        // if current site is no neighbor remove it else go on to next site
        if(itSites->second.type == UNDEF) {
            // Debug output
            if(debugOutput) ev << "  VAST: Site at [" << itSites->second.coord.x << ", " << itSites->second.coord.y << "] has been removed from list." << endl;
            Positions.erase(itSites->second.coord);
            Sites.erase(itSites++);
        }
        else ++itSites;
    }
    updateDisplayString();
}

void NeighborsList::updateDisplayString()
{
    // Displays the current number of neighbors in the list.
    if(ev.isGUI()) {
        std::stringstream cur;
        if (Sites.size() == 1) {
            cur << "1 neighbor";
        }
        else {
            cur << Sites.size() << " neigbors";
        }
        displayString().setTagArg("t", 0, cur.str().c_str());
        displayString().setTagArg("t", 2, "blue");
    }
}

int NeighborsList::getSize()
{
    Enter_Method_Silent();
    return Sites.size();
}

void NeighborsList::clearList()
{
    Enter_Method_Silent();
    if(Sites.size()) {
        Sites.clear();
        Positions.clear();
    }
}

NeighborsList::~NeighborsList()
{
    clearList();
}
