//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Vast.cc
 * @author Helge Backhaus
 */

#include "Vast.h"

Define_Module(Vast);

void Vast::initializeOverlay(int stage)
{
    // because of IPAddressResolver, we need to wait until interfaces are registered,
    // address auto-assignment takes place etc.
    if(stage != MIN_STAGE_OVERLAY) return;

    // fetch parameters
    debugVoronoiOutput = par("debugVastOutput");
    AOI_size = par("AOI_size");
    joinTimeout = par("joinTimeout");
    pingTimeout = par("pingTimeout");
    onlyCommonAPIMessages = false;

    // find friend modules
    neighborsList = check_and_cast<NeighborsList*>(parentModule()->submodule("neighborsList"));

    // set node key
    thisNode.key = OverlayKey::random();

    // initialize neighborsList
    neighborsList->setDebug(debugVoronoiOutput);

    // self-messages
    join_timer = new cMessage("join_timer");
    ping_timer = new cMessage("ping_timer");
    sec_timer = new cMessage("sec_timer");

    // statistics
    joinRequestBytesSent = 0;
    joinAcknowledgeBytesSent = 0;
    nodeMoveBytesSent = 0;
    newNeighborsBytesSent = 0;
    nodeLeaveBytesSent = 0;
    enclosingNeighborsRequestBytesSent = 0;
    pingBytesSent = 0;
    pongBytesSent = 0;
    discardNodeBytesSent = 0;

    maxBytesPerSecondSent = 0;
    averageBytesPerSecondSent = 0;
    bytesPerSecond = 0;

    // watch some variables
    WATCH(position);
    WATCH(thisNode);
    WATCH(AOI_size);

    WATCH(joinRequestBytesSent);
    WATCH(joinAcknowledgeBytesSent);
    WATCH(nodeMoveBytesSent);
    WATCH(newNeighborsBytesSent);
    WATCH(nodeLeaveBytesSent);
    WATCH(enclosingNeighborsRequestBytesSent);
    WATCH(pingBytesSent);
    WATCH(pongBytesSent);
    WATCH(discardNodeBytesSent);

    WATCH(maxBytesPerSecondSent);
    WATCH(bytesPerSecond);

    // set initial state
    changeState(INIT);
    changeState(JOINING);
}

void Vast::changeState(int state)
{
    switch(state) {
        case INIT: {
            this->state = INIT;
            bootstrapOracle->removePeer(thisNode);
        } break;
        case JOINING: {
            this->state = JOINING;
            scheduleAt(simulation.simTime(), join_timer);
            scheduleAt(simulation.simTime() + 1.0, sec_timer);
        } break;
        case READY: {
            this->state = READY;
            cancelEvent(join_timer);
            neighborsList->initializeList(position, thisNode, AOI_size);
            bootstrapOracle->registerPeer(thisNode);
            scheduleAt(simulation.simTime() + pingTimeout, ping_timer);
            // tell the application we are ready
            SimpleClientMessage *scMsg = new SimpleClientMessage("OVERLAY_READY");
            scMsg->setCommand(OVERLAY_READY);
            sendToApp(scMsg);
        } break;
    }
    setBootstrapedIcon();
    // debug output
    if(debugOutput) {
        ev << "VAST: Node " << thisNode.ip << " entered ";
        switch(state) {
            case INIT: ev << "INIT"; break;
            case JOINING: ev << "JOINING"; break;
            case READY: ev << "READY"; break;
        }
        ev << " state." << endl;
    }
}

void Vast::handleTimerEvent(cMessage* msg)
{
    if(msg->isName("join_timer")) {
        //reset timer
        cancelEvent(join_timer);
        scheduleAt(simulation.simTime() + joinTimeout, msg);
        // handle event
        processJoinTimer();
    }
    else if(msg->isName("ping_timer")) {
        //reset timer
        cancelEvent(ping_timer);
        scheduleAt(simulation.simTime() + pingTimeout, msg);
        // handle event
        processPingTimer();
    }
    else if(msg->isName("sec_timer")) {
        //reset timer
        cancelEvent(sec_timer);
        scheduleAt(simulation.simTime() + 1, msg);
        // handle event
        processSecTimer();
    }
}

void Vast::handleAppMessage(cMessage* msg)
{
    if(dynamic_cast<SimpleClientMessage*>(msg)) {
        SimpleClientMessage* simpleClientMsg = check_and_cast<SimpleClientMessage*>(msg);
        // debug output
        if(debugOutput) ev << "VAST: Node " << thisNode.ip << " received " << simpleClientMsg->name() << " from application." << endl;
        switch(simpleClientMsg->getCommand()) {
            case MOVE_INDICATION: {
                SimpleClientMoveMessage* simpleClientMoveMsg = check_and_cast<SimpleClientMoveMessage*>(msg);
                if(state == JOINING) {
                    handleJoin(simpleClientMoveMsg);
                    delete msg;
                }
                else if(state == READY) {
                    handleMove(simpleClientMoveMsg);
                    delete msg;
                }
            } break;
            //case EVENT: {
                // insert handling for additional application messages like events etc. here
            //} break;
            default: {
                delete msg;
            }
        }
    }
    else delete msg;
}

void Vast::handleUDPMessage(BaseOverlayMessage* msg)
{
    if(state == INIT) {
        delete msg;
        return;
    }
    if(dynamic_cast<VastMessage*>(msg)) {
        VastMessage* vastMsg = check_and_cast<VastMessage*>(msg);
        if(vastMsg->getDestKey() == thisNode.key) {
            // debug output
            if(debugOutput) ev << "VAST: Node " << thisNode.ip << " received " << vastMsg->name() << " from " << vastMsg->getSourceNode().ip << "." << endl;
            if(state == READY) {
                switch(vastMsg->getCommand()) {
                    case JOIN_REQUEST: {
                        handleJoinRequest(vastMsg);
                    } break;
                    case NODE_MOVE: {
                        VastMoveMessage* vastMoveMsg = check_and_cast<VastMoveMessage*>(msg);
                        handleNodeMove(vastMoveMsg);
                    } break;
                    case NEW_NEIGHBORS: {
                        VastListMessage* vastListMsg = check_and_cast<VastListMessage*>(msg);
                        handleNewNeighbors(vastListMsg);
                    } break;
                    case NODE_LEAVE: {
                        VastListMessage* vastListMsg = check_and_cast<VastListMessage*>(msg);
                        handleNodeLeave(vastListMsg);
                    } break;
                    case ENCLOSING_NEIGHBORS_REQUEST: {
                        handleEnclosingNeighborsRequest(vastMsg);
                    } break;
                    case PING: {
                        handlePing(vastMsg);
                    } break;
                    case PONG: {
                        handlePong(vastMsg);
                    } break;
                    case DISCARD_NODE: {
                        VastDiscardMessage* vastDiscardMsg = check_and_cast<VastDiscardMessage*>(msg);
                        handleDiscardNode(vastDiscardMsg);
                    } break;
                }
                // update timestamp
                if(vastMsg != NULL && vastMsg->getCommand() != JOIN_REQUEST) {
                    neighborsList->updateTimeStamp(vastMsg->getSourceNode());
                    delete msg;
                }
            }
            else if(state == JOINING && vastMsg->getCommand() == JOIN_ACKNOWLEDGE) {
                VastListMessage* vastListMsg = check_and_cast<VastListMessage*>(msg);
                handleJoinAcknowledge(vastListMsg);
                delete msg;
            }
            else delete msg;
        }
        else {
            sendDiscardNode(vastMsg);
            delete msg;
        }
    }
    else delete msg;
}

void Vast::receiveChangeNotification(int category, cPolymorphic *details)
{
    /*if(category == NF_OVERLAY_NODE_LEAVE && state == READY) {
        int i;
        // debug output
        if(debugOutput) ev << "OVERLAY: Node " << thisNode.ip << " is leaving the overlay." << endl;

        // generate node leave messages
        VoronoiListMessage *voroListMsg = new VoronoiListMessage("NODE_LEAVE");
        voroListMsg->setCommand(NODE_LEAVE);
        voroListMsg->setSourceNode(thisNode);
        voroListMsg->setPosX(position.x);
        voroListMsg->setPosY(position.y);

        voroListMsg->setNeighborNodeArraySize(neighborsList->getSize());
        voroListMsg->setNeighbor_posXArraySize(neighborsList->getSize());
        voroListMsg->setNeighbor_posYArraySize(neighborsList->getSize());

        i = 0;
        neighborsList->initIterator();
        Site *currentSite = neighborsList->next();
        while(currentSite) {
            if(currentSite->type & ENCLOSING) {
                voroListMsg->setNeighborNode(i, currentSite->addr);
                voroListMsg->setNeighbor_posX(i, currentSite->coord.x);
                voroListMsg->setNeighbor_posY(i, currentSite->coord.y);
                ++i;
            }
            currentSite = neighborsList->next();
        }
        voroListMsg->setEntry_count(i);
        voroListMsg->setNeighborNodeArraySize(i);
        voroListMsg->setNeighbor_posXArraySize(i);
        voroListMsg->setNeighbor_posYArraySize(i);

        voroListMsg->setLength(BASE_L + COUNT_L + voroListMsg->getEntry_count() * ENTRY_L);
        if(voroListMsg->getEntry_count() > 0) {
            neighborsList->initIterator();
            currentSite = neighborsList->next();
            while(currentSite) {
                if(!(currentSite->type & THIS)) {
                    VoronoiListMessage *voroCopyMsg = new VoronoiListMessage(*voroListMsg);
                    // prevent notification board from taking the message
                    take(voroCopyMsg);
                    sendMessage(voroCopyMsg, currentSite->addr);
                }
                currentSite = neighborsList->next();
            }
        }
        delete voroListMsg;

        changeState(INIT);
    }*/
}

void Vast::processJoinTimer()
{
    SimpleClientMessage *scMsg = new SimpleClientMessage("MOVE_REQUEST");
    scMsg->setCommand(MOVE_REQUEST);
    sendToApp(scMsg);
}

void Vast::processPingTimer()
{
    bool abnormalLeave = false;
    bool boundaryLeave = false;
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        if(itSites->second.tstamp < 0.0) { // node is dropped cause no pong has been received see below
            abnormalLeave = true;
            if(!(itSites->second.type & NEIGHBOR)) boundaryLeave = true;
            itSites->second.type = UNDEF;
        }
        else if(itSites->second.tstamp < simulation.simTime() - pingTimeout) { // node showed no activity for some time request pong and mark it to be dropped next time
            VastMessage *vastMsg = new VastMessage("PING");
            vastMsg->setCommand(PING);
            vastMsg->setSourceNode(thisNode);
            vastMsg->setPosX(position.x);
            vastMsg->setPosY(position.y);
            vastMsg->setLength(VAST_L(vastMsg));
            sendMessage(vastMsg, itSites->second.addr);
            itSites->second.tstamp = -1.0;
        }
        ++itSites;
    }
    if(abnormalLeave) {
        synchronizeApp();
        neighborsList->removeNeighbors();
        if(boundaryLeave) {
            itSites = neighborsList->Sites.begin();
            while(itSites != neighborsList->Sites.end()) {
                if(itSites->second.type & BOUNDARY) {
                    VastMessage *vastMsg = new VastMessage("ENCLOSING_NEIGHBORS_REQUEST");
                    vastMsg->setCommand(ENCLOSING_NEIGHBORS_REQUEST);
                    vastMsg->setSourceNode(thisNode);
                    vastMsg->setPosX(position.x);
                    vastMsg->setPosY(position.y);
                    vastMsg->setLength(VAST_L(vastMsg));
                    sendMessage(vastMsg, itSites->second.addr);
                }
                ++itSites;
            }
        }
        neighborsList->buildVoronoi();
        //neighborsList->removeNeighbors(); should be superfluous
    }
}

void Vast::processSecTimer()
{
    if(bytesPerSecond > maxBytesPerSecondSent) {
        maxBytesPerSecondSent = bytesPerSecond;
    }
    averageBytesPerSecondSent += bytesPerSecond;
    bytesPerSecond = 0;
}

void Vast::handleJoin(SimpleClientMoveMessage *scMsg)
{
    NodeHandle joinNode = bootstrapOracle->getBootstrapNode();
    position.x = scMsg->getPosX();
    position.y = scMsg->getPosY();
    // check if this is the only node in the overlay
    if(joinNode.isUnspecified()) {
        changeState(READY);
    }
    else {
        VastMessage *vastMsg = new VastMessage("JOIN_REQUEST");
        vastMsg->setCommand(JOIN_REQUEST);
        vastMsg->setSourceNode(thisNode);
        vastMsg->setPosX(position.x);
        vastMsg->setPosY(position.y);
        vastMsg->setLength(VAST_L(vastMsg));
        sendMessage(vastMsg, joinNode);
    }
}

void Vast::handleMove(SimpleClientMoveMessage* scMsg)
{
    Point pos;
    pos.x = scMsg->getPosX();
    pos.y = scMsg->getPosY();
    // test if new position is legal
    if(!neighborsList->validatePosition(pos)) {
        SimpleClientMessage *scMsg = new SimpleClientMessage("MOVE_REQUEST");
        scMsg->setCommand(MOVE_REQUEST);
        sendToApp(scMsg);
        return;
    }
    // send position update to neighbors
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        VastMoveMessage *vastMoveMsg = new VastMoveMessage("NODE_MOVE");
        vastMoveMsg->setCommand(NODE_MOVE);
        vastMoveMsg->setSourceNode(thisNode);
        vastMoveMsg->setPosX(position.x);
        vastMoveMsg->setPosY(position.y);
        vastMoveMsg->setNew_posX(pos.x);
        vastMoveMsg->setNew_posY(pos.y);
        vastMoveMsg->setIs_boundary(itSites->second.type & BOUNDARY);
        vastMoveMsg->setLength(VASTMOVE_L(vastMoveMsg));
        sendMessage(vastMoveMsg, itSites->second.addr);
        ++itSites;
    }
    // set new position
    position = pos;
    neighborsList->changePosition(pos);
    // update voronoi
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
}

void Vast::handleJoinRequest(VastMessage *vastMsg)
{
    Site *forwardSite = NULL;
    double min_dist;
    // start with this node
    min_dist = distance_square(neighborsList->thisSite, vastMsg);
    forwardSite = &neighborsList->thisSite;
    // iterate through all neighbors
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        if(distance_square(itSites->second, vastMsg) < min_dist) {
            min_dist = distance_square(itSites->second, vastMsg);
            forwardSite = &itSites->second;
        }
        ++itSites;
    }
    // do nothing and let node retry with new position if current position is illegal
    if(min_dist == 0.0) {
        delete vastMsg;
        vastMsg = NULL;
    }
    else {
        // send an acknowledge or forward request if any of our neighbors is closer to joining node
        if(forwardSite->type & THIS) {
            VastListMessage *vastListMsg = new VastListMessage("JOIN_ACKNOWLEDGE");
            vastListMsg->setCommand(JOIN_ACKNOWLEDGE);
            vastListMsg->setSourceNode(thisNode);
            vastListMsg->setPosX(position.x);
            vastListMsg->setPosY(position.y);
            // fill neighbors list
            vastListMsg->setNeighborNodeArraySize(neighborsList->getSize());
            vastListMsg->setNeighbor_posXArraySize(neighborsList->getSize());
            vastListMsg->setNeighbor_posYArraySize(neighborsList->getSize());
            int i = 0;
            itSites = neighborsList->Sites.begin();
            while(itSites != neighborsList->Sites.end()) {
                vastListMsg->setNeighborNode(i, itSites->second.addr);
                vastListMsg->setNeighbor_posX(i, itSites->second.coord.x);
                vastListMsg->setNeighbor_posY(i, itSites->second.coord.y);
                ++i;
                ++itSites;
            }
            vastListMsg->setEntry_count(i);
            vastListMsg->setNeighborNodeArraySize(i);
            vastListMsg->setNeighbor_posXArraySize(i);
            vastListMsg->setNeighbor_posYArraySize(i);

            vastListMsg->setLength(VASTLIST_L(vastListMsg));
            sendMessage(vastListMsg, vastMsg->getSourceNode());
            delete vastMsg;
            vastMsg = NULL;
        }
        else {
            sendMessage(vastMsg, forwardSite->addr);
        }
    }
}

void Vast::handleJoinAcknowledge(VastListMessage *vastListMsg)
{
    Point p, p_join;
    p_join.x = vastListMsg->getPosX();
    p_join.y = vastListMsg->getPosY();
    // add acceptor node
    neighborsList->addNode(p_join, vastListMsg->getSourceNode());
    changeState(READY);
    // add new neighbors
    for(int i=0; i<vastListMsg->getEntry_count(); i++) {
        p.x = vastListMsg->getNeighbor_posX(i);
        p.y = vastListMsg->getNeighbor_posY(i);
        neighborsList->addNode(p, vastListMsg->getNeighborNode(i));
    }
    // update voronoi with new neighbors
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
    // contact new neighbors
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        VastMoveMessage *vastMoveMsg = new VastMoveMessage("NODE_MOVE");
        vastMoveMsg->setCommand(NODE_MOVE);
        vastMoveMsg->setSourceNode(thisNode);
        vastMoveMsg->setPosX(p_join.x);
        vastMoveMsg->setPosY(p_join.y);
        vastMoveMsg->setNew_posX(position.x);
        vastMoveMsg->setNew_posY(position.y);
        vastMoveMsg->setIs_boundary(itSites->second.type & BOUNDARY);
        vastMoveMsg->setLength(VASTMOVE_L(vastMoveMsg));
        sendMessage(vastMoveMsg, itSites->second.addr);
        ++itSites;
    }
}

void Vast::handleNodeMove(VastMoveMessage *vastMoveMsg)
{
    Point old_p, new_p;
    old_p.x = vastMoveMsg->getPosX();
    old_p.y = vastMoveMsg->getPosY();
    new_p.x = vastMoveMsg->getNew_posX();
    new_p.y = vastMoveMsg->getNew_posY();
    neighborsList->addNode(new_p, vastMoveMsg->getSourceNode());
    // update voronoi with new neighbor detection or without
    if(vastMoveMsg->getIs_boundary()) {
        neighborsList->buildVoronoi(old_p, new_p);
        synchronizeApp();
        neighborsList->removeNeighbors();
        // send new neighbors
        VastListMessage *vastListMsg = new VastListMessage("NEW_NEIGHBORS");
        vastListMsg->setCommand(NEW_NEIGHBORS);
        vastListMsg->setSourceNode(thisNode);
        vastListMsg->setPosX(position.x);
        vastListMsg->setPosY(position.y);

        vastListMsg->setNeighborNodeArraySize(neighborsList->getSize());
        vastListMsg->setNeighbor_posXArraySize(neighborsList->getSize());
        vastListMsg->setNeighbor_posYArraySize(neighborsList->getSize());

        int i = 0;
        SiteMap::iterator itSites = neighborsList->Sites.begin();
        while(itSites != neighborsList->Sites.end()) {
            if(itSites->second.type & NEW) {
                vastListMsg->setNeighborNode(i, itSites->second.addr);
                vastListMsg->setNeighbor_posX(i, itSites->second.coord.x);
                vastListMsg->setNeighbor_posY(i, itSites->second.coord.y);
                ++i;
            }
            ++itSites;
        }
        vastListMsg->setEntry_count(i);
        vastListMsg->setNeighborNodeArraySize(i);
        vastListMsg->setNeighbor_posXArraySize(i);
        vastListMsg->setNeighbor_posYArraySize(i);

        vastListMsg->setLength(VASTLIST_L(vastListMsg));
        if(vastListMsg->getEntry_count() > 0) {
            sendMessage(vastListMsg, vastMoveMsg->getSourceNode());
        }
        else {
            delete vastListMsg;
        }
    }
    else {
        neighborsList->buildVoronoi();
        synchronizeApp();
        neighborsList->removeNeighbors();
    }
}

void Vast::handleNewNeighbors(VastListMessage *vastListMsg)
{
    Point p;
    // add new neighbors
    for(int i=0; i<vastListMsg->getEntry_count(); i++) {
        p.x = vastListMsg->getNeighbor_posX(i);
        p.y = vastListMsg->getNeighbor_posY(i);
        neighborsList->addNode(p, vastListMsg->getNeighborNode(i));
    }
    // update voronoi with new neighbors
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
}

void Vast::handleNodeLeave(VastListMessage *vastListMsg)
{
    Point p;
    neighborsList->removeNode(vastListMsg->getSourceNode());
    // add possible new neighbors
    for(int i=0; i<vastListMsg->getEntry_count(); i++) {
        p.x = vastListMsg->getNeighbor_posX(i);
        p.y = vastListMsg->getNeighbor_posY(i);
        neighborsList->addNode(p, vastListMsg->getNeighborNode(i));
    }
    // update voronoi with new neighbors
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
}

void Vast::handleEnclosingNeighborsRequest(VastMessage *vastMsg)
{
    // send new neighbors
    VastListMessage *vastListMsg = new VastListMessage("NEW_NEIGHBORS");
    vastListMsg->setCommand(NEW_NEIGHBORS);
    vastListMsg->setSourceNode(thisNode);
    vastListMsg->setPosX(position.x);
    vastListMsg->setPosY(position.y);

    vastListMsg->setNeighborNodeArraySize(neighborsList->getSize());
    vastListMsg->setNeighbor_posXArraySize(neighborsList->getSize());
    vastListMsg->setNeighbor_posYArraySize(neighborsList->getSize());

    int i = 0;
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        if((itSites->second.type & ENCLOSING) && itSites->second.addr != vastMsg->getSourceNode()) {
            vastListMsg->setNeighborNode(i, itSites->second.addr);
            vastListMsg->setNeighbor_posX(i, itSites->second.coord.x);
            vastListMsg->setNeighbor_posY(i, itSites->second.coord.y);
            ++i;
        }
        ++itSites;
    }
    vastListMsg->setEntry_count(i);
    vastListMsg->setNeighborNodeArraySize(i);
    vastListMsg->setNeighbor_posXArraySize(i);
    vastListMsg->setNeighbor_posYArraySize(i);

    vastListMsg->setLength(VASTLIST_L(vastListMsg));
    if(vastListMsg->getEntry_count() > 0) {
        sendMessage(vastListMsg, vastMsg->getSourceNode());
    }
    else {
        delete vastListMsg;
    }
}

void Vast::handlePing(VastMessage *vastMsg)
{
    VastMessage *vastPongMsg = new VastMessage("PONG");
    vastPongMsg->setCommand(PONG);
    vastPongMsg->setSourceNode(thisNode);
    vastPongMsg->setPosX(position.x);
    vastPongMsg->setPosY(position.y);
    vastPongMsg->setLength(VAST_L(vastPongMsg));
    sendMessage(vastPongMsg, vastMsg->getSourceNode());
}

void Vast::handlePong(VastMessage *vastMsg)
{
    Point p;
    // replace entry cause it was probably outdated
    p.x = vastMsg->getPosX();
    p.y = vastMsg->getPosY();
    neighborsList->addNode(p, vastMsg->getSourceNode());
    // update voronoi
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
}

void Vast::handleDiscardNode(VastDiscardMessage *vastMsg)
{
    // discard outdated entry
    neighborsList->removeNode(vastMsg->getDiscardNode());
    // update voronoi
    neighborsList->buildVoronoi();
    synchronizeApp();
    neighborsList->removeNeighbors();
}

void Vast::sendDiscardNode(VastMessage *vastMsg)
{
    NodeHandle discardNode;
    discardNode.ip = thisNode.ip;
    discardNode.key = vastMsg->getDestKey();
    // send message
    VastDiscardMessage *vastDiscardMsg = new VastDiscardMessage("DISCARD_NODE");
    vastDiscardMsg->setCommand(DISCARD_NODE);
    vastDiscardMsg->setSourceNode(thisNode);
    vastDiscardMsg->setPosX(position.x);
    vastDiscardMsg->setPosY(position.y);
    vastDiscardMsg->setDiscardNode(discardNode);
    // debug output
    if(debugOutput) ev << "VAST: Node " << thisNode.ip << " received message for an outdated node from " << vastMsg->getSourceNode().ip << "." << endl;
    vastDiscardMsg->setLength(VASTDISCARD_L(vastDiscardMsg));
    sendMessage(vastDiscardMsg, vastMsg->getSourceNode());
}

void Vast::synchronizeApp()
{
    SimpleClientListMessage *scMsg = new SimpleClientListMessage("UPDATE_NEIGHBORS");
    scMsg->setCommand(UPDATE_NEIGHBORS);

    scMsg->setRemove_neighborArraySize(neighborsList->getSize());
    scMsg->setAdd_neighborArraySize(neighborsList->getSize());
    scMsg->setAdd_neighbor_posXArraySize(neighborsList->getSize());
    scMsg->setAdd_neighbor_posYArraySize(neighborsList->getSize());

    int i, j;
    i = j = 0;
    SiteMap::iterator itSites = neighborsList->Sites.begin();
    while(itSites != neighborsList->Sites.end()) {
        if(itSites->second.type == UNDEF) {
            scMsg->setRemove_neighbor(i, itSites->second.addr);
            ++i;
        }
        else if(!itSites->second.isAdded) {
            scMsg->setAdd_neighbor(j, itSites->second.addr);
            scMsg->setAdd_neighbor_posX(j, itSites->second.coord.x);
            scMsg->setAdd_neighbor_posY(j, itSites->second.coord.y);
            itSites->second.isAdded = true;
            ++j;
        }
        ++itSites;
    }
    scMsg->setRemove_entry_count(i);
    scMsg->setRemove_neighborArraySize(i);
    scMsg->setAdd_entry_count(j);
    scMsg->setAdd_neighborArraySize(j);
    scMsg->setAdd_neighbor_posXArraySize(j);
    scMsg->setAdd_neighbor_posYArraySize(j);

    if(scMsg->getAdd_entry_count() || scMsg->getRemove_entry_count()) {
        sendToApp(scMsg);
    }
    else {
        delete scMsg;
    }
}

void Vast::sendToApp(SimpleClientMessage *scMsg)
{
    // debug output
    if(debugOutput) ev << "VAST: Node " << thisNode.ip << " sending " << scMsg->name() << " to application." << endl;
    send(scMsg, "to_app");
}

void Vast::sendMessage(VastMessage *vastMsg, NodeHandle destAddr)
{
    // collect statistics
    switch(vastMsg->getType()) {
        case JOIN_REQUEST: {
            joinRequestBytesSent += vastMsg->byteLength();
        } break;
        case JOIN_ACKNOWLEDGE: {
            joinAcknowledgeBytesSent += vastMsg->byteLength();
        } break;
        case NODE_MOVE: {
            nodeMoveBytesSent += vastMsg->byteLength();
        } break;
        case NEW_NEIGHBORS: {
            newNeighborsBytesSent += vastMsg->byteLength();
        } break;
        case NODE_LEAVE: {
            nodeLeaveBytesSent += vastMsg->byteLength();
        } break;
        case ENCLOSING_NEIGHBORS_REQUEST: {
            enclosingNeighborsRequestBytesSent += vastMsg->byteLength();
        } break;
        case PING: {
            pingBytesSent += vastMsg->byteLength();
        } break;
        case PONG: {
            pongBytesSent += vastMsg->byteLength();
        } break;
        case DISCARD_NODE: {
            discardNodeBytesSent += vastMsg->byteLength();
        } break;
    }
    bytesPerSecond += vastMsg->byteLength();

    // debug output
    if(debugOutput) ev << "VAST: Node " << thisNode.ip << " sending " << vastMsg->name() << " to " << destAddr.ip << "." << endl;
    vastMsg->setDestKey(destAddr.key);
    sendMessageToUDP(destAddr, vastMsg);

    //set UDP control info
    /*vastMsg->setKind(UDP_C_DATA);

    UDPControlInfo* ctrl = new UDPControlInfo();
    ctrl->setSrcAddr(thisNode.ip);
    ctrl->setSrcPort(thisNode.port);
    ctrl->setDestAddr(destAddr.ip);
    ctrl->setDestPort(thisNode.port);
    vastMsg->setControlInfo(ctrl);
    send(vastMsg, "to_udp");*/
}

double Vast::distance_square(Site s, VastMessage *msg)
{
    double dx, dy;
    dx = s.coord.x - msg->getPosX();
    dy = s.coord.y - msg->getPosY();
    return dx*dx + dy*dy;
}

void Vast::setBootstrapedIcon()
{
    if(ev.isGUI()) {
        if(state == READY) {
            parentModule()->parentModule()->displayString().setTagArg("i2", 1, "green");
            displayString().setTagArg("i", 1, "green");
        }
        else if(state == JOINING) {
            parentModule()->parentModule()->displayString().setTagArg("i2", 1, "yellow");
            displayString().setTagArg("i", 1, "yellow");
        }
        else {
            parentModule()->parentModule()->displayString().setTagArg("i2", 1, "red");
            displayString().setTagArg("i", 1, "red");
        }
    }
}

void Vast::finishOverlay()
{
    // collect statistics
	globalStatistics->addStdDev("JOIN_REQUEST bytes sent", joinRequestBytesSent);
    globalStatistics->addStdDev("JOIN_ACKNOWLEDGE bytes sent", joinAcknowledgeBytesSent);
    globalStatistics->addStdDev("NODE_MOVE bytes sent", nodeMoveBytesSent);
    globalStatistics->addStdDev("NEW_NEIGHBORS bytes sent", newNeighborsBytesSent);
    globalStatistics->addStdDev("NODE_LEAVE bytes sent", nodeLeaveBytesSent);
    globalStatistics->addStdDev("ENCLOSING_NEIGHBORS_REQUEST bytes sent", enclosingNeighborsRequestBytesSent);
    globalStatistics->addStdDev("PING bytes sent", pingBytesSent);
    globalStatistics->addStdDev("PONG bytes sent", pongBytesSent);
    globalStatistics->addStdDev("DISCARD_NODE bytes sent", discardNodeBytesSent);

    globalStatistics->addStdDev("max bytes/second sent", maxBytesPerSecondSent);
    globalStatistics->addStdDev("average bytes/second sent", (long)(averageBytesPerSecondSent / simTime()));

    bootstrapOracle->removePeer(thisNode);
}

int Vast::getState()
{
    Enter_Method_Silent();
    return state;
}

double Vast::getAOI()
{
    Enter_Method_Silent();
    return AOI_size;
}

Point Vast::getPosition()
{
    Enter_Method_Silent();
    return position;
}

NeighborsList* Vast::getList()
{
    Enter_Method_Silent();
    return neighborsList;
}

Vast::~Vast()
{
    // destroy self timer messages
    cancelAndDelete(join_timer);
    cancelAndDelete(ping_timer);
    cancelAndDelete(sec_timer);
}
