//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#ifndef __YMMOG_H_
#define __YMMOG_H_

#include <vector>
#include <list>

#include "YMMOGBase.h"
#include "ResponsibleNodeData.h"
#include "BackupNodeData.h"
#include "IntermediateNodeData.h"
#include "NodeLogEntry.h"
#include "SubSpaceTable.h"

class YMMOG : public YMMOGBase
{
	public:
		virtual int numInitStages() const {return 6;}
		// OMNeT++
		void initialize();
		virtual void finish();
		friend std::ostream& operator<<(std::ostream& Stream,const YMMOG y);

	protected:
		std::vector<ResponsibleNodeData*> responsible_nodes;
		std::vector<ResponsibleNodeData*>::iterator i;
		std::list<BackupNodeData> backup_nodes;
		std::list<BackupNodeData>::iterator b;
		std::list<IntermediateNodeData> intermediate_nodes;
		std::list<IntermediateNodeData>::iterator in;
	private:

		void sendEvent(int subSpaceId, int playerId, Event event);
		void sendEventList(int serviceType, int subSpaceId, int timeSlot, std::vector<Event> eventList, IPAddress destip);
		void sendRequestIp( OpenRequestIp openRequest);
		void sendPlayerList(int action, int subSpaceId);
		void sendDelay(int subSpaceId, double lastDelay, IPAddress fromIp, IPAddress destip);
		void sendPing( PingMessage* openPing);
		void sendStabilize(int subSpaceId, int nodeType, SubSpace subSpace);

		void requestIntermediateNode(int subSpaceId); 

		void processPingTimer();
		void processPlayerPingTimer();
		void processDelayTimer();

		void processResendTimer();
		void processBandwithTimer();
		void processStabilizeTimer();

		void processSelfTimers(cMessage* msg);
		void processAppMessages(cMessage* msg);

		/** Message Handling **/
		void processLogin(LoginMessage* loginMsg);		
		void processLogout(LogoutMessage* logoutMsg);
		void processCreate(CreateNodeMessage* createMsg);
		void processSubscribe(SubscribeMessage* subscribeMsg);
		void processUnSubscribe(UnSubscribeMessage* unsubscribeMsg);
		void processPublish(PublishMessage* publishMsg);
		void processPublishEventList(PublishEventListMessage* publishMsg);
		void processRequestIp(RequestIpMessage* requestMsg);
		void processDelay(DelayMessage* delayMsg);
		void processReplaceIntermediateNode(ReplaceIntermediateNodeMessage* replaceMsg);
		void processPlayerList(PlayerListMessage* playerListMsg);
		void processRequestNode(RequestNodeMessage* requestMsg);
		void processUpdateRespIp(UpdateRespIpMessage* updateMsg);		
		void processReplacement(ReplacementMessage* replaceMsg);
		void processPing(PingMessage* pingMsg);
		void processStabilize(StabilizeMessage* smsg);

			
		// NodeLog
		std::list<NodeLogEntry> nodeLog;
		/** global **/
		int logLevel;
		double timeSlotSize;
		int maxPlayerLeafs;
		int treeWidth;
		bool sending;
		
		/** Player specific **/
		int player_id;
		int xpos;
		int ypos;
		int lastCheckedX;
		int lastCheckedY;
		int subSpaceSize;
		int gameSpaceSize;
		int currentSubSpaceId;

		double thisNodeBandwith;
		double thisNodeDelay;
		IPAddress lobbyIp;
		std::vector<SubSpace> subscribed_SubSpaces;
		int subscriptionRange;
		int unsubscriptionRange;
		
		void subscribe(int subSpaceId, IPAddress destip);
		void unsubscribe(int subSpaceId, IPAddress destip);
		void unsubscribeAll();
		bool findSubSpace(int id);
		SubSpace* getSubSpace(int id);
		IPAddress findIpThroughNeighbor(int subSpaceId);
		void checkSubscriptions();
		int findNeighborSubSpace(int subSpaceId);
		void requestIpFromNeighbor(int neighborId, int requestedId);
		void removeSubSpace( int subSpaceId );
		void requestIp(int subSpaceId, int nodeType);

		//NodeFunctions
		void informNeighbors(SubSpace subSpace);
		void processTimeSlotEnd();
		void sendTreeMessages(ResponsibleNodeData* respNode, std::list<TreeMessage> msgList );
		

		ResponsibleNodeData* findResponsibleNodeDataById(int subSpaceId);
		BackupNodeData* findBackupNodeDataById(int subSpaceId);
		IntermediateNodeData* findIntermediateNodeDataById( int subSpaceId);

		bool deleteResponsibleNodeData(int id);
		bool deleteIntermediateNodeDataById( int subSpaceId );
		void deleteBackupNodeDataById( int subSpaceId );
		
		
		// Self-Timer
		cMessage* timeSlotTimer;
		cMessage* delayTimer;
		cMessage* playerPingTimer;
		
		int maxFailedPings;
		// Delays
		double delayDelay;
		
		
		// Statistics:
		int computedEvents;
		int maxPlayerOnline;
		
		int savedRequests;
		int passedSubSpaces;
				
		//module references
		SubSpaceTable* magicLog;
};

#endif
