//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file UdpOutDevice.cc
 * @author Stephan Krause
 */

#include <netinet/ip.h>
#include <netinet/udp.h>
#include "IPDatagram_m.h"
#include "UDPPacket.h"

#include "UdpOutDevice.h"


Define_Module(UdpOutDevice);

char* UdpOutDevice::encapsulate(cMessage *msg,
		                unsigned int* length,
			        sockaddr** addr,
			        socklen_t* addrlen)
{
    cMessage* payloadMsg = NULL;
    char* payload = NULL;
    sockaddr_in* addrbuf = NULL;

    unsigned int payloadLen;
    
    IPDatagram* IP = check_and_cast<IPDatagram*>(msg);
    // FIXME: Cast ICMP-Messages
    UDPPacket* UDP = dynamic_cast<UDPPacket*>(IP->decapsulate());
    if (!UDP) {
	EV << "Can't parse non-UDP packets (e.g. ICMP) (yet)...\n";
	goto parse_error;
    }
    payloadMsg = UDP->decapsulate();

    // parse payload
    payload = parser->encapsulatePayload(payloadMsg, &payloadLen);
    if (!payload ) {
	EV << "Can't parse packet payload, dropping packet\n";
	goto parse_error;
    }

    if(payloadLen > mtu) {
	EV << "RealworldApp: Encapsulating packet failed: packet too long\n";
	goto parse_error;
    }
    *length = payloadLen;

    // create sockaddr
    addrbuf = new sockaddr_in;
    addrbuf->sin_family = AF_INET;
    addrbuf->sin_port = htons(UDP->destinationPort());
    addrbuf->sin_addr.s_addr = htonl(IP->destAddress().getInt());
    *addrlen = sizeof(sockaddr_in);
    *addr = (sockaddr*) addrbuf;

    delete IP;
    delete UDP;
    delete payloadMsg;
    return payload;

parse_error:
    delete IP;
    delete UDP;
    delete payloadMsg;
    delete payload;
    return NULL;

}

cMessage* UdpOutDevice::decapsulate(char* buf,
		                    uint32_t length,
				    sockaddr* addr,
				    socklen_t addrlen)
{
    if (!addr) {
	opp_error("UdpOutDevice::decapsulate called without providing sockaddr (addr = NULL)");
    }
    if ( addrlen != sizeof(sockaddr_in) ) {
	opp_error("UdpOutDevice::decapsulate called with wron sockaddr length. "
		"Only IPv4 is supported at the moment!");
    }
    sockaddr_in* addrbuf = (sockaddr_in*) addr;

    IPDatagram* IP = new IPDatagram;
    UDPPacket* UDP = new UDPPacket;
    cMessage* payload = 0; 
    const char* myaddr = 0;

    // Parse Payload
    payload = parser->decapsulatePayload( buf, length);
    if (!payload) {
        ev << "UdpOutDevice: Parsing of Payload failed, dropping packet.\n";
	goto parse_error;
    }

    myaddr = parentModule()->parentModule()->submodule("underlayConfigurator", 0)->par("nodeIP").stringValue();

    // Create IP + UDP header
    IP->setSrcAddress( IPAddress( ntohl(addrbuf->sin_addr.s_addr) ));
    IP->setDestAddress( IPAddress( myaddr ));
    IP->setTransportProtocol( IPPROTO_UDP );
    IP->setTimeToLive( 42 ); // Does not matter, packet ends here
    IP->setIdentification( 42 ); // Faked: There is no way to get the real ID
    IP->setMoreFragments( false );
    IP->setDontFragment( true );
    IP->setFragmentOffset( 0 );
    IP->setDiffServCodePoint( 0 ); // Faked...
    IP->setLength( 160 );

    UDP->setSourcePort( ntohs( addrbuf->sin_port ));
    UDP->setDestinationPort(
	    parentModule()->submodule("overlay", 0)->par("localPort").longValue()
    );
    UDP->setByteLength( 8 );

    // Done...
    UDP->encapsulate( payload );
    IP->encapsulate( UDP );
    delete buf;
    delete addr;
    return IP;

parse_error:
    delete IP;
    delete UDP;
    delete addr;
    delete payload;
    return NULL;
}

